@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -48,15 +49,21 @@ SToolBar::SToolBar (const SProperties props)
   xmanagerCludge = true;
   unsigned int i;
   listener = 0;
-  right = true;
   modified = false;
   printing = false;
+  editorLR = true;
+
+  overrideLR = false;
+  embedLR = false;
+  currentLR = true;
+
   fonts = SStringVector (props["yudit.editor.fonts"]);
 
   if (fonts.size()==0) fonts.append ("default");
   fontsizes = SStringVector (props["yudit.editor.fontsizes"]);
   if (fontsizes.size()==0) fontsizes.append ("16");
   inputs = SStringVector (props["yudit.editor.inputs"]);
+  buildDisplayInputs();
   if (inputs.size()==0) inputs.append ("straight");
 
   SString sf = props["yudit.editor.font"];
@@ -74,17 +81,16 @@ SToolBar::SToolBar (const SProperties props)
   if (i==fontsizes.size()) i=0; currentFontSize = i;
 
   /* be int sync with SFormatIndex */
-  linebreaks.append (SS_LB_UNIX); 
-  linebreakStrings.append (translate ("Unix"));
-  linebreaks.append (SS_LB_DOS);
-  linebreakStrings.append (translate ("DOS"));
-  linebreaks.append (SS_LB_MAC);
-  linebreakStrings.append (translate ("MAC"));
-  linebreaks.append (SS_LB_LS);
-  linebreakStrings.append (translate ("LS"));
-  linebreaks.append (SS_LB_PS);
-  linebreakStrings.append (translate ("PS"));
-  currentLineBreak = 0;
+  paragraphBreaks.append (SS_LB_UNIX); 
+  paragraphBreakStrings.append (translate ("Unix"));
+  paragraphBreaks.append (SS_LB_DOS);
+  paragraphBreakStrings.append (translate ("DOS"));
+  paragraphBreaks.append (SS_LB_MAC);
+  paragraphBreakStrings.append (translate ("MAC"));
+  paragraphBreaks.append (SS_LB_PS);
+  paragraphBreakStrings.append (translate ("PS"));
+  currentParagraphSeparator = 0;
+  currentEmbedding = SS_EmbedNone;
 
   /* create the buttons and set the preferred size */
   buttons[SS_OPEN] = new SButton (SIconFactory::getIcon("Open"), "O");
@@ -95,15 +101,19 @@ SToolBar::SToolBar (const SProperties props)
   buttons[SS_GOTO] = new SButton (SIconFactory::getIcon("Goto"), "G"), 
   buttons[SS_UNDO] = new SButton (SIconFactory::getIcon("Undo"), "U");
   buttons[SS_REDO] = new SButton (SIconFactory::getIcon("Redo"), "R");
-  buttons[SS_FINGER] = new SButton (SIconFactory::getIcon("GreenRight"), "D");
+  buttons[SS_SET_OVERRIDE] = new SButton (SIconFactory::getIcon("RLO"), "D");
+
+  buttons[SS_SET_EMBEDDING] = new SButton (SIconFactory::getIcon("RLE"), "E");
+  buttons[SS_YIELD_EMBEDDING] = new SButton (SIconFactory::getIcon("LR"), "Y");
+  buttons[SS_DOCUMENT_EMBEDDING] = new SButton (SIconFactory::getIcon("EmbedNone"), "T");
 
   buttons[SS_FONT] = new SButton ("", SIconFactory::getIcon("Font"));
   setFAccelerator (buttons[SS_FONT], fonts.size(), true);
   buttons[SS_INPUT] = new SButton ("", SIconFactory::getIcon("Right"));
   setFAccelerator (buttons[SS_INPUT], inputs.size(), false);
-  buttons[SS_LINE_BREAK] = new SButton ("", SIconFactory::getIcon("UNIX"));
+  buttons[SS_PARAGRAPH_BREAK] = new SButton ("", SIconFactory::getIcon("UNIX"));
+
 
-  right = true;
   buttons[SS_MAGNIFY_PLUS] = new SButton (
        SIconFactory::getIcon("MagnifyPlus"), "Z");
   buttons[SS_MAGNIFY_MINUS] = new SButton (
@@ -133,9 +143,9 @@ SToolBar::SToolBar (const SProperties props)
     SDimension wp;
     switch  (i)
     {
-    case SS_LINE_BREAK:
-      wp = getMaxSize (buttons[i], linebreakStrings, 
-           linebreakStrings[currentLineBreak]);
+    case SS_PARAGRAPH_BREAK:
+      wp = getMaxSize (buttons[i], paragraphBreakStrings, 
+           paragraphBreakStrings[currentParagraphSeparator]);
       break;
     case SS_FONT:
       wp = getMaxSize (buttons[i], fonts, fonts[currentFont]);
@@ -147,7 +157,7 @@ SToolBar::SToolBar (const SProperties props)
       break;
 */
     case SS_INPUT:
-      wp = getMaxSize (buttons[i], inputs, inputs[currentInput]);
+      wp = getMaxSize (buttons[i], displayInputs, displayInputs[currentInput]);
       break;
     default:
       wp = buttons[i]->getPreferredSize();
@@ -204,7 +214,9 @@ SToolBar::inputChanged (const SStringVector& _inputs, unsigned int _current)
   /* move all buttons */
   inputs = _inputs;
   currentInput = _current;
-  SDimension wp = getMaxSize (cb, inputs, inputs[currentInput]);
+  buildDisplayInputs();
+  
+  SDimension wp = getMaxSize (cb, displayInputs, displayInputs[currentInput]);
   int delta = (int) wp.width - (int) oldSize.width;
   unsigned int i;
   if (delta != 0)
@@ -217,7 +229,39 @@ SToolBar::inputChanged (const SStringVector& _inputs, unsigned int _current)
                buttons[i]->getLocation().y));
     }
   }
-  cb->setText (inputs[currentInput]);
+  cb->setText (displayInputs[currentInput]);
+}
+
+/**
+ * To save display space we don't display the locale at
+ * the end of the x-input-locale input methods.
+ */
+void
+SToolBar::buildDisplayInputs()
+{
+  SStringVector plain; 
+  for (unsigned i=0; i<inputs.size(); i++)
+  {
+    SString s = inputs[i];
+    if (s.size()>2 && (s[0]=='x' || s[0]=='X') && s[1] == '-')
+    {
+      for (unsigned j=2; j<s.size(); j++)
+      {
+        if (s[j] == ':')
+        {
+          SString m = s;
+          m.truncate (j);
+          m.lower();
+          /* This one will have the locale displayed */
+          if (m=="x-utf-8") break;
+          s.truncate (j);
+          break;
+        }
+      }
+    }
+    plain.append (s); 
+  }
+  displayInputs  = plain;
 }
 
 SToolBar::~SToolBar ()
@@ -292,22 +336,22 @@ SToolBar::resize(const SDimension & d)
 
   /**
    * This is XManager bug wrokaround. 
-   * We dont really need to resize these guys.
+   * We don't really need to resize these guys.
    */
   for (unsigned int i=0; i<(unsigned int) SS_MAX; i++)
   {
     SDimension wp;
     switch  (i)
     {
-    case SS_LINE_BREAK:
-      wp = getMaxSize (buttons[i], linebreakStrings, 
-           linebreakStrings[currentLineBreak]);
+    case SS_PARAGRAPH_BREAK:
+      wp = getMaxSize (buttons[i], paragraphBreakStrings, 
+           paragraphBreakStrings[currentParagraphSeparator]);
       break;
     case SS_FONT:
       wp = getMaxSize (buttons[i], fonts, fonts[currentFont]);
       break;
     case SS_INPUT:
-      wp = getMaxSize (buttons[i], inputs, inputs[currentInput]);
+      wp = getMaxSize (buttons[i], displayInputs, displayInputs[currentInput]);
       break;
     default:
       wp = buttons[i]->getPreferredSize();
@@ -337,21 +381,34 @@ SToolBar::buttonPressed (void* source, const SAccelerator* accel)
   {
     if (source == buttons[i])
     {
-      if (i == SS_FINGER)
+      if (i == SS_SET_OVERRIDE)
       {
-        right = !right;
-        /* listener may chage right */
         if (listener) listener->buttonPressed (this, i, accel!=0);
-        buttons[SS_FINGER]->setIcon (SIconFactory::getIcon(
-          right ? "GreenRight" : "GreenLeft"));
+        /* it is listener->buttonPressed that can change it */
+        break;
+      }
+      if (i == SS_SET_EMBEDDING)
+      {
+        if (listener) listener->buttonPressed (this, i, accel!=0);
+        /* it is listener->buttonPressed that can change it */
+        break;
+      }
+      if (i == SS_YIELD_EMBEDDING)
+      {
+        if (listener) listener->buttonPressed (this, i, accel!=0);
+        /* it is listener->buttonPressed that can change it */
         break;
       }
       int indx = getFAccelerator (accel);
       switch (i)
       {
-      case SS_LINE_BREAK:
-        indx = (currentLineBreak+1) % linebreaks.size();
-        setLineBreak (indx);
+      case SS_DOCUMENT_EMBEDDING:
+        indx = (((int)currentEmbedding)+1) % 3;
+        setEmbedding ((SS_Embedding)indx, true);
+        break;
+      case SS_PARAGRAPH_BREAK:
+        indx = (currentParagraphSeparator+1) % paragraphBreaks.size();
+        setParagraphSeparator (indx, true);
         break;
       case SS_INPUT:
         if (accel==0)
@@ -446,7 +503,10 @@ SToolBar::enterWindow (void* source)
        }
       case SS_UNDO: str = translate ("__U__ndo"); break;
       case SS_REDO:  str = translate ("__R__edo"); break;
-      case SS_FINGER: str = translate ("Text __D__irection"); break;
+      case SS_DOCUMENT_EMBEDDING: str = translate ("Document __T__ext Embedding"); break;
+      case SS_SET_OVERRIDE: str = translate ("__D__irection Override"); break;
+      case SS_SET_EMBEDDING: str = translate ("__E__mbedding Override"); break;
+      case SS_YIELD_EMBEDDING: str = translate ("__Y__ield Embedding"); break;
       case SS_MAGNIFY_MINUS: str = translate ("Smaller(__A__)"); break;
       case SS_MAGNIFY_PLUS: str = translate ("Bigger(__Z__)"); break;
       case SS_FONT: str = translate ("Font");
@@ -455,7 +515,7 @@ SToolBar::enterWindow (void* source)
       case SS_INPUT: str = translate ("Input");
         max = (int)inputs.size();
         break;
-      case SS_LINE_BREAK: str = translate ("Line Break"); break;
+      case SS_PARAGRAPH_BREAK: str = translate ("Line Break"); break;
       default:
          break;
       }
@@ -464,14 +524,14 @@ SToolBar::enterWindow (void* source)
         char aa[64];
         if (max > 1)
         {
-           sprintf (aa, "(__%sF1..F%d__)", 
+           sprintf (aa, "%s(__%sF1..F%d__)%s", SS_LB_LRE,
              (i==(unsigned int)SS_FONT) ? "shift-" : "",
-             (max));
+             (max), SS_LB_PDF);
         }
         else
         {
-           sprintf (aa, "(__%sF1__)", 
-             (i==(unsigned int)SS_FONT) ? "shift-" : "");
+           sprintf (aa, "%s(__%sF1__)%s", SS_LB_LRE, 
+             (i==(unsigned int)SS_FONT) ? "shift-" : "", SS_LB_PDF);
         }
         str.append (SString(aa));
       }
@@ -594,7 +654,7 @@ SToolBar::setFont (int index)
   if (index == (int) currentFont) return;
   if (fonts[(unsigned int)index] == "none" || fonts[(unsigned int)index] == "None") return;
   currentFont = index;
-  setButton (SS_FONT, fonts[currentFont]); 
+  setButton (SS_FONT, fonts[currentFont], true); 
 }
 
 void
@@ -605,7 +665,7 @@ SToolBar::setFontSize (int index)
   if (index == (int) currentFontSize) return;
   bool up = (index > (int)currentFontSize);
   currentFontSize = index;
-  setButton (up ? SS_MAGNIFY_PLUS : SS_MAGNIFY_MINUS, fontsizes[currentFontSize]); 
+  setButton (up ? SS_MAGNIFY_PLUS : SS_MAGNIFY_MINUS, fontsizes[currentFontSize], true); 
 }
 
 void
@@ -617,25 +677,34 @@ SToolBar::setInput (int index)
   if (inputs[(unsigned int)index] == "none" 
          || inputs[(unsigned int)index] == "None") return;
   currentInput = index;
-  setButton (SS_INPUT, inputs[currentInput]); 
+  setButton (SS_INPUT, displayInputs[currentInput], true); 
 }
 
 void
-SToolBar::setLineBreak (int index)
+SToolBar::setParagraphSeparator (int index, bool notify)
 {
   if (index < 0) return;
-  if (index >= (int) linebreaks.size()) return;
-  if (index == (int) currentLineBreak) return;
-  currentLineBreak = index;
-  setButton (SS_LINE_BREAK, linebreakStrings[currentLineBreak]); 
+  if (index >= (int) paragraphBreaks.size()) return;
+  if (index == (int) currentParagraphSeparator) return;
+  currentParagraphSeparator = (unsigned int) index;
+  setButton (SS_PARAGRAPH_BREAK, paragraphBreakStrings[currentParagraphSeparator], notify); 
 }
 
 void
-SToolBar::setButton (SButtonIndex index, const SString& txt)
+SToolBar::setEmbedding (SS_Embedding index, bool notify)
+{
+  if (index ==  currentEmbedding) return;
+  currentEmbedding = index;
+  setButton (SS_DOCUMENT_EMBEDDING, SString(), notify); 
+}
+
+void
+SToolBar::setButton (SButtonIndex index, const SString& txt, bool notify)
 {
   switch (index)
   {
-  case SS_LINE_BREAK:
+  case SS_DOCUMENT_EMBEDDING:
+  case SS_PARAGRAPH_BREAK:
   case SS_INPUT:
   case SS_MAGNIFY_PLUS:
   case SS_MAGNIFY_MINUS:
@@ -645,10 +714,10 @@ SToolBar::setButton (SButtonIndex index, const SString& txt)
     return;
   }
   SButton* b = buttons[index];
-  if (index == SS_LINE_BREAK)
+  if (index == SS_PARAGRAPH_BREAK)
   {
     SIcon* icon = 0;
-    switch (currentLineBreak)
+    switch (currentParagraphSeparator)
     {
     case SS_FORMAT_UNIX:
       icon = SIconFactory::getIcon("UNIX"); break;
@@ -656,8 +725,6 @@ SToolBar::setButton (SButtonIndex index, const SString& txt)
       icon = SIconFactory::getIcon("DOS"); break;
     case SS_FORMAT_MAC:
       icon = SIconFactory::getIcon("MAC"); break;
-    case SS_FORMAT_LS:
-      icon = SIconFactory::getIcon("LS"); break;
     case SS_FORMAT_PS:
       icon = SIconFactory::getIcon("PS"); break;
     default:
@@ -665,11 +732,31 @@ SToolBar::setButton (SButtonIndex index, const SString& txt)
     }
     b->setIcon (icon);
   }
-  if (index != SS_MAGNIFY_PLUS && index != SS_MAGNIFY_MINUS)
+  if (index == SS_DOCUMENT_EMBEDDING)
+  {
+    SIcon* icon = 0;
+    switch (currentEmbedding)
+    {
+    case SS_EmbedNone:
+      icon = SIconFactory::getIcon("EmbedNone"); break;
+    case SS_EmbedLeft:
+      icon = SIconFactory::getIcon("EmbedLeft"); break;
+    case SS_EmbedRight:
+      icon = SIconFactory::getIcon("EmbedRight"); break;
+    default:
+      break;
+    }
+    b->setIcon (icon);
+  }
+  if (index != SS_MAGNIFY_PLUS 
+     && index != SS_MAGNIFY_MINUS && index != SS_DOCUMENT_EMBEDDING)
   {
     b->setText (txt);
   }
-  if (listener) listener->buttonPressed (this, (unsigned int) index, true);
+  if (notify && listener)
+  {
+    listener->buttonPressed (this, (unsigned int) index, true);
+  }
 }
 
 void
@@ -705,3 +792,30 @@ SToolBar::setPrinting (bool flag)
   }
 }
 
+/**
+ * The direction has changed in editor - redo ours.
+ */
+void
+SToolBar::directionChanged (bool lr)
+{
+  if (overrideLR == lr)
+  {
+    buttons[SS_SET_OVERRIDE]->setIcon (SIconFactory::getIcon(
+      !overrideLR ? "LRO" : "RLO"));
+  }
+  if (embedLR == lr)
+  {
+   buttons[SS_SET_EMBEDDING]->setIcon (SIconFactory::getIcon(
+      !embedLR ? "LRE" : "RLE"));
+  }
+  if (currentLR != lr)
+  {
+    buttons[SS_YIELD_EMBEDDING]->setIcon (SIconFactory::getIcon(
+      !currentLR ? "LR" : "RL"));
+  }
+
+  overrideLR = !lr;
+  embedLR = !lr;
+  currentLR = lr;
+
+}
