@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -27,8 +28,10 @@
 #include "stoolkit/SIOStream.h"
 #include "swidget/SUniPrint.h"
 #include <swindow/SPrinter.h>
+#include <swindow/SScriptProcessor.h>
 #include <stdlib.h>
 
+static int translateHighlightMode (const SString& hl);
 
 /**
  * A caret that redraws itself differently for lr and rl text 
@@ -39,11 +42,12 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
   freehand = 0;
   freeHandHeight = 0;
   setLayout (SLayout (SLocation (0,0), SLocation (3000,2000)));
+  originalProperties = props;
 
   /* TODO: move this to config file */
   helpStrings.put ("hilfe", "de");
   SEncoder kana("Kana");
-  SEncoder utf8;
+  SEncoder utf8("utf-8-s");
   helpStrings.put (utf8.encode (kana.decode ("HERUPU", false)), "ja");
   SEncoder hung("Hungarian");
   helpStrings.put ("segit", "hu");
@@ -51,13 +55,13 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
   helpStrings.put (utf8.encode (hung.decode ("segi'tse'g", false)), "hu");
   SEncoder yidd("Yiddish");
   SV_UCS4 yi0;
-  yi0.append (SGlyph::RLO);
+  yi0.append (SD_CD_RLO);
   yi0.append (yidd.decode ("gehilF", false));
-  yi0.append (SGlyph::PDF);
+  yi0.append (SD_CD_PDF);
   SV_UCS4 yi1;
-  yi1.append (SGlyph::RLO);
+  yi1.append (SD_CD_RLO);
   yi1.append (yidd.decode ("helF", false));
-  yi1.append (SGlyph::PDF);
+  yi1.append (SD_CD_PDF);
   helpStrings.put (utf8.encode (yi0), "yi");
   helpStrings.put (utf8.encode (yi1), "yi");
   helpStrings.put (utf8.encode (yidd.decode ("gehilF", false)), "yi");
@@ -166,11 +170,14 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
 
   slider = new SSlider ();
   slider->setLayout (sliderLayout);
+  /* on mac somehow it did not want to resize - force it */
+  slider->resize (SDimension (10, 200));
   
   SString currentFont = toolbar->fonts[toolbar->currentFont];
   SString currentFontSize = toolbar->fontsizes[toolbar->currentFontSize];
   SString currentInput = toolbar->inputs[toolbar->currentInput];
-  SString currentLineBreak = toolbar->linebreaks[toolbar->currentLineBreak];
+  SString currentParagraphSeparator = toolbar->paragraphBreaks[toolbar->currentParagraphSeparator];
+  
 
   editor = new STextEdit ();
   currentFontSize.append ((char)0);
@@ -180,6 +187,8 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
 
   editor->setMultiline (true);
   editor->setLineEndMark (true);
+  editor->setDocumentEmbedding (toolbar->currentEmbedding);
+
   editorLayout = 
     SLayout (
       SLocation (2, toobarHeight+2),
@@ -188,11 +197,54 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
       SLocation (100, 100)
     );
   editor->setLayout (editorLayout);
-  editor->setDirection (toolbar->right?'>':'<');
-  command->setDirection (toolbar->right?'>':'<');
 
+  /*SGC */
+  SString inputStyle = originalProperties["yudit.editor.xinputs.style"];
+
+  /* allowed styles */
+  if (inputStyle=="preedit-over-status-under" 
+       || inputStyle == "preedit-under-status-under"
+       || inputStyle=="preedit-over-status-over"
+       || inputStyle == "preedit-root-status-root" )
+  {
+    SProperties p;
+    SColor cbg(originalProperties["yudit.background"]);
+    SColor cfg(originalProperties["yudit.label.foreground"]);
+
+    sprintf (fnc, "%lu,%lu", (unsigned long)cbg.getValue(),
+      (unsigned long)cfg.getValue());
+    p.put ("InputClientColor", fnc);
+
+    SColor sbg(originalProperties["yudit.background"]);
+    SColor sfg(originalProperties["yudit.title.foreground"]);
+
+    sprintf (fnc, "%lu,%lu", (unsigned long)sbg.getValue(),
+      (unsigned long)sfg.getValue());
+    p.put ("InputStatusColor", fnc);
+    editor->setXIMProperties (p);
+
+    p.put ("InputStyle", inputStyle);
+    editor->setXIMProperties (p);
+  }
+  else
+  {
+    fprintf (stderr, "Yudit: unkown yudit.editor.xinputs.style=%*.*s\n",
+       SSARGS(inputStyle));
+  }
+
+  /* showInputStatus has set up the xim properties for editor */
   editor->setInputMethod (currentInput);
-  if (currentInput != "kinput2")  command->setInputMethod (currentInput);
+
+  if (currentInput.size() > 2) 
+  {
+    if (currentInput != SS_KINPUT2_IM  
+       && currentInput != SS_WINDOWS_IM
+       && currentInput[0] != 'x' 
+       && currentInput[1] != 'X')
+    {
+      command->setInputMethod (currentInput);
+    }
+  }
   if (props["yudit.editor.showbreak"]=="true")
   {
     editor->setLineEndMark (true);
@@ -201,7 +253,22 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
   {
     editor->setLineEndMark (false);
   }
-  SEditor e(currentLineBreak);
+  if (props["yudit.editor.wordwrap"]=="true")
+  {
+    editor->setWordWrap (true);
+  }
+  else
+  {
+    editor->setWordWrap (false);
+  }
+  const SString& hl = props["yudit.editor.syntax"];
+  int mode = translateHighlightMode (hl);
+  if (mode >= 0) 
+  {
+    editor->setHighlightMode ((STextView::SS_HighlightMode)mode);
+  }
+
+  SEditor e(currentParagraphSeparator);
   editor->setEditor (e);
   messagebar->setFontSize(df);
   command->setEditor(SEditor());
@@ -241,13 +308,12 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
 
   caretNow = STextIndex (0,0);
   messagebar->setRow(1);
-  messagebar->setColumn (1);
-  editor->caretGoTo(STextIndex(0,0));
+  messagebar->setColumn (0);
+  editor->setCursorIndex(SCursorIndex(0,0));
   editor->addTextEditLS (this);
   command->addTextEditLS (this);
   command->setHistorySize(30);
 
-  originalProperties = props;
   kmapDialog = 0;
   fileDialog = 0;
   textDialog = 0;
@@ -257,12 +323,12 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
 #else 
   SToolBar::SFormatIndex ft=SToolBar::SS_FORMAT_UNIX;
 #endif
-  toolbar->setLineBreak (ft);
+  toolbar->setParagraphSeparator (ft, true);
 
   if (args.size())
   {
     SString comm ("open -yes ");
-    SString enc ((lastencoding.size()==0)?"utf-8":lastencoding);
+    SString enc ((lastencoding.size()==0)?SString("utf-8"):lastencoding);
     SString filename;
     bool help=false;
     unsigned int i;
@@ -277,6 +343,18 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
       {
         help = true;
       }
+      else if (args[i] == SString("-us"))
+      {
+        SScriptProcessor::support(true);
+      }
+      else if (args[i] == SString("-nus"))
+      {
+        SScriptProcessor::support(false);
+      }
+      else if (args[i] == SString("-noinit"))
+      {
+        SScriptProcessor::doInit(false);
+      }
       else if ((args[i] == SString("-d") || args[i] == SString("-geometry")) 
 	      && i+1 < args.size()) {
         i++;
@@ -294,6 +372,7 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
     if (help)
     {
       filename = SString("help");
+      command->putHistory (filename);
     }
     else
     {
@@ -301,16 +380,16 @@ SYudit::SYudit (const SStringVector& args, const SProperties& props)
             completeFileName(filename));
        /* now it comes handy - args starts with 1 */
        unsigned int j;
-       for (j=args.size()-1; j>i; j--)
+       for (j=args.size(); j>i; j--)
        {
           SString ss = composeFileNameCommand (comm, enc, 
-                completeFileName (args[j]));
+                completeFileName (args[j-1]));
+          command->putHistory (ss);
        }
     }
     SStringVector v;
     v.smartSplit (filename);
     load (v, filename);
-    command->putHistory (filename);
   }
 
   toolbar->setFileName(currentFileName, lastencoding);
@@ -373,15 +452,14 @@ SYudit::buttonPressed (void* src, int which, bool accel)
   SString currentFont = toolbar->fonts[toolbar->currentFont];
   SString currentFontSize = toolbar->fontsizes[toolbar->currentFontSize];
   SString currentInput = toolbar->inputs[toolbar->currentInput];
-  SString currentLineBreak = toolbar->linebreaks[toolbar->currentLineBreak];
+  SString currentParagraphSeparator = toolbar->paragraphBreaks[toolbar->currentParagraphSeparator];
   currentFontSize.append ((char)0);
   double df;
   SString pstr;
   SStringVector pstrVector;
   if (sscanf (currentFontSize.array(), "%lf", &df)==0 || df < 2.0) df = 2.0;
   bool ok = false;
-  bool dirc = false;
-  bool dire = false;
+  int dire = false;
   bool wasmodified = toolbar->modified;
  
   switch (which)
@@ -417,6 +495,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
        pstr.append (" ");
        pstr.append (quoteFileName (fileDialog->getFileName()));
        command->setText (pstr);
+       command->setCursorIndex (SCursorIndex(0,10000));
        command->putHistory (pstr);
        pstrVector.smartSplit(pstr);
        load (pstrVector, pstr);
@@ -428,6 +507,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
       SString str = composeFileNameCommand ("open -yes ", 
          lastencoding, currentFileName);
       command->setText(str);
+      command->setCursorIndex (SCursorIndex(0,10000));
     }
     else
     {
@@ -436,6 +516,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
       SString str = composeFileNameCommand (
              "open ", lastencoding, currentFileName);
       command->setText(str);
+      command->setCursorIndex (SCursorIndex(0,10000));
     }
     command->setFocus();
     break;
@@ -471,6 +552,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
        pstr.append (" ");
        pstr.append (quoteFileName (fileDialog->getFileName()));
        command->setText (pstr);
+       command->setCursorIndex (SCursorIndex(0,10000));
        command->putHistory (pstr);
        pstrVector.smartSplit(pstr);
        save (pstrVector, pstr);
@@ -482,6 +564,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
       SString str = composeFileNameCommand (
              "save ", lastencoding, currentFileName);;
       command->setText(str);
+      command->setCursorIndex (SCursorIndex(0,10000));
     }
     command->setFocus();
     break;
@@ -489,12 +572,14 @@ SYudit::buttonPressed (void* src, int which, bool accel)
     messagelabel->setMessage (translate ("find string"), 
        SMessageLabel::SS_INFO);
     command->setText("find ");
+    command->setCursorIndex (SCursorIndex(0,10000));
     command->setFocus();
     break;
   case SToolBar::SS_GOTO:
     messagelabel->setMessage (translate ("go line [column]"), 
        SMessageLabel::SS_INFO);
     command->setText("go ");
+    command->setCursorIndex (SCursorIndex(0,10000));
     command->setFocus();
     break;
   case SToolBar::SS_PRINT:
@@ -503,6 +588,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
     pstr.append ("print ");
     pstr.append (lastprinterOption);
     command->setText(pstr);
+    command->setCursorIndex (SCursorIndex(0,10000));
     command->setFocus();
     break; 
   case SToolBar::SS_PRINT_PREVIEW:
@@ -511,6 +597,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
     pstr.append ("print ");
     pstr.append (lastpreviewOption);
     command->setText(pstr);
+    command->setCursorIndex (SCursorIndex(0,10000));
     command->setFocus();
     break; 
   case SToolBar::SS_UNDO:
@@ -520,6 +607,7 @@ SYudit::buttonPressed (void* src, int which, bool accel)
          SMessageLabel::SS_WARN);
       toolbar->setModified (false);
     }
+    toolbar->setEmbedding(editor->getDocumentEmbedding(), false);
     break;
   case SToolBar::SS_REDO: 
     if (!editor->redo())
@@ -527,24 +615,57 @@ SYudit::buttonPressed (void* src, int which, bool accel)
       messagelabel->setMessage (translate ("Nothing to redo."), 
          SMessageLabel::SS_WARN);
     }
+    toolbar->setEmbedding(editor->getDocumentEmbedding(), false);
     break;
-  case SToolBar::SS_FINGER: 
-    /* if change direction of selected text dont change direction ! */
-    dire = editor->changeDirection (editor->getDirection());
-    dirc = command->changeDirection(command->getDirection());
-    if (dire || dirc)
+  case SToolBar::SS_DOCUMENT_EMBEDDING: 
+    editor->setDocumentEmbedding (toolbar->currentEmbedding);
+    break;
+  case SToolBar::SS_SET_OVERRIDE: 
+    /* change it back */ 
+    dire = editor->changeDirection (toolbar->overrideLR ? SS_DR_LO : SS_DR_RO);
+    if (dire)
     {
-      toolbar->right = !toolbar->right;
+      messagelabel->setMessage("");
+    }
+    else 
+    {
+      /* leave it back */ 
+      if (!wasmodified) toolbar->setModified (true);
       messagelabel->setMessage (
         translate ("Direction of selected text has been set."), 
         SMessageLabel::SS_INFO);
     }
-    else
+    break;
+  case SToolBar::SS_YIELD_EMBEDDING: 
+    dire = editor->changeDirection  (toolbar->currentLR ? SS_DR_L : SS_DR_R);
+    if (dire)
     {
-      if (!wasmodified) toolbar->setModified (false);
       messagelabel->setMessage("");
-      editor->setDirection (toolbar->right?'>':'<');
-      command->setDirection (toolbar->right?'>':'<');
+    }
+    else 
+    {
+      /* leave it back */ 
+      if (!wasmodified) toolbar->setModified (true);
+      messagelabel->setMessage (
+        translate ("Direction of selected text has been re-set."), 
+        SMessageLabel::SS_INFO);
+    }
+    break;
+    
+  case SToolBar::SS_SET_EMBEDDING: 
+    /* change it back */ 
+    dire = editor->changeDirection (toolbar->embedLR ? SS_DR_LE : SS_DR_RE);
+    if (dire)
+    {
+      messagelabel->setMessage("");
+    }
+    else 
+    {
+      /* leave it back */ 
+      if (!wasmodified) toolbar->setModified (true);
+      messagelabel->setMessage (
+        translate ("Direction of selected text has been set."), 
+        SMessageLabel::SS_INFO);
     }
     break;
   case SToolBar::SS_MAGNIFY_PLUS: 
@@ -574,7 +695,10 @@ SYudit::buttonPressed (void* src, int which, bool accel)
     editor->setInputMethod (currentInput);
     showFreeHand (currentInput=="freehand");
    
-    if (currentInput != "kinput2")
+    if (currentInput != SS_KINPUT2_IM
+       && currentInput != SS_WINDOWS_IM
+       && currentInput[0] != 'x' 
+       && currentInput[1] != 'X')
     {
       command->setInputMethod (currentInput);
     }
@@ -582,10 +706,11 @@ SYudit::buttonPressed (void* src, int which, bool accel)
     {
       command->setInputMethod ("straight");
     }
+    /* update glyph info */
+    caretMoved();
     break;
-  case SToolBar::SS_LINE_BREAK: 
-    editor->setLineBreak (currentLineBreak);
-    messagebar->setGlyphInfo (editor->glyphAt (caretNow));
+  case SToolBar::SS_PARAGRAPH_BREAK: 
+    editor->setParagraphSeparator (currentParagraphSeparator, true);
     break;
   default:
     break;
@@ -641,6 +766,7 @@ SYudit::textEntered (void *source)
       messagelabel->setMessage (translate ("Unsaved changes. Enter quits!"), 
          SMessageLabel::SS_ERR);
       command->setText("quit -yes");
+      command->setCursorIndex (SCursorIndex(0,10000));
       command->setFocus();
       return;
     }
@@ -649,6 +775,7 @@ SYudit::textEntered (void *source)
       messagelabel->setMessage (translate ("Pending printing job. Enter quits!"), 
          SMessageLabel::SS_ERR);
       command->setText("quit -yes");
+      command->setCursorIndex (SCursorIndex(0,10000));
       command->setFocus();
       return;
     }
@@ -672,6 +799,11 @@ SYudit::textEntered (void *source)
     load (cv, text);
     return;
   }
+  if (cv.size()>0 && (cv[0] == "configure" || cv[0] == "config"))
+  {
+    load (cv, text);
+    return;
+  }
   if (cv.size()>0 && (cv[0] == "save" || cv[0] == "save"))
   {
     save (cv, text);
@@ -692,9 +824,50 @@ SYudit::textEntered (void *source)
     replace (cv, text);
     return;
   }
-  SString mess ("syntax error: `");
+  if (cv.size() > 0 && cv[0] == "syntax")
+  {
+    int smode = -1;
+    if (cv.size() != 2 || (smode = translateHighlightMode (cv[1])) < 0)
+    {
+      SString err = translate ("usage: ");
+      err.append ("syntax none|simple|simple-dark");
+      messagelabel->setMessage (err, SMessageLabel::SS_ERR);
+      return;
+    }
+    editor->setHighlightMode ((STextView::SS_HighlightMode)smode);
+    userProps.put ("yudit.editor.syntax", cv[1]);
+    editor->setFocus();
+    command->putHistory(text);
+    return;
+  }
+  if (cv.size() > 0 && cv[0] == "wordwrap")
+  {
+    if (cv.size() != 2 || (cv[1] != "true" && cv[1] != "false"))
+    {
+      SString err = translate ("usage: ");
+      err.append ("wordwrap true|false");
+      messagelabel->setMessage (err, SMessageLabel::SS_ERR);
+      return;
+    }
+    if (cv[1]=="true")
+    {
+      editor->setWordWrap (true);
+    }
+    else
+    {
+      editor->setWordWrap (false);
+    }
+    userProps.put ("yudit.editor.wordwrap", cv[1]);
+    editor->setFocus();
+    command->putHistory(text);
+    return;
+  }
+  SString mess ("syntax error: ");
+  mess.append (SS_LB_LRE);
+  mess.append ("'");
   mess.append (text);
   mess.append ("'");
+  mess.append (SS_LB_PDF);
   messagelabel->setMessage (mess, SMessageLabel::SS_ERR);
   command->clear();
   command->historyEnd();
@@ -822,9 +995,12 @@ SYudit::print (const SStringVector& args, const SString& str)
     os = p.getOutputStream();
     if (!os.isOK())
     {
-      SString ems ("Can not execute: `");
+      SString ems ("Can not execute: ");
+      ems.append (SS_LB_LRE);
+      ems.append ("'");
       ems.append (executable);
       ems.append ("'");
+      ems.append (SS_LB_PDF);
       messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
       return;
     }
@@ -835,9 +1011,12 @@ SYudit::print (const SStringVector& args, const SString& str)
     os = f.getOutputStream();
     if (!os.isOK())
     {
-      SString ems ("Can not open: `");
+      SString ems ("Can not open: ");
+      ems.append (SS_LB_LRE);
+      ems.append ("");
       ems.append (filename);
       ems.append ("'");
+      ems.append (SS_LB_PDF);
       messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     }
   }
@@ -845,7 +1024,10 @@ SYudit::print (const SStringVector& args, const SString& str)
   SString fontName = toolbar->fonts[toolbar->currentFont];
   SUniPrint uniPrint(writer, fontName, hsized, fontName, messagebar->fontSize);
   SString text = editor->getText();
+
+  uniPrint.setDocumentEmbedding (toolbar->currentEmbedding);
   uniPrint.setLineEndMark(shownl);
+  uniPrint.setWordWrap (editor->getWordWrap());
   toolbar->setPrinting(true);
 
   command->clear();
@@ -857,7 +1039,7 @@ SYudit::print (const SStringVector& args, const SString& str)
   if (!uniPrint.print (currentFileName, text, true))
   {
       toolbar->setPrinting (false);
-      SString ems ("Printing failed on`");
+      SString ems ("Printing failed on '");
       ems.append (executable);
       ems.append (filename);
       ems.append ("'");
@@ -1003,7 +1185,19 @@ SYudit::load (const SStringVector& args, const SString& str)
     encode = "utf-8";
     filename = internal;
   }
-
+  if (args[0] =="configure" || args[0] =="config")
+  {
+    SString afile = currentFileName;
+    currentFileName = "/dev/null";
+    /* trick to save properties */
+    saveProperties();
+    currentFileName = afile;
+    SString internal = getHome();;
+    internal.append ("/.yudit/yudit.properties");
+    i = args.size();
+    encode = "utf-8";
+    filename = internal;
+  }
 
   if (filename.size()==0)
   {
@@ -1028,12 +1222,18 @@ SYudit::load (const SStringVector& args, const SString& str)
     return;
   }
 
-  SEncoder utf8; 
+  SEncoder utf8("utf-8-s"); 
   SEncoder enc(encode);
   if (!enc.isOK())
   {
     SString ems = translate ("Unknown encoding: ");
-    ems.append ("`"); ems.append (encode); ems.append ("'");
+    
+    ems.append (SS_LB_LRE);
+    ems.append ("'"); 
+    ems.append (encode); 
+    ems.append ("'");
+    ems.append (SS_LB_PDF);
+
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     return;
   }
@@ -1052,7 +1252,11 @@ SYudit::load (const SStringVector& args, const SString& str)
       toolbar->setModified(false);
     }
     SString ems = translate ("Can not read: ");
-    ems.append ("`"); ems.append (filename); ems.append ("'");
+    ems.append (SS_LB_LRE);
+    ems.append ("'"); 
+    ems.append (filename); 
+    ems.append ("'");
+    ems.append (SS_LB_PDF);
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     return;
   }
@@ -1067,38 +1271,62 @@ SYudit::load (const SStringVector& args, const SString& str)
 #else 
   SToolBar::SFormatIndex ft=SToolBar::SS_FORMAT_UNIX;
 #endif
+  bool hadlinebreak = false;
   if (ucs4.size() 
     || ucs4[ucs4.size()-1] == (SS_UCS4) '\r' 
     || ucs4[ucs4.size()-1] == (SS_UCS4) '\n' 
-    || ucs4[ucs4.size()-1] == (SS_UCS4) 0x2029 
-    || ucs4[ucs4.size()-1] == (SS_UCS4) 0x2028)
+    || ucs4[ucs4.size()-1] == (SS_UCS4) 0x2029) 
   { 
     switch (ucs4[ucs4.size()-1])
     {
     case '\n': 
       ft =  (ucs4.size()>1 && ucs4[ucs4.size()-2]=='\r')
          ? SToolBar::SS_FORMAT_DOS : SToolBar::SS_FORMAT_UNIX; 
-         break;
-    case '\r': ft = SToolBar::SS_FORMAT_MAC; break;
-    case 0x2028: ft = SToolBar::SS_FORMAT_LS; break; // LS
-    case 0x2029: ft = SToolBar::SS_FORMAT_PS; break; // PS
+      hadlinebreak = true;
+      break;
+    case '\r': ft = SToolBar::SS_FORMAT_MAC; 
+      hadlinebreak = true;
+      break;
+    case 0x2029: ft = SToolBar::SS_FORMAT_PS;
+      hadlinebreak = true;
+      break;
     }
   }
   else
   {
     ft = SToolBar::SS_FORMAT_UNIX;
   }
-  sft = (toolbar->currentLineBreak != (unsigned int) ft);
+  sft = (toolbar->currentParagraphSeparator != (unsigned int) ft);
   text = utf8.encode (ucs4);
   editor->clear();
   editor->setText(text);
-  editor->caretGoTo(STextIndex(0,0));
+  editor->setCursorIndex(SCursorIndex(0,0));
 
   SString ems = translate ("Opened: ");
-  ems.append ("`"); ems.append (filename); ems.append ("'");
-  if (sft) ems.append (translate (" Line Break:"));
-  if (sft) ems.append (translate (toolbar->linebreakStrings[ft]));
-  if (sft) toolbar->setLineBreak ((int) ft);
+  ems.append (SS_LB_LRE);
+  ems.append ("'"); 
+  ems.append (filename); 
+  ems.append ("'");
+  ems.append (SS_LB_PDF);
+  ems.append (".");
+  if (sft)
+  {
+    ems.append (translate (" Line Break:"));
+    ems.append (translate (toolbar->paragraphBreakStrings[ft]));
+  }
+  if (!hadlinebreak)
+  {
+    ems.append (translate(" Incomplete Last Line..."));
+  }
+
+  if (sft)
+  {
+    toolbar->setParagraphSeparator ((int) ft, false);
+    SString currentParagraphSeparator = 
+         toolbar->paragraphBreaks[toolbar->currentParagraphSeparator];
+    editor->setParagraphSeparator (currentParagraphSeparator, false);
+  }
+  
   toolbar->setModified (false);
   editor->setFocus();
 
@@ -1112,7 +1340,7 @@ SYudit::load (const SStringVector& args, const SString& str)
   const STextIndex* sindex = caretIndex.get (filename);
   if (sindex)
   {
-    editor->caretGoTo(*sindex);
+    editor->setCursorIndex(SCursorIndex(sindex->line, sindex->index));
   }
   editor->setFocus();
 }
@@ -1120,9 +1348,6 @@ SYudit::load (const SStringVector& args, const SString& str)
 void 
 SYudit::save (const SStringVector& args, const SString& str)
 {
-  SString currentLineBreak = toolbar->linebreaks[toolbar->currentLineBreak];
-  editor->setLineBreak (currentLineBreak);
-  messagebar->setGlyphInfo (editor->glyphAt (caretNow));
   SString encode="utf-8";
   SString filename;
   unsigned int i;
@@ -1165,12 +1390,16 @@ SYudit::save (const SStringVector& args, const SString& str)
     return;
   }
 
-  SEncoder utf8; 
+  SEncoder utf8("utf-8-s"); 
   SEncoder enc(encode);
   if (!enc.isOK())
   {
     SString ems = translate ("Unknown encoding: ");
-    ems.append ("`"); ems.append (encode); ems.append ("'");
+    ems.append (SS_LB_LRE);
+    ems.append ("'"); 
+    ems.append (encode); 
+    ems.append ("'");
+    ems.append (SS_LB_PDF);
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     return;
   }
@@ -1179,7 +1408,11 @@ SYudit::save (const SStringVector& args, const SString& str)
   if (!os.isOK())
   {
     SString ems = translate ("Can not write: ");
-    ems.append ("`"); ems.append (filename); ems.append ("'");
+    ems.append (SS_LB_LRE);
+    ems.append ("'"); 
+    ems.append (filename); 
+    ems.append ("'");
+    ems.append (SS_LB_PDF);
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     return;
   }
@@ -1190,25 +1423,19 @@ SYudit::save (const SStringVector& args, const SString& str)
   if (ucs4.size() 
     && ucs4[ucs4.size()-1] != (SS_UCS4) '\r' 
     && ucs4[ucs4.size()-1] != (SS_UCS4) '\n' 
-    && ucs4[ucs4.size()-1] != (SS_UCS4) 0x2029 
-    && ucs4[ucs4.size()-1] != (SS_UCS4) 0x2028)
+    && ucs4[ucs4.size()-1] != (SS_UCS4) 0x2029)
   { 
-    switch (toolbar->currentLineBreak)
-    {
-    case SToolBar::SS_FORMAT_UNIX: ucs4.append ((SS_UCS4)'\n'); break; // unix
-    case SToolBar::SS_FORMAT_DOS: 
-       ucs4.append ((SS_UCS4)'\r'); ucs4.append ((SS_UCS4)'\n'); break; // dos
-    case SToolBar::SS_FORMAT_MAC: ucs4.append ((SS_UCS4)'\r'); break; // mac
-    case SToolBar::SS_FORMAT_LS: ucs4.append ((SS_UCS4)0x2028); break; // LS
-    case SToolBar::SS_FORMAT_PS: ucs4.append ((SS_UCS4)0x2029); break; // PS
-    }
     completed = true;
   }
   text = enc.encode (ucs4);
   if ((text.size() > 0 && !writer.write (text)) || !writer.close())
   {
     SString ems = translate ("Error while writing: ");
-    ems.append ("`"); ems.append (filename); ems.append ("'");
+    ems.append (SS_LB_LRE);
+    ems.append ("'"); 
+    ems.append (filename); 
+    ems.append ("'");
+    ems.append (SS_LB_PDF);
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     return;
   }
@@ -1219,8 +1446,13 @@ SYudit::save (const SStringVector& args, const SString& str)
   command->historyEnd();
 
   SString ems = translate ("Wrote: ");
-  ems.append ("`"); ems.append (filename); ems.append ("'");
-  if (completed) ems.append (translate(" Completed last line."));
+  ems.append (SS_LB_LRE);
+  ems.append ("'"); 
+  ems.append (filename); 
+  ems.append ("'");
+  ems.append (SS_LB_PDF);
+  ems.append (".");
+  if (completed) ems.append (translate(" Incomplete Last Line..."));
   messagelabel->setMessage (ems, 
          (completed) ? SMessageLabel::SS_WARN : SMessageLabel::SS_INFO);
   toolbar->setModified (false);
@@ -1249,7 +1481,19 @@ SYudit::goLineRow (const SStringVector& args, const SString& str)
   int ci=0;
   sscanf (l.array(), "%d", &li);
   sscanf (c.array(), "%d", &ci);
-  editor->caretGoTo(STextIndex((unsigned int)(li-1),(unsigned int) (ci-1)));
+  if (li <= 0)
+  {
+    editor->setCursorIndex(SCursorIndex(0, 0, true));
+  }
+  else if (ci <= 0)
+  {
+    editor->setCursorIndex(SCursorIndex((unsigned int)(li-1), 0, true));
+  }
+  else
+  {
+    editor->setCursorIndex(SCursorIndex((unsigned int)(li-1),
+       (unsigned int) (ci-1), false));
+  }
 }
 
 /**
@@ -1267,9 +1511,11 @@ SYudit::find (const SStringVector& args, const SString& str)
   if (!editor->find (args[1]))
   {
     SString ems = translate ("search string not found.");
+    ems.append (SS_LB_LRE);
     ems.append ("'");
     ems.append (args[1]);
     ems.append ("'");
+    ems.append (SS_LB_PDF);
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
     return;
   }
@@ -1290,9 +1536,11 @@ SYudit::replace (const SStringVector& args, const SString& str)
   if (!editor->replace (args[1], args[2]))
   {
     SString ems = translate ("search string not found.");
+    ems.append (SS_LB_LRE);
     ems.append ("'");
     ems.append (args[1]);
     ems.append ("'");
+    ems.append (SS_LB_PDF);
     messagelabel->setMessage (ems, SMessageLabel::SS_ERR);
   }
   return;
@@ -1315,18 +1563,45 @@ SYudit::focusChanged (void *source, bool in)
 {
 }
 void
-SYudit::caretMoved (void *source, unsigned int line, unsigned int column)
+SYudit::caretMoved (void *source, 
+   unsigned int _line, unsigned int _column, bool _before)
 {
   if (source != editor) return;
-  if (line != caretNow.line) messagebar->setRow(line+1);
-  if (column != caretNow.index) messagebar->setColumn (column+1);
-  if ((caretNow.line != line || caretNow.index != column) 
+  caretMoved();
+}
+
+void 
+SYudit::caretMoved ()
+{
+  STextIndex ti = editor->getCaretArrowFrom ();
+
+  if (ti.line != caretNow.line) messagebar->setRow(ti.line+1);
+  if (ti.index != caretNow.index) messagebar->setColumn (ti.index);
+
+  if ((caretNow.line != ti.line || caretNow.index != ti.index) 
       && !toolbar->printing)
   {
     messagelabel->setMessage("");
   }
-  caretNow = STextIndex (line, column);
-  messagebar->setGlyphInfo (editor->glyphAt (caretNow));
+  caretNow = ti;
+  
+  toolbar->directionChanged (editor->isEmbedStateLR());
+
+  if (ti.index == 0)
+  {
+    messagebar->setGlyphInfo (0, SString());
+  }
+  else
+  {
+    STextIndex tn (ti.line, ti.index-1);
+    const SGlyph* g = editor->glyphAt(tn);
+    SV_UCS4 chars;
+    if (g!=0)
+    {
+      chars = g->getChars();
+    }
+    messagebar->setGlyphInfo (g, editor->encode (chars));
+  }
 }
 
 /**
@@ -1397,6 +1672,11 @@ SYudit::newFileExists (const SString& filename)
 void
 SYudit::saveProperties ()
 {
+  /* do not save proeprties if it is being edited */
+  SString cfile = getHome();;
+  cfile.append ("/.yudit/yudit.properties");
+  if (currentFileName == cfile) return;
+
   //userProps.put ("yudit.default.printer", lastprinterOption);
   //userProps.put ("yudit.default.previewer", lastpreviewOption);
 
@@ -1416,43 +1696,6 @@ SYudit::saveProperties ()
     }
   }
 
-/* dont save it. it makes confusion */
-#if 0
-  createFileDialog ();
-  SStringVector enc = fileDialog->getFileTypes();
-  SStringVector benc = SEncoder::builtin();
-  unsigned int i=0;
-  SBinHashtable<bool> ph;
-  for (i=0; i<benc.size(); i++)
-  {
-    /* this is an input mehtod and it is misleading. */
-    if (benc[i] == "unicode") continue;
-    if (benc[i].match("iso-8859*")) continue;
-    if (benc[i].match("viscii")) continue;
-    if (benc[i].match("wingreek")) continue;
-    if (benc[i].match("tcvn")) continue;
-    if (benc[i].match("rovas")) continue;
-    if (benc[i].match("koi8-*")) continue;
-    if (benc[i].match("cp-1251")) continue;
-    ph.put (benc[i], true);
-  }
-  i=0;
-  while (i<enc.size())
-  {
-    if (ph[enc[i]])
-    {
-      enc.remove (i);
-    }
-    else
-    {
-      i++;
-    }
-  }
-  s = enc.join (",");
-
-  //dont save it.
-  userProps.put ("yudit.editor.filetypes", s);
-#endif
   /* create a property file that has the original properties and 
    * the ones that I changed.
    */
@@ -1483,6 +1726,22 @@ SYudit::saveProperties ()
   /* merge in changes */
   full.merge (changes); 
 
+  /* That stupid old properties header comment. */
+  const SString *pstr = full.get ("yudit.background.#");
+  /* compatibility. remove this. */
+  if (pstr!=0)
+  { 
+    /* Test it */
+    if (pstr->find ("#   Autosave will eliminate all comments and '\\' (line joiner) characters.") > 0)
+    {
+      SString a;
+      a.append ("# From version 2.7.6 Yudit preserves comment lines\n");
+      a.append ("# that start with a '#' and preceed the property  \n");
+      a.append ("# line they refer to. ");
+      full.put ("yudit.background.#", a);
+    }
+  }
+
   if (!::saveProperties (configFile, full))
   {
     SString str = translate ("Could not save preferences");
@@ -1492,7 +1751,6 @@ SYudit::saveProperties ()
     fprintf (stderr, "Could not save properties into %*.*s.\n",
        SSARGS(configFile));
   }
-  fprintf (stderr, "saved `%*.*s'.\n", SSARGS(configFile));
 }
 
 bool
@@ -1508,9 +1766,11 @@ SYudit::dropped (void* p, const SString& mimetype, const SString& data)
        || strncmp (file.array(), "file:", strlen ("file:"))!=0)
     {
       SString msg = translate ("unsupported text/uri: ");
+      msg.append (SS_LB_LRE);
       msg.append ("'");
       msg.append (file);
       msg.append ("'");
+      msg.append (SS_LB_PDF);
       messagelabel->setMessage (msg, SMessageLabel::SS_ERR);
       return false;
     }
@@ -1531,6 +1791,7 @@ SYudit::dropped (void* p, const SString& mimetype, const SString& data)
     cm.append ("-yes ");
     cm.append (file);
     command->setText(cm);
+    command->setCursorIndex (SCursorIndex(0,10000));
     command->setFocus();
   }
   else
@@ -1539,6 +1800,7 @@ SYudit::dropped (void* p, const SString& mimetype, const SString& data)
        SMessageLabel::SS_INFO);
     cm.append (file);
     command->setText(cm);
+    command->setCursorIndex (SCursorIndex(0,10000));
     command->putHistory(cm);
     SStringVector cv; cv.smartSplit (cm);
     load (cv, cm);
@@ -1577,7 +1839,7 @@ SYudit::createFileDialog () // If needed
   unsigned int ib=0;
   for (ib=0; ib<benc.size(); ib++)
   {
-    /* this is an input mehtod and it is misleading. */
+    /* this is an input method and it is misleading. */
     if (benc[ib] == "unicode") continue;
     ph.put (benc[ib], true);
   }
@@ -1585,7 +1847,7 @@ SYudit::createFileDialog () // If needed
   
   for (ib=0; ib<bext.size(); ib++)
   {
-    /* this is an input mehtod and it is misleading. */
+    /* this is an input method and it is misleading. */
     if (bext[ib].match("iso-8859*"))
     {
       ph.put (bext[ib], true);
@@ -1618,6 +1880,10 @@ SYudit::createFileDialog () // If needed
     {
       ph.put (bext[ib], true);
     }
+    else if (bext[ib] == "cp-1250")
+    {
+      ph.put (bext[ib], true);
+    }
   }
  
   SStringVector addEnc;
@@ -1650,6 +1916,7 @@ SYudit::createKMapDialog ()
   if (kmapDialog) return;
   kmapDialog = new SKMapDialog ();
   kmapDialog->setModal (this);
+  kmapDialog->setXInputs (originalProperties["yudit.editor.xinputs"]);
   kmapDialog->setBackground (originalProperties["yudit.background"]);
   kmapDialog->setLabelForeground (originalProperties["yudit.label.foreground"]);
   kmapDialog->setTitleForeground (originalProperties["yudit.title.foreground"]);
@@ -1829,9 +2096,11 @@ SYudit::addMytoolPrefix (SString* executable)
   if (f.size () <= 0) 
   {
     SString message=translate ("Command not found: ");
+    message.append (SS_LB_RLE);
     message.append ("'");
     message.append (pr);
     message.append ("'");
+    message.append (SS_LB_PDF);
     messagelabel->setMessage (message, SMessageLabel::SS_ERR);
     return false;
   }
@@ -1854,9 +2123,11 @@ SYudit::addMytoolPrefix (SString* executable)
     if (f0.size () <= 0) 
     {
       SString message=translate ("Command not found: ");
+      message.append (SS_LB_LRE);
       message.append ("'");
       message.append (pr);
       message.append ("'");
+      message.append (SS_LB_PDF);
       messagelabel->setMessage (message, SMessageLabel::SS_ERR);
       return false;
     }
@@ -1867,9 +2138,11 @@ SYudit::addMytoolPrefix (SString* executable)
   if (!commandExists (v[0]))
   {
     SString message=translate ("Command not found: ");
+    message.append (SS_LB_LRE);
     message.append ("'");
     message.append (v[0]);
     message.append ("'");
+    message.append (SS_LB_PDF);
     messagelabel->setMessage (message, SMessageLabel::SS_ERR);
     return false;
   }
@@ -1879,3 +2152,29 @@ SYudit::addMytoolPrefix (SString* executable)
   executable->insert (0, pr);
   return true;
 }
+
+/**
+ * Translate our string highlight mode to an integer
+ * @return -1 if it can not be translated.
+ */
+static int 
+translateHighlightMode (const SString& hl)
+{
+  if (hl == "none")
+  {
+    return (int) STextView::HM_None;
+  }
+  else if (hl == "simple")
+  {
+    return (int) STextView::HM_Simple;
+  }
+  else if (hl == "simple-dark")
+  {
+    return (int) STextView::HM_SimpleDark;
+  }
+  else if (hl == "html")
+  {
+    return (int) STextView::HM_HTML;
+  }
+  return -1;
+}
