@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -169,7 +170,7 @@ main (int argc, char* argv[])
     fclose (output);
     /* call argv[2]...argc - with this file */
     SStringVector vargs;
-    /* on linux it is foregound by default */
+    /* on linux it is foreground by default */
     for (int i=2; i<argc; i++)
     {
        SString arg (argv[i]);
@@ -185,8 +186,9 @@ main (int argc, char* argv[])
 
     SPipe pipe (vargs.join (" "));
     {
-      SOutputStream os = pipe.getOutputStream();
-      os.close();
+      /* we have to use input stream if we want to wait. Hack in SIO.cpp */
+      SInputStream is = pipe.getInputStream();
+      is.close();
     }
 
     /* wait till it exits */
