@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -73,7 +74,7 @@ typedef enum {
 }  SS_FieldType;
 
 
-char* textMapPrefix[FT_MAX+1]={
+const char* textMapPrefix[FT_MAX+1]={
   "# ", 
   "#\n#-----------------------------------------\n#### THE NAME OF THIS MAP ###.\nNAME=", 
   "\n#--------------- NEW SECTION -------------\nSECTION=", 
@@ -89,7 +90,7 @@ char* textMapPrefix[FT_MAX+1]={
   "#",
   "",
 };
-char* cMapPrefix[FT_MAX+1]={
+const char* cMapPrefix[FT_MAX+1]={
   "/*--- Magic --*/\n", 
   "\n/*--- Map Name --*/\n", 
   "\n/*------------------New section---------------*/\n", 
@@ -890,7 +891,7 @@ SBMap::circle (SBMapType _type, const void *in, int in_size,
   // There is no next.
   stateModel->needReset = 1;
 
-  // This has only one encoder/decoder. Dont spoil output!
+  // This has only one encoder/decoder. Don't spoil output!
   if (stateModel->circleCount==1) return res;
 
   switch (stateModel->out.toWordSize)
@@ -958,7 +959,7 @@ SBMap::encode (int mapIndex, const void *in, int in_size,
 
   SBMapItem* map = maps[mapIndex];
 
-  // We go through the input and do the search byet by byte
+  // We go through the input and do the search byte by byte
   // This sucks. Should not do this. 
   stateModel->in.appendToNet (in, in_size);
   if (stateModel->in.length==0) return SS_ACCEPT;
@@ -1575,7 +1576,7 @@ SBMapItem::convertFromBumap()
   char  fullLine[64];
   int  status;
 
-  // Go thought the array and map everything to everything.
+  // Go through the array and map everything to everything.
   oldBase = base;
   base = 0;
   for (i=highMin; i<=highMax; i++) for (j=lowMin; j<=lowMax; j++)
@@ -1633,7 +1634,7 @@ SBMapItem::add (const unsigned char* _key, int _keySize, int _matchSize,
       const unsigned char* _value, int _valueSize,
       const unsigned char* _comment, int _commentSize)
 {
-  // Use name and commnet. Then convert the rest.
+  // Use name and comment. Then convert the rest.
   if (itemType == SBMapBumap)
   {
     convertFromBumap();
@@ -1674,7 +1675,7 @@ SBMapItem::add (const unsigned char* _key, int _keySize, int _matchSize,
     codeMapType = SS_DYNAMIC;
   }
 
-  // Mae a new buffer.
+  // Make a new buffer.
   if (baseType == SS_STATIC)
   {
     baseBufferSize = (baseSize == 0) ? 32 : (baseSize << 1) + 2;
@@ -2148,10 +2149,10 @@ SBMapItem::getKey (int at, unsigned int *size, unsigned int *matchedSize)
 
 /**
  * find the character at position. This function should be called in sequence
- * - the inner loop that goes trhough the string has been moved out.
+ * - the inner loop that goes through the string has been moved out.
  * It takes care of multiple keys.
  * @param in the input normalized string
- * @param pos the position of the string more to be copyed from - from[0]
+ * @param pos the position of the string more to be copied from - from[0]
  * @from the item from which we search - inclusive
  * @to the item till what which we search - exclusive
  * @returns 
@@ -2159,7 +2160,7 @@ SBMapItem::getKey (int at, unsigned int *size, unsigned int *matchedSize)
  *   <li> MATCH - the input matches exactly no more </li>
  *   <li> MATCH_MORE - the input matches exactly with more </li>
  *   <li> MORE - the input matches with more </li>
- *   <li> REJECT - the input wont match </li>
+ *   <li> REJECT - the input won't match </li>
  *  </ul>
  * In case itemType is SBMapBumap, from will be mapped into 'to'.
  * For straight maps, only bounds are checked.
@@ -2197,7 +2198,7 @@ SBMapItem::find(const unsigned char in, unsigned int pos, int* from, int* to)
     }
     arrayIndex = ((int)high - highMin)
       * (lowMax - lowMin+1) + ((int)in - lowMin);
-    // Array  fo 16 bit integers
+    // Array of 16 bit integers
     mp = &base[arrayIndex<<1];
     *to = TO16 (mp);
     if (*to==0) return REJECT;
@@ -2590,7 +2591,7 @@ int
 SBMap::serializeUMAP (SString* _fd, SFileFormat _format)
 {
   int    w;
-  SBMapItem*  mapItem;
+  SBMapItem*  mapItem =0;
 
   // Serialize header.
   if (sizeof (magicBM) != 16)
@@ -2657,8 +2658,8 @@ SBMap::serializeUMAP (SString* _fd, SFileFormat _format)
   const unsigned char* cvle;
   SS_WORD64 key;  
   SS_WORD64 vle;  
-  unsigned int i;
-  unsigned int j;
+  unsigned int i=0;
+  unsigned int j=0;
   for (i=0; i<mapItem->codeSize; i++)
   {
     ckey = mapItem->getKey (i, &keylen, &matched);
@@ -2704,7 +2705,7 @@ SBMap::serializeUMAP (SString* _fd, SFileFormat _format)
     delete mat[1];
     return -1;
   }
-  // Determine migh and low.
+  // Determine high and low.
   SS_WORD16 high[2][2];
   SS_WORD16 low[2][2];
   high[0][0] = 255; high[0][1] = 0; high[1][0] = 255; high[1][1] = 0;
@@ -2715,7 +2716,7 @@ SBMap::serializeUMAP (SString* _fd, SFileFormat _format)
   static bool warned = false;
   for (j=0; j<2; j++) for (i=0; i<0x10000; i++)
   {
-    /* previously thought I will use the oppisite
+    /* previously thought I will use the opposite
        array - but it could be used only for reversible
        maps. So we have to use OUR OWN array. gaspar */
     //if (i==0) mat[1-j][i] = 0;
@@ -2866,7 +2867,7 @@ SBMapItem::addState(SS_WORD32 oldState, const unsigned char in, unsigned int pos
     from = _from; to = _to;
 
     // If divisable by 2 it means that this is the 
-    // second nibbble and we may have to do some work...
+    // second nibble and we may have to do some work...
     // Determine what STT to pass. That can be MATCH,
     // REJECT or MORE.
     if (newPos&1)
@@ -2906,7 +2907,7 @@ SBMapItem::addState(SS_WORD32 oldState, const unsigned char in, unsigned int pos
       if (found == MORE)
       {
         stt = (oldState >> 30);
-        // Dont downgrade match;
+        // Don't downgrade match;
         if (stt==MATCH)
         {
           stt = oldState;
@@ -3447,7 +3448,7 @@ static int
 write8 (SString* _fd, SS_WORD8 number, SFileFormat _format, const SS_FieldType ftype, const SS_LineEnd lend)
 {
   unsigned char  out[64];
-  char*  pref;
+  const char*  pref;
   switch (_format)
   {
   case SS_BINARY:
@@ -3526,7 +3527,7 @@ write16 (SString* _fd, SS_WORD16 number, SFileFormat _format, const SS_FieldType
 {
   unsigned char  out[64];
   unsigned char  num[2];
-  char*  pref;
+  const char*  pref;
 
   switch (_format)
   {
@@ -3605,7 +3606,8 @@ write32 (SString* _fd, SS_WORD32 number, SFileFormat _format, const SS_FieldType
 {
   unsigned char  out[64];
   unsigned char  num[4];
-  char*  pref;
+  const char*  pref;
+
   switch (_format)
   {
   case SS_BINARY:
@@ -3696,7 +3698,7 @@ SBMap::getDecoderMap (SStringVector* key, SStringVector* value,
     
     if (getType(i) == SBMap_DECODE)
     {
-      /* crucular maps are not supported */
+      /* circular maps are not supported */
       if (found >= 0) return 0;
       found = i;
     }
@@ -3707,6 +3709,23 @@ SBMap::getDecoderMap (SStringVector* key, SStringVector* value,
   SBMapItem* map = maps[(unsigned int)found];
   return map->getDecoderMap (key, value, _size);
 }
+/* SGC For maps with holes */
+unsigned int
+SBMap::getLinearPosition (unsigned int _index, SS_UCS4 key)
+{
+  return maps[_index]->getLinearPosition(key);
+}
+SS_UCS4 
+SBMap::getLinearKey (unsigned int _index, unsigned int position)
+{
+  return maps[_index]->getLinearKey(position);
+}
+SS_UCS4 
+SBMap::getLinearValue (unsigned int _index, unsigned int position)
+{
+  return maps[_index]->getLinearValue(position);
+}
+
 
 /**
  * return key value map to see what decodes to what
@@ -3734,7 +3753,7 @@ SBMapItem::getDecoderMap (SStringVector* key, SStringVector* value,
         SS_UCS4 in = (i << 8) + j;
         unsigned int arrayIndex = ((int)i - highMin)
           * (lowMax - lowMin+1) + ((int)j - lowMin);
-        // Array  fo 16 bit integers
+        // Array of 16 bit integers
         const unsigned char* mp = &base[arrayIndex<<1];
         SS_UCS4 u4out = TO16 (mp);
         if (u4out==0) continue;
@@ -3798,6 +3817,93 @@ SBMapItem::getDecoderMap (SStringVector* key, SStringVector* value,
   return codeSize;
 }
 
+/**
+ * Get the position where key at index is less or equal to input 
+ * argument key. This can be used in interpolated linear maps.
+ * @param key is the input argument key.
+ * @return input position.
+ */
+unsigned int
+SBMapItem::getLinearPosition (SS_UCS4 key)
+{
+  if (itemType!=SBMapNToN) return 0;
+  /* 64 bit map is not supported by getLinearPosition */
+  if (inWordSize>2) return 0;
+  unsigned int bkeySize = 1<<inWordSize;
+  unsigned char* bkey = new unsigned char [bkeySize];
+  CHECK_NEW (bkey);
+  switch (inWordSize)
+  {
+  case 0:
+    if (key>0xff) { delete bkey; return 0;}
+    bkey[0] = key&0xff;
+    break;
+  case 1:
+    if (key>0xffff)  { delete bkey; return 0;}
+    bkey[1] = key&0xff;
+    bkey[0] = (key>>8)&0xff;
+    break;
+  case 2:
+    bkey[3] = key&0xff;
+    bkey[2] = (key>>8)&0xff;
+    bkey[1] = (key>>16)&0xff;
+    bkey[0] = (key>>24)&0xff;
+    break;
+  default:
+    delete bkey;
+    return 0;
+  }
+  unsigned int from = nextSorted(bkey, bkeySize);
+  delete bkey;
+  if (from ==0) return 0;
+  /* nextSorted returns next element  - the insertion point */
+  return (from-1);
+}
+
+/**
+ * Get the key value in SS_UCS4 corresponding to index
+ */
+SS_UCS4 
+SBMapItem::getLinearKey (unsigned int position)
+{
+  if (itemType!=SBMapNToN) return 0;
+  /* 64 bit map is not supported by getLinearKey */
+  if (position >= codeSize || inWordSize>2) return 0;
+  unsigned int len;
+  const unsigned char* key = getKey(position, &len);
+  /* Multiple words are  not supported by getLinearKey */
+  if (len != (unsigned int)(1<<inWordSize)) return 0;
+  SS_UCS4 ret = 0;
+  for (unsigned int i=0; i<len; i++) 
+  {
+    ret = ret << 8;
+    ret += key[i];
+  }
+  return ret;
+}
+
+/**
+ * Get the decoded value in SS_UCS4 corresponding to index
+ */
+SS_UCS4 
+SBMapItem::getLinearValue (unsigned int position)
+{
+  if (itemType!=SBMapNToN) return 0;
+  /* 64 bit map is not supported by getLinearValue */
+  if (position >= codeSize || outWordSize>2) return 0;
+  unsigned int len;
+  const unsigned char* value = getValue(position, &len);
+  /* Multiple words are  not supported by getLinearValue */
+  if (len != (unsigned int)(1<<outWordSize)) return 0;
+  SS_UCS4 ret = 0;
+  for (unsigned int i=0; i<len; i++) 
+  {
+    ret = ret << 8;
+    ret += value[i];
+  }
+  return ret;
+}
+
 /**
  * This routine can be called only if temType==SBMapBumap 
  */
@@ -3811,7 +3917,7 @@ SBMap::decode (SS_UCS2 in)
   if ((int)j<map->lowMin || (int)j>map->lowMax) return 0;
   unsigned int arrayIndex = ((int)i - map->highMin)
      * (map->lowMax - map->lowMin+1) + ((int)j - map->lowMin);
-  // Array  fo 16 bit integers
+  // Array of 16 bit integers
   const unsigned char* mp = &map->base[arrayIndex<<1];
   SS_UCS4 u4out = TO16 (mp);
   return u4out;
@@ -3830,7 +3936,7 @@ SBMap::encode (SS_UCS4 in)
   if ((int)j<map->lowMin || (int)j>map->lowMax) return 0;
   unsigned int arrayIndex = ((int)i - map->highMin)
      * (map->lowMax - map->lowMin+1) + ((int)j - map->lowMin);
-  // Array  fo 16 bit integers
+  // Array of 16 bit integers
   const unsigned char* mp = &map->base[arrayIndex<<1];
   SS_UCS2 u4out = TO16 (mp);
   return u4out;
