@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -171,7 +172,7 @@ SEventBSD::addTimer (SEventSource* s, SEventTarget* t)
   SClient* c = new SClient (s, t, SClient::TIMER);
   CHECK_NEW (c);
 
-  /* Check if there is alreay somebody there...*/
+  /* Check if there is already somebody there...*/
   targetHashtable.put (t, c, false);
   sourceHashtable.put (s, c, false);
   clientHashtable.put((const SString&)c->id, c);
@@ -223,7 +224,7 @@ SEventBSD::addServer (SServerStream* st, SEventTarget* t)
   CHECK_NEW (c);
   nonblockHack (c->value);
 
-  /* Check if there is alreay somebody there...*/
+  /* Check if there is already somebody there...*/
   targetHashtable.put (t, c, false);
   sourceHashtable.put (s, c, false);
   clientHashtable.put ((const SString&)c->id, c);
@@ -251,7 +252,7 @@ SEventBSD::addInput (SInputStream* ins, SEventTarget* t)
   CHECK_NEW (c);
   nonblockHack (c->value);
 
-  /* Check if there is alreay somebody there...*/
+  /* Check if there is already somebody there...*/
   targetHashtable.put (t, c, false);
   sourceHashtable.put (s, c, false);
   clientHashtable.put((const SString&)c->id, c);
@@ -311,7 +312,7 @@ SEventBSD::addOutput (SOutputStream* outs, SEventTarget* t, const SString& m)
   c->data = m;
   
   nonblockHack (c->value);
-  /* Check if there is alreay somebody there...*/
+  /* Check if there is already somebody there...*/
   targetHashtable.put (t, c, false);
   sourceHashtable.put (s, c, false);
   clientHashtable.put((const SString&)c->id,c);
@@ -646,7 +647,7 @@ SEventBSD::next()
   memcpy (&ro, &((SEventDelegate*)delegate)->readFD, sizeof (fd_set)); 
   memcpy (&wo, &((SEventDelegate*)delegate)->writeFD, sizeof (fd_set)); 
 
-//Does not always get the SIGPIPE on LInux Alpha
+//Does not always get the SIGPIPE on Linux Alpha
 //fprintf (stderr, "Before select %d %d\n", FD_ISSET (5, &wo), FD_ISSET (5, &ro));
   //fprintf (stderr, "s=%u %u %u %u\n", 
   //  readVector.size(), writeVector.size(), timeoutVector.size(),
@@ -690,7 +691,7 @@ SEventBSD::next()
     }
     else
     {
-       // FIXME: xomething is wrong with this array.
+       // FIXME: something is wrong with this array.
        // maybe it is the sorted nature.
        SClient* c = timeoutClient[0];
        SEventSource *s=c->source;
@@ -724,7 +725,7 @@ SEventBSD::next()
     }
   }
 
-  // See sockets. Lets see reading...
+  // See sockets. Let's see reading...
   for (i=0; i<readVector.size(); i++)
   {
    int fd = (int)readVector[i];
@@ -776,7 +777,7 @@ SEventBSD::next()
    }
   }
 
-  // See sockets. Lets see writing...
+  // See sockets. Let's see writing...
   for (i=0; i<writeVector.size(); i++)
   {
    int fd = (int)writeVector[i];
@@ -1043,7 +1044,11 @@ initSignals(bool init)
 }
 #else
 
-static  void signalHandler (int signum);
+extern "C"
+{
+  typedef void(*signalHandler_t)(int);
+  static  void signalHandler (int signum);
+};
 
 void
 initSignals(bool init)
@@ -1058,7 +1063,8 @@ initSignals(bool init)
     sigemptyset(&act.sa_mask);
     act.sa_flags = 0;
 
-    act.sa_handler = signalHandler;
+    /* Hack: on some systems signalHandler does not take an int argument */
+    *((signalHandler_t*) &act.sa_handler) = (signalHandler_t) signalHandler;
 #ifdef SA_INTERRUPT
     act.sa_flags |= SA_INTERRUPT;
 #endif
@@ -1076,24 +1082,32 @@ initSignals(bool init)
   }
 }
 
-static  void
-signalHandler (int signum)
-{
-    switch (signum)
-    {
-    case SIGPIPE:
-        fprintf (stderr, "SEventBSD: disconnected - SIGPIPE\n");
-        break;
+/**
+ * Be aware: on some systems signalHandler does not take an int argument
+ */
+extern "C" {
+  static  void
+  signalHandler (int signum)
+  {
+  #if 0
+      switch (signum)
+      {
+      case SIGPIPE:
+          fprintf (stderr, "SEventBSD: disconnected - SIGPIPE\n");
+          break;
 
-    case SIGALRM:
-        fprintf (stderr, "SEventBSD: timeout - SIGALRM\n");
-        break;
-
-    default:
-        fprintf (stderr, "SEventBSD: signal %d\n", signum);
-    }
-    sEventBSDSignal = signum;
-    return;
+      case SIGALRM:
+          fprintf (stderr, "SEventBSD: timeout - SIGALRM\n");
+          break;
+  
+      default:
+          fprintf (stderr, "SEventBSD: signal %d\n", signum);
+      }
+      sEventBSDSignal = signum;
+  #endif
+     sEventBSDSignal = 1;
+     return;
+  }
 }
 
 #endif
