@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -269,7 +270,7 @@ SHWData::reindex()
     image = SFileImage();
     return;
   }
-  /* Version 1 uses mistic letters for directions */
+  /* Version 1 uses mystic letters for directions */
   if (buffer[5] == '1')
   {
     version = 1; 
@@ -417,7 +418,7 @@ SHWData::convert(const SLineCurves& strokes, bool directed)
     {
       unsigned int k=i+3;
       while (k<j && buffer[k] != '\n' && buffer[k] != ' ') k++;
-      /* we are a bit currupt */
+      /* we are a bit corrupt */
       if (k>=j || buffer[k] != ' ')
       {
         break;
@@ -440,7 +441,7 @@ SHWData::convert(const SLineCurves& strokes, bool directed)
   unsigned int vsize = scores.size();
   for (i=0; i<vsize; i++)
   {
-     /* the biggest the score the better it is */
+     /* the bigger the score the better it is */
      unsigned int order = sorted.appendSorted(-scores[i]);
      bestOnes.insert (order, candidates[i]);
   }
@@ -456,7 +457,7 @@ SHWData::convert(const SLineCurves& strokes, bool directed)
      strokes.size(), sorted.size(), fullsize, -sorted[0]/lineCurve32[16].y);
   }
 
-  /* the most netaives come first. */
+  /* the most negatives come first. */
   int delta = sorted[0] * SD_CADIDATE_PERCENT / 100; /* 20 percent, negative */
   if (delta<0) delta = -delta; // always
   for (i=0; i<sorted.size(); i++)
@@ -502,7 +503,7 @@ SHWData::scoreAll (const SLineCurves& strokes, const char* buffer,
         if (debugLevel > 0)
         {
           fprintf (stderr, "Not enough strokes in '%*.*s' expected=%u got=%u\n",
-            to-from, to-from, &buffer[from], strokes.size(), i);
+            (int)(to-from), (int)(to-from), &buffer[from], strokes.size(), i);
         }
         return -1;
       }
@@ -514,7 +515,7 @@ SHWData::scoreAll (const SLineCurves& strokes, const char* buffer,
         if (debugLevel > 0)
         {
           fprintf (stderr, "Not enough strokes in '%*.*s' expected=%u got=%u\n",
-            to-from, to-from, &buffer[from], strokes.size(), i);
+            (int)(to-from), (int)(to-from), &buffer[from], strokes.size(), i);
         }
         return -1;
       }
@@ -528,7 +529,7 @@ SHWData::scoreAll (const SLineCurves& strokes, const char* buffer,
           if (debugLevel > 0)
           {
             fprintf (stderr, "Bad syntax in '%*.*s' strokes=%u\n",
-              to-from, to-from, &buffer[from], strokes.size());
+              (int)(to-from), (int)(to-from), &buffer[from], strokes.size());
           }
           return -1;
         }
@@ -552,7 +553,7 @@ SHWData::scoreAll (const SLineCurves& strokes, const char* buffer,
     }
     
     /*
-     * Dont add ranks up. Propbabilities are multiplied.
+     * Don't add ranks up. Probabilities are multiplied.
      * gsinai@yudit.org 2001-11-04
      */
     currentLength += projectedLength;
@@ -592,7 +593,7 @@ SHWData::fitOneStroke (const SLineCurve& stroke,  unsigned int strokefrom,
   /* Should not happen unless we have a corrupt guide */
   if (guideend-guidestart < 1 || stroke.length()==0) return 0; 
 
-  /* End of recursize loop - nod dividable any more */
+  /* End of recursize loop - not dividable any more */
   if (strokeend-strokefrom < 2 || guideend-guidestart==1)
   {
     if (strokeend-strokefrom < 1) return 0;
@@ -690,7 +691,7 @@ SHWData::getFilterValue (const SLineCurves& strokes,
   strokeIndex[0] = strokeIndex[1] = 0;
   bool must = false;
 
-  unsigned index=0; /* 0 or 1 */
+  unsigned idx=0; /* 0 or 1 */
   int currentLength = 0;
   int minLength = 0;
   int maxLength = 0;
@@ -706,7 +707,7 @@ SHWData::getFilterValue (const SLineCurves& strokes,
     case 'a':
     case 'b':
     case 'l':
-      oparand [index] = filter[i];
+      oparand [idx] = filter[i];
       break;
     case '0':
     case '1':
@@ -718,24 +719,24 @@ SHWData::getFilterValue (const SLineCurves& strokes,
     case '7':
     case '8':
     case '9': 
-      strokeIndex[index] = strokeIndex[index] * 10 + (filter[i] - '0');
+      strokeIndex[idx] = strokeIndex[idx] * 10 + (filter[i] - '0');
       break;
     case '-': /* next operand */
-      index = 1;
+      idx = 1;
       break;
     case '!': /* filter must be applied */
       must = true;
       break;
     case ' ':
-      if (index==0) break;
+      if (idx==0) break;
     default:
-      if (index != 1 || strokeIndex[0] > strokes.size() || strokeIndex[0] == 0
+      if (idx != 1 || strokeIndex[0] > strokes.size() || strokeIndex[0] == 0
          || strokeIndex[1] > strokes.size() || strokeIndex[1] == 0)
       {
         if (debugLevel > 0)
         {
           fprintf (stderr, "Filter syntax error in %*.*s at '%c'\n", 
-               filterSize, filterSize, filter, filter[i]);
+               (int)filterSize, (int)filterSize, filter, filter[i]);
         }
         return 0;/* neutral */
       }
@@ -750,12 +751,12 @@ SHWData::getFilterValue (const SLineCurves& strokes,
       oparand[0] = oparand[1] = 0;
       strokeIndex[0] = strokeIndex[1] = 0;
       must = false;
-      index = 0;
+      idx = 0;
     }
     /* end case */
   }
   /* last filter */
-  if (index == 1)
+  if (idx == 1)
   {
       /* apply filter */
       length = applyFilter(strokes, oparand, strokeIndex, must);
