@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -578,22 +579,28 @@ SFile::getOutputStream()
  * Find the file by the path..
  * @parma filen is the base filename. 
  * @param pt is the list of paths it could be found.
+ * @aram lookcwd is true if we also need to look into current working dir
  */
-SFile::SFile (const SString& filen, const SStringVector& pt): SIO(SEventSource::FILE) 
+SFile::SFile (const SString& filen, const SStringVector& pt, bool lookcwd): SIO(SEventSource::FILE) 
 {
+  name = "";
   /* Serach for file */
-  name = filen;
   for (unsigned int i=0; i<pt.size() && size() < 0; i++)
   {
     SString newName = pt[i];
     newName.append ("/");
     newName.append (filen);
     newName.replaceAll("//", "/");
+    if (!lookcwd && newName == filen)
+    {
+      name = "";
+      continue;
+    }
     name = newName;
   }
-  if (size() < 0)
+  if (size() < 0 && lookcwd)
   {
-    name = filen;
+    name =  filen;
   }
 }
 
@@ -639,8 +646,12 @@ SFile::getFileImage(bool writeflag)
     return map;
   }
   SFileImage newMap((long) fd, s, writeflag);
-  delete fileName;
   map = newMap;
+  if (map.size() < 0)
+  {
+    fprintf (stderr, "SIO: Open succeeded but mmap failed for %s\n", fileName);
+  }
+  delete fileName;
   return map;
 }
 
@@ -674,6 +685,40 @@ SFile::size()
   return (long) buf.st_size;
 }
 
+bool
+SFile::truncate (long _size)
+{
+  SString newName = name;
+#ifdef USE_WINAPI
+  newName.replaceAll("/", "\\");
+#endif
+  char* cname = newName.cString();
+  int d = open (cname, O_WRONLY | O_CREAT | O_APPEND | S_MYOFLAGS, 0666);
+  if (d <=0)
+  {
+    delete cname;
+    return false;
+  }
+#ifdef USE_WINAPI
+  if (::_chsize (d, _size) != 0)
+  {
+    close (d);
+    delete cname;
+    return false;
+  }
+#else
+  if (::ftruncate (d, _size) != 0)
+  {
+    close (d);
+    delete cname;
+    return false;
+  }
+#endif
+  close (d);
+  delete cname;
+  return true;
+}
+
 /**
  * Destruct the file.
  */
@@ -786,7 +831,7 @@ SSocket::openSocket(SOutputStream* o)
   char* hstring = host.cString();
 
   // We could use saddr.s_addr = inet_addr(address);
-  // But we dont want at this stage
+  // But we don't want at this stage
   if (sscanf (hstring, "%d.%d.%d.%d", 
      &address[0], &address[1], &address[2], &address[3])!=4)
   {
@@ -1401,8 +1446,12 @@ SPipe::openPipe(bool output)
 {
 #ifdef USE_WINAPI
    HANDLE pp[2];
-   
-   if (!CreatePipe (&pp[0], &pp[1], 0, 0))
+   SECURITY_ATTRIBUTES sa;
+   memset (&sa, 0, sizeof (sa)); 
+   sa.nLength = sizeof (sa);
+   sa.lpSecurityDescriptor = 0; 
+   sa.bInheritHandle = 0; /* without this pipe would close */
+   if (!CreatePipe (&pp[0], &pp[1], &sa, 0))
    {
      fprintf (stderr, "openPipe failed.\n");
      return -1;
@@ -1411,53 +1460,65 @@ SPipe::openPipe(bool output)
    memset (&sinfo, 0, sizeof (sinfo));
    sinfo.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
    sinfo.wShowWindow = SW_SHOWDEFAULT;
+   HANDLE hProcess = GetCurrentProcess();
+   HANDLE dupedHandle = INVALID_HANDLE_VALUE; 
+   HANDLE myHandle = INVALID_HANDLE_VALUE; 
    /* ugly but works...*/
    if (output)
    {
-     sinfo.hStdInput = pp[0];
-     sinfo.hStdOutput = 0;
-     sinfo.hStdError = 0;
+     /* dup the read side, make it inheritible, and close the original */
+     myHandle = pp[1];
+     DuplicateHandle(hProcess, pp[0], hProcess, &dupedHandle,
+                0, TRUE, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE);
+     sinfo.hStdInput = dupedHandle;
+     sinfo.hStdOutput = INVALID_HANDLE_VALUE;
+     sinfo.hStdError = INVALID_HANDLE_VALUE;
    }
    else
    {
-     sinfo.hStdInput = 0;
-     sinfo.hStdOutput = pp[1];
-     sinfo.hStdError = 0;
+     /* dup the read side, make it inheritible, and close the original */
+     myHandle = pp[0];
+     DuplicateHandle(hProcess, pp[1], hProcess, &dupedHandle,
+                0, TRUE, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE);
+     sinfo.hStdInput = INVALID_HANDLE_VALUE;
+     sinfo.hStdOutput = dupedHandle;
+     sinfo.hStdError = INVALID_HANDLE_VALUE;
    }
 
    PROCESS_INFORMATION* pinfo = new PROCESS_INFORMATION[1];
 
-   // TODO: nt
-   //CreateProcess (0, "cmd.exe"...
-   //char * cm = isWindows() ? "command.com" : "cmd.exe";
-   //SString commandline (cm);
-   //commandline.append (" /c ");
-   //commandline.append (command);
    SString commandline (command);
-//fprintf (stderr, "CreateProcess: [%*.*s]\n", SSARGS(commandline));
    char* cl = commandline.cString();
-   
-   if (!CreateProcess (NULL, cl, NULL, NULL, TRUE, 
-        //CREATE_DEFAULT_ERROR_MODE | CREATE_NEW_CONSOLE,
-        CREATE_DEFAULT_ERROR_MODE | DETACHED_PROCESS,
-        NULL, NULL, &sinfo, pinfo))
+  
+   /* create process with inherit handle */ 
+   if (!CreateProcess (
+        NULL, // module name 
+        cl, // command line 
+        NULL, // not inheritable handles
+        NULL, // not inheritable thread handle
+        TRUE, // handles not inherited 
+        CREATE_DEFAULT_ERROR_MODE | DETACHED_PROCESS, // no console
+        NULL, // use parent's environment
+        NULL, // use parent's starting directory
+        &sinfo, // STARTUP_INFO
+        pinfo)) // PROCESS_INFO
    {
      fprintf (stderr, "Could not create process [%s]\n", cl);
-     CloseHandle (pp[0]);
-     CloseHandle (pp[1]);
+     CloseHandle (myHandle);
+     CloseHandle (dupedHandle);
      delete pinfo;
      if (cl) delete cl;
      return -1;
    }
-   waitHandles.append ((long)pinfo);
    if (cl) delete cl;
-   if (output)
-   {
-     CloseHandle (pp[0]);
-     return (long) pp[1];
-   }
-   CloseHandle (pp[1]);
-   return (long) pp[0];
+   /* we need to close this because it got duped */
+   CloseHandle (dupedHandle);
+
+   /* wait till program starts */
+   // wont link ::WaitForInputIdle (pinfo->hProcess, 5000);
+
+   waitHandles.append ((long)pinfo);
+   return (long) myHandle;
 #else
    int pp[2];
    if (pipe (pp) == -1)
