@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -641,7 +642,7 @@ SUniMap::decode (const SString& in)
     // Provision for the bad maps.
     else 
     {
-      fprintf (stderr, "SUniMap:: decode bad map zero matched!\n");
+      //fprintf (stderr, "SUniMap:: decode bad map zero matched!\n");
       ucs4vOut.append ((SS_UCS4)buffer[i]);
       i++;
     }
@@ -824,7 +825,7 @@ SUniMap::lift (const SV_UCS4& in, unsigned int inindex,
   if (!ok) return inindex;
   if (delegate==0) return inindex;
 
-  /* make sure we dont have too big values */
+  /* make sure we don't have too big values */
   SBMap *map = ((SUniMapDelegate*) delegate)->map;
 
   int mindex = isdecode ? dindex : eindex;
@@ -874,7 +875,7 @@ SUniMap::getDecoderMap (SStringVector* key, SStringVector* value,
         unsigned int _size)
 {
   if (delegate==0) return 0;
-  /* make sure we dont have too big values */
+  /* make sure we don't have too big values */
   SBMap *map = ((SUniMapDelegate*) delegate)->map;
   return map->getDecoderMap (key, value, _size);
 }
@@ -900,6 +901,56 @@ SUniMap::isClustered() const
   return ((((SUniMapDelegate*) delegate)->map)->mapType==4);
 }
 
+  /* For maps with holes */
+unsigned int
+SUniMap::getDecodePosition (SS_UCS4 key)
+{
+  if (dindex < 0) return 0;
+  SUniMapDelegate* d = (SUniMapDelegate*) delegate;
+  SBMap *map = d->map;
+  return map->getLinearPosition((unsigned int)dindex, key);
+}
+SS_UCS4
+SUniMap::getDecodeKey (unsigned int position)
+{
+  if (dindex < 0) return 0;
+  SUniMapDelegate* d = (SUniMapDelegate*) delegate;
+  SBMap *map = d->map;
+  return map->getLinearKey((unsigned int)dindex, position);
+}
+SS_UCS4
+SUniMap::getDecodeValue (unsigned int position)
+{
+  if (dindex < 0) return 0;
+  SUniMapDelegate* d = (SUniMapDelegate*) delegate;
+  SBMap *map = d->map;
+  return map->getLinearValue((unsigned int)dindex, position);
+}
+
+unsigned int
+SUniMap::getEncodePosition (SS_UCS4 key)
+{
+  if (eindex < 0) return 0;
+  SUniMapDelegate* d = (SUniMapDelegate*) delegate;
+  SBMap *map = d->map;
+  return map->getLinearPosition((unsigned int)eindex, key);
+}
+SS_UCS4
+SUniMap::getEncodeKey (unsigned int position)
+{
+  if (eindex < 0) return 0;
+  SUniMapDelegate* d = (SUniMapDelegate*) delegate;
+  SBMap *map = d->map;
+  return map->getLinearKey((unsigned int)eindex, position);
+}
+SS_UCS4
+SUniMap::getEncodeValue (unsigned int position)
+{
+  if (eindex < 0) return 0;
+  SUniMapDelegate* d = (SUniMapDelegate*) delegate;
+  SBMap *map = d->map;
+  return map->getLinearValue((unsigned int)eindex, position);
+}
 
 /**
  * Convert mys length to max value.
@@ -908,7 +959,7 @@ SUniMap::isClustered() const
 static SS_UCS4
 getMaxBytes (unsigned int inp)
 {
-  /* FIXME: maxBytes algorithm is fixe here. please fix it
+  /* FIXME: maxBytes algorithm is fixed here. please fix it
    * in decode and encode - look for maxBytes.
    */
   switch (inp)
@@ -917,6 +968,6 @@ getMaxBytes (unsigned int inp)
   case 1: return (0xffff);
   case 2: return (0xffffffff);
   }
-  /* dont support 64 bit */
+  /* don't support 64 bit */
   return 0;
 }
