@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -269,20 +270,8 @@ saveProperties (const SString& file, const SProperties& out)
 
   SWriter writer(os);
   if (!writer.isOK()) return false;
-  SString sh("#Yudit-");
-  sh.append (SD_YUDIT_VERSION);
-  sh.append (" properties\n#\n");
-  sh.append ("# This file is automatically saved when the application exits.\n");
-  sh.append ("# Format\n");
-  sh.append ("#   property=value\n");
-  sh.append ("#   property=listelement1,listelement2\n");
-  sh.append ("# Entries should not contain any white space.\n");
-  sh.append ("# Character '\\' is a line joiner when it stands at the end of line.\n");
-  sh.append ("#   Autosave will eliminate all comments and '\\' (line joiner) characters.\n");
-  sh.append ("#\n");
-  if (!writer.write (sh)) return false;
   SString s(orig.toString());
-  s.append ("\n#end of Yudit-");
+  s.append ("\n# End of Yudit-");
   s.append (SD_YUDIT_VERSION);
   s.append (" properties\n");
   if (!writer.write (s)) return false;
@@ -366,7 +355,7 @@ SString translate (const SString& str)
     } 
     SString s (currentBind);
     s.append (".mo");
-    SFile file = SFile (s, plist);
+    SFile file = SFile (s, plist, false);
     moFileImage = file.getFileImage();
   }
   if (moFileImage.size()<=0) return SString(str);
@@ -514,7 +503,7 @@ unicodeValueOf (const SString& str)
 unsigned long
 ss_sqrtlong (unsigned long sq)
 {
-  /* we dont check for negative. assume...*/
+  /* we don't check for negative. assume...*/
   if (sq<2) return sq; /* zero or 1 */
 
   register unsigned long x=sq;
@@ -568,7 +557,7 @@ static long maxsqrtlong=0;
 
 /**
  * Get the maximum long value on the machine. 
- * The hard way.  Dont trust header files.  Gaspar Sinai 
+ * The hard way.  Don't trust header files.  Gaspar Sinai 
  */
 long getMaxLong()
 {
@@ -600,7 +589,7 @@ static long maxint=0;
 static long maxsqrtint=0;
 /**
  * Get the maximum long value on the machine. 
- * The hard way.  Dont trust header files.  Gaspar Sinai 
+ * The hard way.  Don't trust header files.  Gaspar Sinai 
  */
 int getMaxInt()
 {
@@ -636,8 +625,19 @@ isWindows ()
   ovi.dwOSVersionInfoSize = sizeof (ovi);
   GetVersionEx (&ovi);
   return (ovi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS);
-#endif /*USE_WINAPI*/
+#else /*USE_WINAPI*/
   return false;
+#endif
+}
+
+bool
+isWinAPI ()
+{
+#ifdef USE_WINAPI
+  return true;
+#else /*USE_WINAPI*/
+  return false;
+#endif
 }
 
 /**
