@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -21,6 +22,7 @@
 
 #include <swidget/SEditor.h>
 #include <swidget/SEditorIF.h>
+#include <stoolkit/SEncoder.h>
 
 /**
  * All editors are like this one.
@@ -28,7 +30,7 @@
 SEditor::SEditor(void)
 {
   cleared = false;
-  lineBreak = "\n";
+  paragBreak = "\n";
   editorIF = 0;
   vimode = false;
   undoIndex = 0;
@@ -39,7 +41,7 @@ SEditor::SEditor(void)
 SEditor::SEditor (const SString& lb)
 {
   cleared = false;
-  lineBreak = lb;
+  paragBreak = lb;
   editorIF = 0;
   vimode = false;
   undoIndex = 0;
@@ -70,7 +72,7 @@ SEditor::SEditor (const SEditor& in)
   inputMethods = in.inputMethods;
   fonts = in.fonts;
   undoIndex = 0;
-  lineBreak = in.lineBreak;
+  paragBreak = in.paragBreak;
 }
 
 /**
@@ -84,7 +86,7 @@ SEditor::operator=(const SEditor& in)
   inputMethods = in.inputMethods;
   fonts = in.fonts;
   undoIndex = 0;
-  lineBreak = in.lineBreak;
+  paragBreak = in.paragBreak;
   return *this;
 }
 
@@ -99,6 +101,9 @@ SEditor::setInterface(SEditorIF* _editorIF)
  editorIF = _editorIF; 
 }
 
+
+
+
 /**
  * called when a key has been pressed.
  */
@@ -107,24 +112,74 @@ SEditor::keyPressed (SWindowListener::SKey key, const SString& s,
             bool ctrl, bool shift, bool meta)
 {
   if (editorIF == 0) return;
-  if (key==SWindowListener::Key_Meta_R || key == SWindowListener::Key_Meta_L
-   || key == SWindowListener::Key_Alt_R || key==SWindowListener::Key_Alt_L)
+
+  if (key == SWindowListener::Key_Shift_R || 				key==SWindowListener::Key_Meta_L ||
+key==SWindowListener::Key_Meta_R ||
+key==SWindowListener::Key_Control_R ||
+key==SWindowListener::Key_Control_L ||  
+key==SWindowListener::Key_Shift_L)
   {
-    editorIF->deselectText();
-    editorIF->startSelect();
-    return;
+     return;
   }
+  
   /* Whoops - we did not receive keyrelease */
-  if (editorIF->isSelecting() && !meta)
+  //Enable selection with SHIFT, 
+  // Addition by Maarten van Gompel <proycon@anaproy.homeip.net>
+  if (editorIF->isSelecting() && !meta && !shift && !ctrl ) 
   {
-    editorIF->endSelect();
+   editorIF->endSelect();  
   }
-  SString string = s;
-  STextIndex b = editorIF->getCaret();
-  STextIndex a;
-  STextIndex ba;
+  SString erasedString ;
+  SCursorIndex b = editorIF->getCursorIndex();
+  SCursorIndex a;
+  SCursorIndex ba;
   SUndo::SType std;
+  SS_Embedding embd =  editorIF->getDocumentEmbedding();
   cleared = false;
+  SString lineBreak = paragBreak;
+  if (ctrl||meta) lineBreak = SS_LB_LS;
+  
+  /* If there's a selection, erase it first before adding new input*/
+  if ((key  != SWindowListener::Key_F1) &&
+    (key  != SWindowListener::Key_F2) &&
+    (key  != SWindowListener::Key_F3) &&
+    (key  != SWindowListener::Key_F4) &&
+    (key  != SWindowListener::Key_F5) &&
+    (key  != SWindowListener::Key_F6) &&
+    (key  != SWindowListener::Key_F7) &&
+    (key  != SWindowListener::Key_F8) &&
+    (key  != SWindowListener::Key_F9) &&
+    (key  != SWindowListener::Key_F10) &&
+    (key  != SWindowListener::Key_F11) &&
+    (key  != SWindowListener::Key_F12) &&
+    (key  != SWindowListener::Key_Control_R) &&
+    (key  != SWindowListener::Key_Control_L) &&
+    (key  != SWindowListener::Key_Meta_R) &&
+    (key  != SWindowListener::Key_Meta_L) &&
+    (key  != SWindowListener::Key_Alt_R) &&
+    (key  != SWindowListener::Key_Alt_L) &&
+    (key  != SWindowListener::Key_Left) &&
+    (key  != SWindowListener::Key_Right) &&
+    (key  != SWindowListener::Key_Up) &&
+    (key  != SWindowListener::Key_Down) &&  
+    (key  != SWindowListener::Key_Prior) &&
+    (key  != SWindowListener::Key_Next) &&
+    (key  != SWindowListener::Key_Home) &&
+    (key  != SWindowListener::Key_End) &&  
+    (key  != SWindowListener::Key_Escape) &&  
+    (!ctrl) && (!meta))
+  {
+    std = SUndo::SS_EraseSelect;
+    STextIndex mba = editorIF->getSelectedIndex(true);
+    ba = SCursorIndex (mba.line, mba.index);
+    erasedString = editorIF->eraseSelectedText();
+
+    if (erasedString.size() != 0)
+    {
+      a = editorIF->getCursorIndex();
+      add (SUndo(std, editorIF->getDirection (), embd, erasedString, b, a, ba));
+    } 
+  }
   switch (key)
   {
   case SWindowListener::Key_Escape:
@@ -139,121 +194,124 @@ SEditor::keyPressed (SWindowListener::SKey key, const SString& s,
     }
     return;
   case SWindowListener::Key_Home:
-    editorIF->caretGoTo(STextIndex(0,0));
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
+    if (ctrl || meta)
+    {
+      editorIF->setCursorIndex(SCursorIndex(0,0));
+      if (editorIF->isSelecting())
+      {
+        if (meta)
+        {
+            editorIF->selectText (editorIF->getCursorIndex());
+        }
+        else // ctrl should just go to end without selecting.
+        {
+           editorIF->deselectText();
+        }
+      }
+    }
+    else
+    {
+      SCursorIndex ti = editorIF->getCursorIndex();
+      ti.textIndex.index = 0;
+      ti.before = true;
+      editorIF->setCursorIndex(ti);
+      if (editorIF->isSelecting())
+      {
+         editorIF->selectText (editorIF->getCursorIndex());
+      }
+    }
     break;
   case SWindowListener::Key_End:
-    /* I live this hack */
-    editorIF->caretGoTo(STextIndex(100000000,0));
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
+    if (ctrl || meta)
+    {
+      editorIF->setCursorIndex(SCursorIndex(100000000,100000000));
+      if (editorIF->isSelecting())
+      {
+        if (meta)
+        {
+           editorIF->selectText (editorIF->getCursorIndex());
+        }
+        else // ctrl should just go to end without selecting.
+        {
+           editorIF->deselectText();
+        }
+      }
+    }
+    else
+    {
+      SCursorIndex ti = editorIF->getCursorIndex();
+      ti.textIndex.index = 100000000;
+      ti.before =true;
+      editorIF->setCursorIndex(ti);
+      if (editorIF->isSelecting())
+      {
+         editorIF->selectText (editorIF->getCursorIndex());
+      }
+    }
     break;
   case SWindowListener::Key_Prior:
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
     editorIF->pageUp();
     return;
   case SWindowListener::Key_Next:
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
     editorIF->pageDown();
     return;
   case SWindowListener::Key_Up:
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
     editorIF->caretUp();
     return;
   case SWindowListener::Key_Down:
-    editorIF->caretDown();
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
+        editorIF->caretDown();
     return;
   case SWindowListener::Key_Left:
-  // Ifound this toggle very annoying.
-#if 0
-    if (editorIF->getDirection () == '>')
-    {
-       editorIF->setDirection ('<');
-       return;
-    }
-#endif
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
     editorIF->caretLeft();
     return;
   case SWindowListener::Key_Right:
-  // Ifound this toggle very annoying.
-#if 0
-    if (editorIF->getDirection () == '<')
-    {
-       editorIF->setDirection ('>');
-       return;
-    }
-#endif
+    if ((shift || ctrl || meta) && !editorIF->isSelecting()) { editorIF->startSelect(); }
     editorIF->caretRight();
     return;
+  /*SGC*/
   case SWindowListener::Key_BackSpace:
-    /* switch - delete */
-    std = SUndo::SS_EraseSelect;
-    ba = editorIF->getSelectedIndex(true);
-    string = editorIF->eraseSelectedText();
-    if (string.size() == 0)
+    /* switch - delete a single character */
+    if (erasedString.size() == 0)
     {
-      if (editorIF->getDirection () == '<')
-      {
-         std = SUndo::SS_EraseOne;
-         string = editorIF->erase();
-      }
-      else
-      {
-         std = SUndo::SS_BackSpaceOne;
-         string = editorIF->backspace();
-      }
+      std = SUndo::SS_BackSpaceOne;
+      erasedString = editorIF->backspace();
       /* not even used */
-      ba = editorIF->getCaret();
-    }
-    if (string.size() != 0)
-    {
-      a = editorIF->getCaret();
-      add (SUndo(std, editorIF->getDirection (), string, b, a, ba));
+      ba = editorIF->getCursorIndex();
+      add (SUndo(std, editorIF->getDirection (), embd, erasedString, b, ba, ba));
     }
     return;
   case SWindowListener::Key_Delete:
     /* switch - backspace */
-    std = SUndo::SS_EraseSelect;
-    ba = editorIF->getSelectedIndex(true);
-    string = editorIF->eraseSelectedText();
-    if (string.size() == 0)
+    if (erasedString.size() == 0)
     {
-      if (editorIF->getDirection () == '<')
-      {
-         std = SUndo::SS_BackSpaceOne;
-         string = editorIF->backspace();
-      }
-      else
-      {
-         std = SUndo::SS_EraseOne;
-         string = editorIF->erase();
-      }
+      std = SUndo::SS_EraseOne;
+      erasedString = editorIF->erase();
+
       /* not even used */
-      ba = editorIF->getCaret();
-    }
-    if (string.size() != 0)
-    {
-      a = editorIF->getCaret();
-      add (SUndo(std, editorIF->getDirection (), string, b, a, ba));
+      ba = editorIF->getCursorIndex();
+      add (SUndo(std, editorIF->getDirection (), embd, erasedString, b, ba, ba));
     }
     return;
   case SWindowListener::Key_Enter:
-    if (editorIF->getDirection () == '<')
-    {
-      b = editorIF->getIndexBeforeLineBreak();
-      if (cleared) return;
-      editorIF->caretGoTo(b);
-    }
-    ba = editorIF->insertDirtyText(lineBreak);
+    editorIF->insertDirtyText(lineBreak);
+    ba = editorIF->getCursorIndex();
     if (cleared) return;
-    a = editorIF->getCaret();
-    add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), lineBreak, b, a, ba));
+    a = editorIF->getCursorIndex();
+    add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), embd, lineBreak, b, a, ba));
     break;
   case SWindowListener::Key_Return:
-    if (editorIF->getDirection () == '<')
-    {
-      b = editorIF->getIndexBeforeLineBreak();
-      if (cleared) return;
-      editorIF->caretGoTo(b);
-    }
-    ba = editorIF->insertDirtyText(lineBreak);
+    editorIF->insertDirtyText(lineBreak);
+    ba = editorIF->getCursorIndex();
     if (cleared) return;
-    a = editorIF->getCaret();
-    add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), lineBreak, b, a, ba));
+    a = editorIF->getCursorIndex();
+    add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), embd, lineBreak, b, a, ba));
     break;
   case SWindowListener::Key_F1:
     fKey (0, ctrl|shift|meta);
@@ -291,117 +349,158 @@ SEditor::keyPressed (SWindowListener::SKey key, const SString& s,
   case SWindowListener::Key_F12:
     fKey (11, ctrl|shift|meta);
     break;
-  case SWindowListener::Key_u:
-  case SWindowListener::Key_U:
+  case SWindowListener::Key_B:
+  case SWindowListener::Key_b:
     if (ctrl|meta)
     {
-      undo ();
+      if (ctrl && editorIF->isSelecting()) editorIF->endSelect();
+      editorIF->pageUp ();
+      break;
+    }
+  case SWindowListener::Key_c:
+   if (ctrl|meta)
+    {
+      if (editorIF->isSelecting()) editorIF->endSelect();
+      return; /* Automatic copy */
+    }
+  /* This is used with accelerator anyway - this is never called here */
+  case SWindowListener::Key_Y:
+  case SWindowListener::Key_y:
+   if (ctrl|meta)
+    {
+      editorIF->setDirection (SS_DR_L);
       break;
     }
-  case SWindowListener::Key_d:
   case SWindowListener::Key_D:
-    if (ctrl|meta)
+  case SWindowListener::Key_d:
+   if (ctrl|meta)
     {
-      if (editorIF->getDirection () == '<')
+      if (editorIF->getDirection () != SS_DR_RO)
       {
-        editorIF->setDirection ('>');
+        editorIF->setDirection (SS_DR_RO);
       }
-      else
+      else 
       {
-        editorIF->setDirection ('<');
+        editorIF->setDirection (SS_DR_LO);
       }
       break;
     }
-  case SWindowListener::Key_r:
-  case SWindowListener::Key_R:
+  /* This is used with accelerator anyway - this is never called here */
+  case SWindowListener::Key_E:
+  case SWindowListener::Key_e:
+   if (ctrl|meta)
+    {
+      if (editorIF->getDirection () != SS_DR_RE)
+      {
+        editorIF->setDirection (SS_DR_RE);
+      }
+      else 
+      {
+        editorIF->setDirection (SS_DR_LE);
+      }
+      break;
+    }
+  case SWindowListener::Key_F:
+  case SWindowListener::Key_f:
     if (ctrl|meta)
+    {
+      if (ctrl && editorIF->isSelecting()) editorIF->endSelect();
+      editorIF->pageDown ();
+      break;
+    }
+  case SWindowListener::Key_H:
+  case SWindowListener::Key_h:
+    if (ctrl|meta)
+    {
+      if (ctrl && editorIF->isSelecting()) editorIF->endSelect();
+      editorIF->caretLeft ();
+      break;
+    }
+  case SWindowListener::Key_J:
+  case SWindowListener::Key_j:
+  case SWindowListener::Key_N:
+  case SWindowListener::Key_n:
+    if (ctrl||meta)
+    {
+      if (ctrl && editorIF->isSelecting()) editorIF->endSelect();
+      editorIF->caretDown ();
+      break;
+    }
+  case SWindowListener::Key_K:
+  case SWindowListener::Key_k:
+    if (ctrl|meta)
+    {
+      if (ctrl && editorIF->isSelecting()) editorIF->endSelect();
+      editorIF->caretUp ();
+      break;
+    }
+  case SWindowListener::Key_L:
+  case SWindowListener::Key_l:
+    if (ctrl||meta)
+    {
+      if (ctrl && editorIF->isSelecting()) editorIF->endSelect();
+      editorIF->caretRight ();
+      break;
+    }
+  case SWindowListener::Key_M:
+  case SWindowListener::Key_m:
+    if (ctrl)
+    {
+      editorIF->endSelect();
+      editorIF->setCursorIndex(SCursorIndex(b.textIndex.line, 0));
+      STextIndex mba = editorIF->getIndexAfterLineBreak ();
+      ba = SCursorIndex (mba.line, mba.index);
+      erasedString =  editorIF->eraseText(mba);
+      if (cleared) return;
+      if (erasedString.size() != 0)
+      {
+        a = editorIF->getCursorIndex();
+        add (SUndo(SUndo::SS_Erase, editorIF->getDirection (), embd,  
+             erasedString, b, a, ba));
+      }
+      break;
+    }
+  case SWindowListener::Key_R:
+  case SWindowListener::Key_r:
+    if (ctrl)
     {
       redo ();
       break;
     }
-  case SWindowListener::Key_f:
-  case SWindowListener::Key_F:
-    if (ctrl|meta)
+  case SWindowListener::Key_U:
+  case SWindowListener::Key_u:
+    if (ctrl)
     {
-      editorIF->pageDown ();
+      undo ();
       break;
     }
   case SWindowListener::Key_v:
+   if (ctrl) 
+    {
+      if (editorIF->isSelecting()) editorIF->endSelect();
+      pasteText(b);
+      return;
+    }
   case SWindowListener::Key_V:
     if (ctrl|meta)
     {
+      if (editorIF->isSelecting()) editorIF->endSelect();
       pasteText (b);
       return;
     }
-  case SWindowListener::Key_x:
   case SWindowListener::Key_X:
-    if (ctrl|meta)
+  case SWindowListener::Key_x:
+    if (ctrl | meta)
     {
       /* switch - delete */
       std = SUndo::SS_EraseSelect;
-      ba = editorIF->getSelectedIndex(true);
-      string = editorIF->eraseSelectedText();
+      STextIndex mba = editorIF->getSelectedIndex(true);
+      erasedString = editorIF->eraseSelectedText();
       if (cleared) return;
-      if (string.size() != 0)
+      if (erasedString.size() != 0)
       {
-        a = editorIF->getCaret();
-        add (SUndo(std, editorIF->getDirection (), string, b, a, ba));
-      }
-      break;
-    }
-  case SWindowListener::Key_b:
-  case SWindowListener::Key_B:
-    if (ctrl|meta)
-    {
-      editorIF->pageUp ();
-      break;
-    }
-  case SWindowListener::Key_p:
-  case SWindowListener::Key_P:
-  case SWindowListener::Key_k:
-  case SWindowListener::Key_K:
-    if (ctrl|meta)
-    {
-      editorIF->caretUp ();
-      break;
-    }
-  case SWindowListener::Key_n:
-  case SWindowListener::Key_N:
-  case SWindowListener::Key_j:
-  case SWindowListener::Key_J:
-    if (ctrl||meta)
-    {
-      editorIF->caretDown ();
-      break;
-    }
-  case SWindowListener::Key_h:
-  case SWindowListener::Key_H:
-    if (ctrl|meta)
-    {
-      editorIF->caretLeft ();
-      break;
-    }
-  case SWindowListener::Key_l:
-  case SWindowListener::Key_L:
-    if (ctrl||meta)
-    {
-      editorIF->caretRight ();
-      break;
-    }
-  case SWindowListener::Key_m:
-  case SWindowListener::Key_M:
-    if (ctrl)
-    {
-      editorIF->endSelect();
-      editorIF->caretGoTo(STextIndex(b.line, 0));
-      ba = editorIF->getIndexAfterLineBreak ();
-      string =  editorIF->eraseText(ba);
-      if (cleared) return;
-      if (string.size() != 0)
-      {
-        a = editorIF->getCaret();
-        add (SUndo(SUndo::SS_Erase, editorIF->getDirection (), 
-             string, b, a, ba));
+        a = editorIF->getCursorIndex();
+        add (SUndo(std, editorIF->getDirection (), embd,  erasedString, b, a, ba));
       }
       break;
     }
@@ -414,13 +513,37 @@ SEditor::keyPressed (SWindowListener::SKey key, const SString& s,
       editorIF->focusOut();
       break;
     }
-    if (s.size())
+    SEncoder enc;
+    SV_UCS4 ustr = enc.decode(s);
+    while (ustr.size() && editorIF->addComposing(ustr[0]))
     {
-      ba =  editorIF->insertDirtyText(s);
-      a = editorIF->getCaret();
-      add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), s, b, a, ba));
+      std = SUndo::SS_ComposeOne;
+      a = editorIF->getCursorIndex();
+      SS_UCS4 u4 = ustr[0];
+      /* hack, string stores u4 */
+      add (SUndo(std, editorIF->getDirection (), embd,  
+        SString((char*)&u4, sizeof (u4)), a, a, a));
+      ustr.remove (0);
+    }
+    SString sstr = enc.encode (ustr);
+    if (sstr.size())
+    {
+      b = editorIF->getCursorIndex();
+      editorIF->insertDirtyText(sstr);
+      ba = editorIF->getCursorIndex();
+      a = editorIF->getCursorIndex();
+      add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), embd,
+          sstr, b, a, ba));
+    }
+    else /* to update glyph info */
+    {
+      a = editorIF->getCursorIndex();
+      editorIF->setCursorIndex(a);
     }
   }
+
+
+  
 }
 
 /**
@@ -432,7 +555,9 @@ SEditor::keyReleased (SWindowListener::SKey key, const SString& s,
 {
   if (editorIF == 0) return;
   if (key==SWindowListener::Key_Meta_R || key == SWindowListener::Key_Meta_L
-   || key == SWindowListener::Key_Alt_R || key==SWindowListener::Key_Alt_L)
+   || key == SWindowListener::Key_Alt_R || key==SWindowListener::Key_Alt_L 
+   || key == SWindowListener::Key_Control_L||key==SWindowListener::Key_Control_R
+   || key == SWindowListener::Key_Shift_R || key==SWindowListener::Key_Shift_L)
   {
     editorIF->endSelect();
     return;
@@ -443,33 +568,43 @@ SEditor::keyReleased (SWindowListener::SKey key, const SString& s,
  * called when a mouse button was pressed.
  */
 void
-SEditor::buttonPressed (int button, const STextIndex& index)
+SEditor::buttonPressed (int button, const SCursorIndex& index)
 {
   if (editorIF == 0) return;
   editorIF->deselectText();
   if (button == 1)
   {
-    pasteText(editorIF->getCaret());
+    pasteText(editorIF->getCursorIndex());
     return;
   }
-  editorIF->caretGoTo (index);
-  editorIF->startSelect();
+  /*mousewheel*/
+  if (button == 4) { 
+    editorIF->caretDown();
+  } else if (button == 3) {
+    editorIF->caretUp();
+  }
+  else {
+    editorIF->setCursorIndex (index);
+    editorIF->startSelect(); 
+  }
 }
 
 /**
  * paste a text ftom clip
  */
 void
-SEditor::pasteText (const STextIndex& index)
+SEditor::pasteText (const SCursorIndex& index)
 {
   if (editorIF == 0) return;
   editorIF->deselectText();
-  editorIF->caretGoTo(index);
-  STextIndex b = index;
-  STextIndex ba = editorIF->insertText();
-  SString s = editorIF->getText (b, ba);
-  STextIndex a = editorIF->getCaret();
-  add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), s, b, a, ba));
+  editorIF->setCursorIndex(index);
+  SCursorIndex b = index;
+  /* stop input methods */
+  editorIF->insertClipboardText();
+  SCursorIndex ba = editorIF->getCursorIndex();
+  SString s = editorIF->getText (b.getTextIndex(), ba.getTextIndex());
+  SCursorIndex a = editorIF->getCursorIndex();
+  add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), editorIF->getDocumentEmbedding(), s, b, a, ba));
 }
 
 /**
@@ -480,18 +615,19 @@ SEditor::insertText (const SString& text)
 {
   if (editorIF == 0) return;
   editorIF->deselectText();
-  STextIndex b = editorIF->getCaret();
-  STextIndex ba = editorIF->insertText(text);
-  SString s = editorIF->getText (b, ba);
-  STextIndex a = editorIF->getCaret();
-  add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), s, b, a, ba));
+  SCursorIndex b = editorIF->getCursorIndex();
+  editorIF->insertText(text);
+  SCursorIndex ba = editorIF->getCursorIndex();
+  SString s = text;
+  SCursorIndex a = editorIF->getCursorIndex();
+  add (SUndo(SUndo::SS_Insert, editorIF->getDirection (), editorIF->getDocumentEmbedding(), s, b, a, ba));
 }
 
 /**
  * called when a mouse button was released.
  */
 void
-SEditor::buttonReleased (int button, const STextIndex& index)
+SEditor::buttonReleased (int button, const SCursorIndex& index)
 {
   if (editorIF == 0) return;
   if (button == 1)
@@ -506,7 +642,7 @@ SEditor::buttonReleased (int button, const STextIndex& index)
  * called when a mouse button was dragged.
  */
 void
-SEditor::buttonDragged (int button, const STextIndex& index)
+SEditor::buttonDragged (int button, const SCursorIndex& index)
 {
   if (button == 1)
   {
@@ -521,7 +657,7 @@ SEditor::buttonDragged (int button, const STextIndex& index)
  * It can come only if button was already released.
  */
 void
-SEditor::multiClicked (int button, const STextIndex& index, int count)
+SEditor::multiClicked (int button, const SCursorIndex& index, int count)
 {
   if (editorIF == 0) return;
   if (button == 1)
@@ -621,31 +757,42 @@ SEditor::undo()
   undoIndex--;
   SUndo u = undoBuffer[undoIndex];
   //fprintf (stderr, "Undo:%d.\n", u.type);
-  STextIndex in;
   switch (u.type)
   {
   case SUndo::SS_Insert:
-    editorIF->caretGoTo(u.beforeAfter);
-    editorIF->eraseText (u.before);
-    editorIF->setDirection(u.caret);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.beforeAfter);
+    editorIF->eraseText (u.before.getTextIndex());
+    editorIF->setDirection(u.direction);
     break;
   case SUndo::SS_Erase:
   case SUndo::SS_EraseOne:
   case SUndo::SS_BackSpaceOne:
-    editorIF->caretGoTo(u.after);
-    in = editorIF->insertText(u.string);
-    editorIF->caretGoTo(u.before);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.beforeAfter);
+    editorIF->setCursorIndex(u.after);
+    editorIF->insertText(u.string, false);
+    editorIF->setCursorIndex(u.before);
+    break;
+  case SUndo::SS_ComposeOne:
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.after);
+    editorIF->setDirection(u.direction);
+    editorIF->removeComposing();
+    editorIF->setCursorIndex(u.after);
+    editorIF->setDirection(u.direction);
+    /* compose does not move caret */
     break;
   case SUndo::SS_EraseSelect:
-    editorIF->caretGoTo(u.after);
-    STextIndex in = editorIF->insertText(u.string);
-    editorIF->caretGoTo(in);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.after);
+    editorIF->insertText(u.string, false);
     editorIF->startSelect();
 
     /* this is the sleection spot */
     editorIF->selectText(u.after);
     editorIF->endSelect();
-    editorIF->caretGoTo(u.after);
+    editorIF->setCursorIndex(u.after);
     break;
   }
   return true;
@@ -676,26 +823,39 @@ SEditor::redo()
   SUndo u = undoBuffer[undoIndex];
   undoIndex++;
   //fprintf (stderr, "redo:%d.\n", u.type);
-  STextIndex in;
-  switch (u.type)
+  switch (u.type) 
   {
   case SUndo::SS_Insert:
-    editorIF->caretGoTo(u.before);
-    in = editorIF->insertText(u.string);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.before);
+    editorIF->insertText(u.string, false);
     break;
   case SUndo::SS_Erase:
   case SUndo::SS_EraseSelect:
-    editorIF->caretGoTo(u.after);
-    editorIF->eraseText(u.beforeAfter);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.after);
+    editorIF->eraseText(u.beforeAfter.getTextIndex());
     break;
   case SUndo::SS_BackSpaceOne:
-    editorIF->caretGoTo(u.before);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.before);
     editorIF->backspace();
     break;
   case SUndo::SS_EraseOne:
-    editorIF->caretGoTo(u.before);
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.before);
     editorIF->erase();
     break;
+  case SUndo::SS_ComposeOne:
+    editorIF->setDocumentEmbedding(u.embedding);
+    editorIF->setCursorIndex(u.after);
+    editorIF->setDirection(u.direction);
+    {
+      const SS_UCS4* u4 = (const SS_UCS4*) u.string.array();
+      editorIF->addComposing(u4[0]);
+      editorIF->setCursorIndex(u.after);
+    }
+    break;
   }
   return true;
 }
@@ -744,13 +904,13 @@ SEditor::clear ()
 }
 
 void
-SEditor::setLineBreak (const SString& _lineBreak)
+SEditor::setParagraphSeparator (const SString& _paragBreak)
 {
-  lineBreak = _lineBreak;
+  paragBreak = _paragBreak;
 }
 
 SString
-SEditor::getLineBreak () const
+SEditor::getParagraphSeparator () const
 {
-  return SString (lineBreak);
+  return SString (paragBreak);
 }
