@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -26,6 +27,7 @@
 #include "stoolkit/SHashtable.h"
 #include "stoolkit/STextData.h"
 #include "stoolkit/SExcept.h"
+#include "stoolkit/SCluster.h"
 
 static SFontFB fallbackFont;
 typedef SHashtable<SFontImplVector> SFontHashtable;
@@ -64,11 +66,22 @@ SFont::SFont (void) : name("default"), xlfd (SD_XLFD_ANY)
   static bool _defaultSet=false;
   if (_defaultSet==false)
   {
-    /* FIXME: code2000.ttf and arial.ttf are not free fonts */
     SStringVector m(
+        /* yudit.hex has ohungarian.hex and part of unifont.hex */
+        // Indic range is not necessary free. But they are the best
+        "yudit.hex,arabforms.hex,syriacforms.hex,unifont.hex,"
+        "markus9x18.bdf,markus18x18ja.bdf,"
         "-*-*-medium-r-normal--16-*-*-*-c-*-iso8859-1,"
         "-*-*-*-*-*--16-*-*-*-c-*-iso8859-1,"
-        "cyberbit.ttf,code2000.ttf,arial.ttf,"
+        "THOOLIUC.TTF:mlym,"
+        "ani.ttf:beng,"
+        "pothana2000.ttf:telu,"
+        "TCRCYoutsoUnicode.ttf:tibt,"
+        "raghu.ttf:deva,"
+        "mangal.ttf:deva,tunga.ttf:knda,code2000.ttf:taml,"
+        "raavi.ttf:guru,shruti.ttf:gujr,"
+        "arialuni.ttf,cyberbit.ttf,"
+        "code2000.ttf,code2001.ttf:unicode:RL,arial.ttf,"
         "yudit.ttf"
     );
     SFontImplVector list;
@@ -122,6 +135,7 @@ SFont::SFont (const SFont& font)
   fontGap = font.fontGap;
   fontVector = font.fontVector;
   fontScale = font.fontScale;
+  fallbackScale = font.fallbackScale;
 }
 
 /**
@@ -138,6 +152,7 @@ SFont::operator = (const SFont& font)
   fontGap = font.fontGap;
   xlfd = font.xlfd;
   fontScale = font.fontScale;
+  fallbackScale = font.fallbackScale;
   fontVector = font.fontVector;
   return *this;
 }
@@ -154,12 +169,14 @@ void
 SFont::setSize(double points)
 {
   fontScale = points;
+  fallbackScale = points;
   SS_Matrix2D m;
   double sc = fallbackFont.scale();
   m.scale (fontScale * sc, fontScale * sc);
-  fontGap =  fallbackFont.gap(m);
-  fontAscent =  fallbackFont.ascent(m);
-  fontDescent =  fallbackFont.descent(m);
+  fontGap = 0.0;
+  /* at least */
+  fontAscent = -1.0;
+  fontDescent = 0.0;
 
   for (unsigned int i=0; i<fontVector.size(); i++)
   {
@@ -174,6 +191,14 @@ SFont::setSize(double points)
     if (a > fontAscent) fontAscent = a;
     if (d > fontDescent) fontDescent = d;
   }
+  if (fontAscent <= 1.0 || fontAscent < points/2)
+  {
+    fontGap =  fallbackFont.gap(m);
+    fontAscent =  fallbackFont.ascent(m);
+    fontDescent =  fallbackFont.descent(m);
+    if (fontAscent < 1.0) fontAscent = 1.0;
+  }
+  fallbackScale = fontAscent+fontDescent;
 }
 
 /**
@@ -213,21 +238,24 @@ SFont::clear()
 static SFontFB::SIGN
 mapFB (const SGlyph& glyph)
 {
-  if (glyph.size() > 1)
+  if (glyph.decompSize() > 1)
   {
-    if (glyph[0] == SGlyph::CR || glyph[1] == SGlyph::LF) return SFontFB::CRLF;
-    if (glyph[1] == SGlyph::CR || glyph[0] == SGlyph::LF) return SFontFB::LFCR;
+    if (glyph[0] == SD_CD_CR || glyph[1] == SD_CD_LF) return SFontFB::CRLF;
+    if (glyph[1] == SD_CD_CR || glyph[0] == SD_CD_LF) return SFontFB::LFCR;
   }
   else
   {
-    if (glyph.getChar() == SGlyph::CR) return SFontFB::CR;
-    if (glyph.getChar() == SGlyph::LF) return SFontFB::LF;
-    if (glyph.getChar() == SGlyph::LS) return SFontFB::LS;
-    if (glyph.getChar() == SGlyph::PS) return SFontFB::PS;
-    if (glyph.getChar() == SGlyph::TAB) return SFontFB::TAB;
-    if (glyph.getChar() < 32) return SFontFB::CTRL;
+    if (glyph.getChar() == SD_CD_CR) return SFontFB::CR;
+    if (glyph.getChar() == SD_CD_LF) return SFontFB::LF;
+    if (glyph.getChar() == SD_CD_LS) return SFontFB::LS;
+    if (glyph.getChar() == SD_CD_PS) return SFontFB::PS;
+    if (glyph.getChar() == SD_CD_TAB) return SFontFB::TAB;
+    if (glyph.getChar() == SD_CD_LRM) return SFontFB::LRM;
+    if (glyph.getChar() == SD_CD_RLM) return SFontFB::RLM;
+    if (glyph.getChar() == SD_CD_ZWJ) return SFontFB::FB_ZWJ;
+    if (glyph.getChar() == SD_CD_ZWNJ) return SFontFB::FB_ZWNJ;
   }
-  return SFontFB::LF;
+  return SFontFB::CTRL;
 }
 
 /**
@@ -246,53 +274,101 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
   {
     SS_Matrix2D sd;
     double sc = fallbackFont.scale();
-    sd.scale (fontScale * sc, fontScale * sc);
+    sd.scale (fallbackScale * sc, fallbackScale * sc);
+    sd.translate (0, -fontDescent);
     SS_Matrix2D sm =  m * sd;
     SFontFB::SIGN sig = mapFB (glyph);
-    fallbackFont.signDraw(canvas, pen, sm, sig, glyph.getChar());
+    /* The markers are not mirrored. */
+    if (!glyph.isLR() && sig != SFontFB::LRM && sig != SFontFB::RLM)
+    {
+      currw =  fallbackFont.signWidth (sm, sig);
+      /* mirroring */
+      sm.x0 = -sm.x0;
+      sm.t0 = sm.t0 + currw;
+    }
+    fallbackFont.signDraw(canvas, pen, sm, sig, glyph.getFirstChar());
     return;
   }
+
+
   /* try to draw it with all fonts. */
   SS_UCS4 comp =  glyph.getShapedChar();
-  if (!glyph.lr && glyph.isMirrorable())
-     comp = glyph.getMirroredChar();
+  unsigned int scriptcode = getLigatureScriptCode (comp);
+  SV_UCS4 ligclust;
+  if (scriptcode == SD_COMBINING_LIGATURE)
+  {
+    unsigned int lgsize  = getLigatureCluster (comp, 0);
+    if (lgsize > 1)
+    {
+      SS_UCS4* lc = new SS_UCS4[lgsize];
+      CHECK_NEW (lc);
+      getLigatureCluster (comp, lc);
+      comp = lc[0];
+      for (unsigned int i=0; i<lgsize; i++)
+      {
+        ligclust.append (lc[i]);
+      }
+      delete lc;
+    }
+  }
+  else if (!glyph.isLR() && glyph.isMirrorable())
+  {
+    comp = glyph.getMirroredChar();
+  }
+
+  /* gsize is the full size */
+  unsigned int gsize = glyph.decompSize();
+  /* This array is the full array */
+  const SS_UCS4* decomp = glyph.getDecompArray(); 
+
+  if (comp == 0 && glyph.decompSize() == 1) comp = decomp[0];
 
   /* first try the precomposed */
+  bool baseOK = false;
+  double baseWidth = 0.0;
+  setBase (comp);
   if (comp != 0 && comp!= 0x200c && comp != 0x200d)
   {
     unsigned int i;
-    /* use mirrored glyphs for Old Hungarian */
-    if (comp >= 0xEE00 && comp < 0xEE9F)
+
+    /* use mirrored glyphs for Old Hungarian, Old Italic */
+    if (  (comp >= 0xEE00 && comp < 0xEE9F)
+        || (comp >= 0x10300 && comp <= 0x1032F))
     {
       /* Try to use lr and rl attributes */
       for (i=0; i<fontVector.size(); i++)
       {
         SFontImpl im = fontVector[i];
-        if (im.isLR() && !glyph.lr) continue;
-        if (im.isRL() && glyph.lr) continue;
+        if (im.isLR() && !glyph.isLR()) continue;
+        if (im.isRL() && glyph.isLR()) continue;
         /* neutral = lr */
-        if (!im.isLR() && !im.isRL() && !glyph.lr) continue;
+        if (!im.isLR() && !im.isRL() && !glyph.isLR()) continue;
+        /* this is for better positioning of diacritical marks. */
         if (im.draw (canvas, pen, m, comp)) 
         {
-          return;
+          im.width (comp, &baseWidth);
+          baseOK=true;
+          break;
         }
       }
       SS_Matrix2D mm =  m;
       /* Try to mirror it - we drew neutrals and same directions */
-      for (i=0; i<fontVector.size(); i++)
+      if (!baseOK) for (i=0; i<fontVector.size(); i++)
       {
         SFontImpl im = fontVector[i];
         bool used = im.width (comp, &currw);
         if (used)
         {
           /* Mirrorring can be done only if we render on our own. */
+          mm.x0 = -m.x0;  /* try to indicate mirroring */
           if (im.isTTF())
           {
-            mm.x0 = -m.x0; 
             mm.t0 = m.t0 + currw;
           }
           im.draw (canvas, pen, mm, comp); 
-          return;
+          baseWidth = currw;
+          baseOK = true;
+          break;
         }
         currw = 0.0;
       }
@@ -305,26 +381,87 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
         SFontImpl im = fontVector[i];
         if (im.draw (canvas, pen, m, comp)) 
         {
-          return;
+          im.width (comp, &baseWidth);
+          baseOK = true;
+          break;
         }
       }
     }
   }
 
-  /* You reach this point if comp did not work*/
-  unsigned int gsize = glyph.size();
-  const SS_UCS4* decomp = glyph.getDecompArray(); 
-  if (gsize > 0)
-   /* there is hope to draw composed chars */
+  /*
+   * If it is shaped and current shape is isolated fallback 
+   */
+  if (!baseOK && glyph.getShapeArray()!=0 && glyph.currentShape == 0)// ISOLATED
+  {
+    SS_UCS4 orig =  glyph.getChar();
+    if (orig!=comp && orig != 0)
+    {
+      setBase (orig);
+      for (unsigned int i=0; i<fontVector.size(); i++)
+      {
+        SFontImpl im = fontVector[i];
+        if (im.draw (canvas, pen, m, orig)) 
+        {
+          im.width (orig, &baseWidth);
+          baseOK = true;
+          break;
+        }
+      }
+    }
+  }
+
+  /* Try precompositions  */
+  const SS_UCS4* fbs = 0;
+  if (!baseOK && gsize > 0)
   {
     double* positions = new double[gsize];
     unsigned int* indeces = new unsigned int[gsize];
     CHECK_NEW (positions);
     CHECK_NEW (indeces);
     bool found = false;
-    bool overstrike =  (glyph.getShapeArray() ==0 
+
+    bool overstrike =  (glyph.getShapeArray()==0 
         && !glyph.isYuditLigature() && !glyph.isCluster());
 
+
+    /*
+     * Hack for special Yudit ligatures 
+     * Normally clusters are not OVERSTRIKE.
+     * MARK Composing Cluster: see bin/cluster/cluster.template 
+    */
+    if (gsize > 1 && 
+        (decomp[1] == 0x309A || decomp[1] == 0x300 || decomp[1] == 0x301))
+    {
+      overstrike = true;
+    }
+    else
+    {
+    }
+    int scode = getLigatureScriptCode (comp); 
+    if (scode == SD_HANGUL_JAMO 
+        || scode == SD_LAO || scode == SD_THAI  || scode == SD_TIBETAN)
+    {
+      overstrike = true;
+    }
+
+    /* shape fallback does not include composing marks */
+    fbs = glyph.getShapeFallback();
+    if (fbs)
+    {
+      overstrike = false; /* well, this is overstrike sometimes */
+      decomp = fbs; /* even if we can not draw it, this will be displayed */
+    }
+    /* we can not do overstrike positioning on a shape fallback */
+    if (overstrike)
+    {
+      setBase (decomp[0]);
+    }
+    else
+    {
+      setBase (0);
+    }
+
     /* build positions */
     unsigned int index = 0;
     unsigned int i=0;
@@ -333,13 +470,15 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
     while (i<fsize) 
     {
       SFontImpl im = fontVector[i];
-      bool used = im.width (decomp[index], &currw);
+      currw = 0.0;
+      bool used = (index > 0 && decomp[index]==0x200d) 
+          ? true: im.width (decomp[index], &currw);
       /* ZWJ and ZWNJ - use fallback if not present.*/
       if (!used && i+1 == fsize && !overstrike)
       {
          SS_Matrix2D sm;
          double sc = fallbackFont.scale();
-         sm.scale (fontScale * sc, fontScale * sc);
+         sm.scale (fallbackScale * sc, fallbackScale * sc);
          used = true;
          i = fsize;
          currw = fallbackFont.width (sm, decomp[index]);
@@ -361,7 +500,16 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
          {
            i++; continue;
          }
-         positions[index] = im.isTTF() ? positions[0] : 0;
+         /* by default don't move caret */
+         positions[index] = 0;
+
+         /* 
+          * Should be in sync with: SFontTTF::getBaseOffsets
+          */
+         if (!im.isLeftAligned(decomp[index]))
+         {
+            positions[index] = positions[0]-currw;
+         }
       }
       else
       {
@@ -379,7 +527,8 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
       /* found if all found */
       if (index == gsize)
       {
-        found = 1;
+        baseWidth = fullsize;
+        found = true;
         break;
       }
     }
@@ -389,7 +538,7 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
       {
         SS_Matrix2D mo = m;
         double translatex = 0.0;
-        if (glyph.lr)
+        if (glyph.isLR())
         {
           if (overstrike)
           {
@@ -415,7 +564,8 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
         {
           SS_Matrix2D sd;
           double sc = fallbackFont.scale();
-          sd.scale (fontScale * sc, fontScale * sc);
+          sd.scale (fallbackScale * sc, fallbackScale * sc);
+          sd.translate (0, -fontDescent);
           SS_Matrix2D sm =  m * sd;
           sm.translate (translatex, (double)0.0);
           fallbackFont.draw(canvas, pen, sm, decomp[i]);
@@ -430,53 +580,143 @@ SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m,
     }
     delete positions;
     delete indeces;
-    if (found) return;
-  }
-  else if (glyph.getChar() != comp)
-   /* gsize is 0. Is it a single shaped glyph? */
-  {
-    SS_UCS4 orchar = glyph.getChar();
-    /* Try all fonts on it */
-    for (unsigned int i=0; i<fontVector.size(); i++)
+    if (found) 
     {
-      SFontImpl im = fontVector[i];
-      /* cool. draw it */
-      if (im.draw (canvas, pen, m, orchar)) 
+      baseOK = true;
+      /* combining marks are not part of fallback */
+      if (fbs==0) ligclust.clear();
+    }
+  }
+
+  /*
+   * If it is shaped and current shape is any fallback 
+   * Isolated fallback has been processed already.
+   */
+  if (!baseOK && glyph.getShapeArray()!=0 && glyph.currentShape != 0) //ISOLATED
+  {
+    SS_UCS4 orig =  glyph.getChar();
+    if (orig!=comp && orig != 0)
+    {
+      setBase (orig);
+      for (unsigned int i=0; i<fontVector.size(); i++)
       {
-         return;
+        SFontImpl im = fontVector[i];
+        if (im.draw (canvas, pen, m, orig)) 
+        {
+          im.width (orig, &baseWidth);
+          baseOK = true;
+          break;
+        }
       }
     }
   }
 
+  /* Add extra composing characters as overstrike - if possible  */
+  if (baseOK) /* we already have set the base with setbase */
+  {
+    gsize = glyph.compSize();
+    decomp = glyph.getCompArray();
+    /* add composing marks for composing clusters */
+    unsigned int postcomp = 0;
+    if (ligclust.size() > 1)
+    {
+      ligclust.remove (0);
+      postcomp = ligclust.size();
+      for (unsigned int i=0; i<gsize; i++)
+      {
+        ligclust.append (decomp[i]);
+      } 
+      gsize = ligclust.size();
+      decomp = ligclust.array();
+    }
+    for (unsigned int i=0; i<gsize; i++)
+    {
+      /* Try all fonts on it */
+      SS_Matrix2D mc = m;
+      for (unsigned int j=0; j<fontVector.size(); j++)
+      {
+        SFontImpl im = fontVector[j];
+        bool used = im.width (decomp[i], &currw);
+        if (!used) continue;
+        /* FIXME: find a better way */
+        /* get these rl scripts right somehow, fonts suppose you go
+          visual order...  */
+
+        /* Measure LR: composing from the end. */
+        if (!im.isLeftAligned (decomp[i]))
+        {
+          mc.t0 = m.t0 + baseWidth - currw;
+        }
+        /* try to play with composing marks in the middle of ligature */
+        /* This trick works only if ligature has two base characters */
+        else if (i<postcomp && (im.isTTF() || fbs != 0))
+        {
+          /* RL: move it more to the end */
+          if ((decomp[i] >= 0x500 && decomp[i] < 0x900))
+          {
+            mc.t0 = m.t0 + baseWidth / 2;
+          }
+        }
+        /* Draw it finally... */
+        if (im.draw (canvas, pen, mc, decomp[i])) 
+        {
+           break;
+        }
+      }
+    }
+  
+    /* return anyway. 
+       Extra composing failures will not be indicated for now. There is one 
+       for sure when you apply the composing to a cluster  */
+    return;
+  }
+
   /* Draw some last resort font. */
   SS_Matrix2D sd;
   double sc = fallbackFont.scale();
-  sd.scale (fontScale * sc, fontScale * sc);
+  sd.scale (fallbackScale * sc, fallbackScale * sc);
+  sd.translate (0, -fontDescent);
   SS_Matrix2D sm =  m * sd;
-
+  SS_Matrix2D mo = sm;
+  SV_UCS4 allchar;
   if (comp != 0)
   {
-    fallbackFont.draw(canvas, pen, sm, comp);
+    allchar.append (comp);
+    gsize = glyph.compSize();
+    decomp = glyph.getCompArray();
   }
   else
   {
-    SS_Matrix2D mo = sm;
-    int myindex = 0;
-    int inc = 1;
-    int limit = gsize;
-    if (!glyph.lr)
+    /* draw everything. CompArray is right after DecompArray */
+    gsize += glyph.compSize();
+  }
+  for (unsigned int k=0; k<gsize; k++)
+  {
+    allchar.append (decomp[k]);
+  }
+  gsize = allchar.size();
+  decomp = allchar.array();
+
+  int myindex = 0;
+  int inc = 1;
+  int limit = gsize;
+  if (!glyph.isLR())
+  {
+    myindex = gsize-1;
+    inc = -1;
+    limit = -1;
+  }
+  while (myindex!=limit)
+  {
+    SS_UCS4 chr = decomp[myindex];
+    if (getLigatureScriptCode (chr) == SD_AS_LITERAL)
     {
-      myindex = gsize-1;
-      inc = -1;
-      limit = -1;
-    }
-    while (myindex!=limit)
-    {
-      fallbackFont.draw(canvas, pen, mo, decomp[myindex]);
-      currw = fallbackFont.width (mo, decomp[myindex]);
-      mo.translate (currw, (double)0.0);
-      myindex = myindex + inc;
+      chr = chr & 0xff;
     }
+    fallbackFont.draw(canvas, pen, mo, chr);
+    currw = fallbackFont.width (mo, chr);
+    mo.translate (currw, (double)0.0);
+    myindex = myindex + inc;
   }
   return;
 }
@@ -498,7 +738,7 @@ SFont::width (const SGlyph& glyph)
   {
     SS_Matrix2D m;
     double sc = fallbackFont.scale();
-    m.scale (fontScale * sc, fontScale * sc);
+    m.scale (fallbackScale * sc, fallbackScale * sc);
     SFontFB::SIGN sig = mapFB (glyph);
     currw =  fallbackFont.signWidth (m, sig);
     return currw;
@@ -506,8 +746,23 @@ SFont::width (const SGlyph& glyph)
 
   /* try to draw it with all fonts. */
   SS_UCS4 comp =  glyph.getShapedChar();
-  if (!glyph.lr && glyph.isMirrorable())
-     comp = glyph.getMirroredChar();
+  unsigned int scriptcode = getLigatureScriptCode (comp);
+  if (scriptcode == SD_COMBINING_LIGATURE)
+  {
+    unsigned int lgsize = getLigatureCluster (comp, 0);
+    if (lgsize > 1)
+    {
+      SS_UCS4* lc = new SS_UCS4[lgsize];
+      CHECK_NEW (lc);
+      getLigatureCluster (comp, lc);
+      comp = lc[0];
+      delete lc;
+    }
+  }
+  else if (!glyph.isLR() && glyph.isMirrorable())
+  {
+    comp = glyph.getMirroredChar();
+  }
 
   /* first try the precomposed */
   if (comp != 0 && comp!= 0x200c && comp != 0x200d)
@@ -522,30 +777,71 @@ SFont::width (const SGlyph& glyph)
       }
     }
   }
+  /*
+   * If it is shaped and current shape is isolated fallback 
+   */
+  if (glyph.getShapeArray()!=0 && glyph.currentShape == 0) // ISOLATED
+  {
+    SS_UCS4 orig =  glyph.getChar();
+    if (orig!=comp && orig != 0)
+    {
+      for (unsigned int i=0; i<fontVector.size(); i++)
+      {
+        SFontImpl im = fontVector[i];
+        bool used = im.width (orig, &currw);
+        if (used && currw >0.0)
+        {
+          return currw;
+        }
+      }
+    }
+  }
+
   /* You reach this point if comp did not work */
-  unsigned int gsize = glyph.size();
+  unsigned int gsize = glyph.decompSize();
   const SS_UCS4* decomp = glyph.getDecompArray(); 
 
-  /*
-   * Shaped glyphs have no overstrikes here, but continuous output. 
-   */
   if (gsize > 0)
   {
     bool overstrike =  (glyph.getShapeArray() ==0 
         && !glyph.isYuditLigature() && !glyph.isCluster());
+
+    /* hack for special Yudit ligatures */
+    /* MARK Composing Cluster: see bin/cluster/cluster.template */
+    if (gsize > 1 && 
+        (decomp[1] == 0x309A || decomp[1] == 0x300 || decomp[1] == 0x301))
+    {
+      overstrike = true;
+    }
+    int scode = getLigatureScriptCode (comp); 
+    if (scode == SD_HANGUL_JAMO 
+        || scode == SD_LAO || scode == SD_THAI  || scode == SD_TIBETAN)
+    {
+      overstrike = true;
+    }
+
+    const SS_UCS4* fbs = glyph.getShapeFallback();
+    if (fbs)
+    {
+      overstrike = false;
+      decomp = fbs; /* even if we can not draw it, this will be displayed */
+    }
+
     unsigned int index = 0;
     unsigned int i=0;
     unsigned int fsize = fontVector.size();
     while (i<fsize)
     {
       SFontImpl im = fontVector[i];
-      bool used = im.width (decomp[index], &currw);
+      currw = 0.0; 
+      bool used = (index > 0 && decomp[index]==0x200d)
+        ? true : im.width (decomp[index], &currw);
       /* can draw fallback in the middle */
       if (!used && i+1 == fsize && !overstrike)
       {
          SS_Matrix2D sm;
          double sc = fallbackFont.scale();
-         sm.scale (fontScale * sc, fontScale * sc);
+         sm.scale (fallbackScale * sc, fallbackScale * sc);
          used = true;
          i = fsize;
          currw = fallbackFont.width (sm, decomp[index]);
@@ -576,17 +872,23 @@ SFont::width (const SGlyph& glyph)
       }
     }
   }
-  else if (glyph.getChar() != comp)
-   /* gsize is 0. Is it a single shaped glyph? */
+  /*
+   * If it is shaped and current shape is any fallback 
+   * Isolated fallback has been processed already.
+   */
+  if (glyph.getShapeArray()!=0 && glyph.currentShape != 0) // ISOLATED
   {
-    SS_UCS4 ocomp = glyph.getChar();
-    for (unsigned int i=0; i<fontVector.size(); i++)
+    SS_UCS4 orig =  glyph.getChar();
+    if (orig!=comp && orig != 0)
     {
-      SFontImpl im = fontVector[i];
-      bool used = im.width (ocomp, &currw);
-      if (used && currw >0.0)
+      for (unsigned int i=0; i<fontVector.size(); i++)
       {
-        return currw;
+        SFontImpl im = fontVector[i];
+        bool used = im.width (orig, &currw);
+        if (used && currw >0.0)
+        {
+          return currw;
+        }
       }
     }
   }
@@ -594,18 +896,32 @@ SFont::width (const SGlyph& glyph)
   /* last resort font */
   SS_Matrix2D sm;
   double sc = fallbackFont.scale();
-  sm.scale (fontScale * sc, fontScale * sc);
+  sm.scale (fallbackScale * sc, fallbackScale * sc);
+  maxw  = 0;
   if (comp != 0)
   {
-    maxw = fallbackFont.width (sm, comp);
+    SS_UCS4 chr = comp;
+    if (getLigatureScriptCode (chr) == SD_AS_LITERAL)
+    {
+      chr = chr & 0xff;
+    }
+    maxw = fallbackFont.width (sm, chr);
+    gsize = glyph.compSize();
+    decomp = glyph.getCompArray();
   }
   else
   {
-    maxw  = 0;
-    for (unsigned int i=0; i< glyph.size(); i++)
+    /* draw everything */
+    gsize += glyph.compSize();
+  }
+  for (unsigned int i=0; i< gsize; i++)
+  {
+    SS_UCS4 chr = decomp[i];
+    if (getLigatureScriptCode (chr) == SD_AS_LITERAL)
     {
-      maxw = maxw + fallbackFont.width (sm, decomp[i]);
+      chr = chr & 0xff;
     }
+    maxw = maxw + fallbackFont.width (sm, chr);
   }
   return maxw;
 }
