@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -25,10 +26,27 @@
  * The following is from freeware 'dixon.zip' font at 
  * http://www.brockmeiers.de/fonts/
  */
-static double fb_scale=0.001;
 
-static short fb_ascent=728;
-static short fb_descent=210;
+#define SD_BOXR 10
+#define SD_BWIDTH (fb_width*SD_BOXR/400)
+
+//static short fb_ascent=728;
+//static short fb_descent=210;
+static short fb_doffset=210;
+static short fb_aoffset=728;
+
+/* we will draw with fb_aoffset ascender but pretend it is 0 */
+static short fb_ascent=fb_aoffset+fb_doffset;
+static short fb_descent=0;
+
+//static double fb_scale=0.001 * (728.0/ (728.0+210.0));
+static double fb_scale= 1.0/( + double(fb_ascent) * (1.0 + double (SD_BOXR) / 50));
+
+//static double fb_scale=0.001;
+static double fb_height = 728.0+210.0;
+
+//static double fb_scale=0.001;
+
 //static short fb_gap=150;
 static short fb_gap=0;
 static short fb_width=721;
@@ -40,11 +58,9 @@ static short fb_width=721;
 #define SD_CLOSE_PATH 32004
 #define SD_END_GLYPH 32005
 
-#define SD_BOXR 10
-#define SD_BWIDTH (fb_width*SD_BOXR/400)
 /**
  * The first number after the SD_BEGIN_GLYPH is the width 
- * dont worry we just use fb_width to scale.
+ * don't worry we just use fb_width to scale.
  */
 short fb_box[] =
 {
@@ -132,7 +148,7 @@ SD_BEGIN_GLYPH, 721, /*  width */
 SD_END_GLYPH
 };
 
-/* linefeed */
+/* Paragraph Separator */
 short fb_ps[] =
 {
 SD_BEGIN_GLYPH, 721, /*  width */ 
@@ -149,6 +165,75 @@ SD_BEGIN_GLYPH, 721, /*  width */
 SD_END_GLYPH
 };
 
+/* LRM */
+short fb_lrm[] =
+{
+SD_BEGIN_GLYPH, 721, /*  width */ 
+ SD_MOVE_TO,  720-500, 300,
+ SD_LINE_TO,  720-300, 300,
+ SD_LINE_TO,  720-300, 200,
+ SD_LINE_TO,  720-50, 325,
+ SD_LINE_TO,  720-300, 450,
+ SD_LINE_TO,  720-300, 350,
+ SD_LINE_TO,  720-500, 350,
+ SD_CLOSE_PATH,
+SD_END_GLYPH
+};
+
+/* RLM */
+short fb_rlm[] =
+{
+SD_BEGIN_GLYPH, 721, /*  width */ 
+ SD_MOVE_TO,  500, 300,
+ SD_LINE_TO,  300, 300,
+ SD_LINE_TO,  300, 200,
+ SD_LINE_TO,  50, 325,
+ SD_LINE_TO,  300, 450,
+ SD_LINE_TO,  300, 350,
+ SD_LINE_TO,  500, 350,
+ SD_CLOSE_PATH,
+SD_END_GLYPH
+};
+
+/* ZWNJ */
+short fb_zwnj[] =
+{
+SD_BEGIN_GLYPH, 721, /*  width */ 
+ SD_MOVE_TO,  550, 325,
+ SD_LINE_TO,  400, 200,
+ SD_LINE_TO,  400, 300,
+
+ SD_LINE_TO,  250, 300,
+ SD_LINE_TO,  250, 200,
+ SD_LINE_TO,  100, 325,
+ SD_LINE_TO,  250, 450,
+ SD_LINE_TO,  250, 350,
+
+ SD_LINE_TO,  400, 350,
+ SD_LINE_TO,  400, 450,
+ SD_CLOSE_PATH,
+SD_END_GLYPH
+};
+
+/* ZWJ */
+short fb_zwj[] =
+{
+SD_BEGIN_GLYPH, 721, /*  width */ 
+ SD_MOVE_TO,  550, 200,
+ SD_LINE_TO,  400, 300,
+
+ SD_LINE_TO,  250, 300,
+ SD_LINE_TO,  100, 200,
+
+ SD_LINE_TO,  100, 450,
+ SD_LINE_TO,  250, 350,
+
+ SD_LINE_TO,  400, 350,
+ SD_LINE_TO,  550, 450,
+ SD_CLOSE_PATH,
+SD_END_GLYPH
+};
+
 /**
  * The first number after the SD_BEGIN_GLYPH is the width 
  */
@@ -819,6 +904,18 @@ SFontFB::signDraw (SCanvas* canvas, const SPen& pen,
   case PS:
     drawOne (canvas, f, fb_ps);
     break;
+  case LRM:
+    drawOne (canvas, f, fb_lrm);
+    break;
+  case RLM:
+    drawOne (canvas, f, fb_rlm);
+    break;
+  case FB_ZWNJ:
+    drawOne (canvas, f, fb_zwnj);
+    break;
+  case FB_ZWJ:
+    drawOne (canvas, f, fb_zwj);
+    break;
   case TAB:
     break;
   case CTRL:
@@ -860,8 +957,9 @@ SFontFB::draw (SCanvas* canvas, const SPen& pen,
     drawOne (canvas, f, fb_box);
 
     r.scale (0.5, 0.9);
-    r.translate (r.x0 * double (fb_width) * double (SD_BOXR) / 100.0, 0);
+    r.translate (r.x0 * double (fb_width) * double (SD_BOXR) / 100.0, 2.0 * (double) SD_BOXR);
 
+    //r.translate (0, (r.y1 * (double)fb_doffset + (double) SD_BOXR));
     SS_Matrix2D m =  f * r;
     drawOne (canvas, m, fb_all[g>>4]);
     m.translate (m.x0 * double (fb_width), 0);
@@ -876,8 +974,9 @@ SFontFB::draw (SCanvas* canvas, const SPen& pen,
     r.scale (0.5, 0.4);
     r.translate (r.x0 * double (fb_width) * double (SD_BOXR) / 100.0, 0);
 
+    r.translate (0, (r.y1 * (double)fb_doffset + 2.0 * (double) SD_BOXR));
     SS_Matrix2D s = r;
-    r.translate (0, (r.y1 * double (fb_ascent + fb_descent)));
+    r.translate (0, (r.y1 * fb_height));
 
     SS_Matrix2D m =  f * r;
     drawOne (canvas, m, fb_all[(g>>12)&0x0f]);
@@ -900,8 +999,9 @@ SFontFB::draw (SCanvas* canvas, const SPen& pen,
     r.scale (0.5, 0.4);
     r.translate (r.x0 * double (fb_width) * double (SD_BOXR) / 100.0, 0);
 
+    r.translate (0, (r.y1 * (double)fb_doffset + 2.0 * (double) SD_BOXR));
     SS_Matrix2D s = r;
-    r.translate (0, (r.y1 * double (fb_ascent + fb_descent)));
+    r.translate (0, r.y1 * fb_height);
 
     SS_Matrix2D m =  f * r;
     drawOne (canvas, m, fb_all[(g>>28)&0x0f]);
@@ -947,25 +1047,25 @@ SFontFB::drawOne (SCanvas* canvas, const SS_Matrix2D& m, short* guide)
        wid = guide[i+1]; i=i+2;
        break;
     case  SD_MOVE_TO:
-       x = m.x0 * double (guide[i+1]) + m.y0 * double (guide[i+2]) + m.t0;
-       y = m.x1 * double (guide[i+1]) + m.y1 * double (guide[i+2]) + m.t1;
+       x = m.x0 * double (guide[i+1]) + m.y0 * double (guide[i+2]+fb_doffset) + m.t0;
+       y = m.x1 * double (guide[i+1]) + m.y1 * double (guide[i+2]+fb_doffset) + m.t1;
        canvas->moveto (x, y);
        i = i + 3;
        break;
 
     case  SD_LINE_TO:
-       x = m.x0 * double (guide[i+1]) + m.y0 * double (guide[i+2]) + m.t0;
-       y = m.x1 * double (guide[i+1]) + m.y1 * double (guide[i+2]) + m.t1;
+       x = m.x0 * double (guide[i+1]) + m.y0 * double (guide[i+2]+fb_doffset) + m.t0;
+       y = m.x1 * double (guide[i+1]) + m.y1 * double (guide[i+2]+fb_doffset) + m.t1;
        canvas->lineto (x, y);
        i = i + 3;
        break;
     case  SD_CURVE_TO:
-       x0 = m.x0 * double (guide[i+1]) + m.y0 * double (guide[i+2]) + m.t0;
-       y0 = m.x1 * double (guide[i+1]) + m.y1 * double (guide[i+2]) + m.t1;
-       x1 = m.x0 * double (guide[i+3]) + m.y0 * double (guide[i+4]) + m.t0;
-       y1 = m.x1 * double (guide[i+3]) + m.y1 * double (guide[i+4]) + m.t1;
-       x2 = m.x0 * double (guide[i+5]) + m.y0 * double (guide[i+6]) + m.t0;
-       y2 = m.x1 * double (guide[i+5]) + m.y1 * double (guide[i+6]) + m.t1;
+       x0 = m.x0 * double (guide[i+1]) + m.y0 * double (guide[i+2]+fb_doffset) + m.t0;
+       y0 = m.x1 * double (guide[i+1]) + m.y1 * double (guide[i+2]+fb_doffset) + m.t1;
+       x1 = m.x0 * double (guide[i+3]) + m.y0 * double (guide[i+4]+fb_doffset) + m.t0;
+       y1 = m.x1 * double (guide[i+3]) + m.y1 * double (guide[i+4]+fb_doffset) + m.t1;
+       x2 = m.x0 * double (guide[i+5]) + m.y0 * double (guide[i+6]+fb_doffset) + m.t0;
+       y2 = m.x1 * double (guide[i+5]) + m.y1 * double (guide[i+6]+fb_doffset) + m.t1;
        canvas->curveto (x0, y0, x1, y1, x2, y2);
        i = i + 7;
        break;
