@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -35,7 +36,7 @@ static unsigned int cacheSize = 100;
 static unsigned int cacheCount = 0;
 
 /**
- * Afetr this size things wont be cached.
+ * After this size things won't be cached.
  */
 void
 SGEngine::setCacheSize(unsigned int size)
@@ -60,12 +61,15 @@ SGEngine::setCacheOn (bool on)
 
 /**
  * This class is meant to be the base class of canvases
+ * @param windingrule is true if nonzero winding areas should
+ * be filled instead of nonzero intersects.
  */
-SGEngine::SGEngine(void)
+SGEngine::SGEngine(bool windingrule)
 {
   SS_Matrix2D m;
   matrix.append (m);
   setup (1.0);
+  winding = windingrule;
 }
 
 SGEngine::~SGEngine ()
@@ -349,8 +353,8 @@ SGEngine::_fill(int _x, int _y, unsigned int _width, unsigned int _height)
     scanHoriz[i] = new SS_InterSection();
     CHECK_NEW (scanHoriz[i]);
   }
-    scan (scanHoriz, origoX, origoY, (int) height, false);
-    render (image, scanHoriz, width, height, false); 
+  scan (scanHoriz, origoX, origoY, (int) height, false);
+  render (image, scanHoriz, width, height, false); 
   for (i=0; i<height; i++)
   {
     delete scanHoriz[i];
@@ -542,6 +546,8 @@ SGEngine::linetoInternal (double x, double y)
 
 /**
  * can the curves into the buffer
+ * If winding is true use the non-zero winding rule rather that
+ * intersect rule.
  * @param inter in the allocated  and initialisez intersection array
  * @param offsetx is the offset in inter buffer
  * @param offsety is the offset in scanline.
@@ -552,7 +558,29 @@ void
 SGEngine::scan (SS_InterSection** intersBuff, 
   int ox, int oy, int height, bool swap)
 {
-  char* crossBuff = new char[height];
+  if (winding)
+  {
+    scanWinding (intersBuff, ox, oy, height, swap);
+  }
+  else
+  {
+    scanCrosses (intersBuff, ox, oy, height, swap);
+  }
+}
+
+/**
+ * Scan the curves into the buffer using nonzero winding rule.
+ * @param inter in the allocated  and initialisez intersection array
+ * @param offsetx is the offset in inter buffer
+ * @param offsety is the offset in scanline.
+ * @parm hieght is the hieght of the inter buffer 
+ * @param xy is 0 if x is x y is y 1 vice versa.
+ */
+void
+SGEngine::scanWinding (SS_InterSection** intersBuff, 
+  int ox, int oy, int height, bool swap)
+{
+  int* crossBuff = new int[height];
   CHECK_NEW (crossBuff);
 
   double is;
@@ -560,9 +588,149 @@ SGEngine::scan (SS_InterSection** intersBuff,
 
   unsigned int swapindx = (swap) ? 1 : 0;
   unsigned int swapindy = (swap) ? 0 : 1;
+  /* all paths */
+  unsigned int i;
+  unsigned int j;
+  SS_InterSection** clockwise = new SS_InterSection*[(unsigned int)height];
+  CHECK_NEW (clockwise);
+  for (i=0; i<(unsigned int)height; i++)
+  {
+    clockwise[i] = new SS_InterSection();
+    CHECK_NEW (clockwise[i]);
+  }
+  for (i=0; i<pathVector.size(); i++)
+  {
+    memset (crossBuff, 0, height * sizeof (int));
+    /* one path */
+    for (j=0; j+3<pathVector[i].size(); j=j+2)
+    {
+      bool lastone = (j+3+2>=pathVector[i].size());
+      /* Blur the image by half a grid to get a better contour */
+      double x0 = pathVector[i][j+swapindx] * (double) oversample  
+           + 0.5 - (double) ox ;
+      double y0 = pathVector[i][j+swapindy] * (double) oversample 
+           + 0.5 - (double) oy ;
+      double x1 = pathVector[i][j+swapindx+2] * (double) oversample
+           + 0.5 - (double) ox ;
+      double y1 = pathVector[i][j+swapindy+2] * (double) oversample 
+           + 0.5 - (double) oy ;
+
+      int from = (int) y0;
+      int to = (int) y1;
+      int increment = (y0 > y1) ? -1 : 1;
+
+      if (increment > 0)
+      {
+        if (from >= height || to < 0)
+        {
+          continue;
+        }
+        if (from < 0) from = 0;
+        if (to >= height) to  = height-1;
+        diff = y1 - y0;
+      }
+      else
+      {
+        if (to >= height || from < 0)
+        {
+          continue;
+        }
+        if (to < 0) to = 0;
+        if (from >= height) from  = height-1;
+        diff = y0 - y1;
+      }
+      /* scan betbeen  y0..y1 */
+      for (int k=from; ; k+=increment)
+      {
+         double cline = double (k);
+         bool crosses  = (increment > 0) 
+               ? (cline >= y0 && cline < y1) 
+               : (cline >= y1 && cline < y0);
+
+         int lastCross = crossBuff[k];
+         if (lastCross == 0)
+         {
+           lastCross =  -increment;
+         }
+
+         if (increment * lastCross > 0 || !crosses)
+         {
+            if (k==to) break;
+            continue;
+         }
+         crossBuff[k] = increment;
+
+         if (diff < 0.1)
+         {
+           is = x0;
+         }
+         else
+         {
+           is = x0 + (x1 - x0) * (cline - y0) / (y1-y0);
+         }
+         unsigned int pos = intersBuff[k]->appendSorted ((int)is);
+         clockwise[k]->insert (pos, increment);
+         /*
+          * This happend if we got inside/out it all wrong
+          * Mainly happens when we get a straight line first and we
+          * think is is positive.
+          */
+         if (lastone && (intersBuff[k]->size() & 1)!=0)
+         {
+           intersBuff[k]->remove (0);
+           clockwise[k]->remove (0);
+         }
+         if (k==to) break;
+      }
+    }
+  }
+  
+  for (i=0; i<(unsigned int)height; i++)
+  {
+    int swinding = 0;
+    for (j=0; j<intersBuff[i]->size(); )
+    {
+      int oldwin = swinding;
+      swinding += clockwise[i]->peek(j);
+      if (oldwin==0 || swinding==0)
+      {
+         j++; continue;
+      }
+      /* remove the ones that don't go from/to zero */
+      intersBuff[i]->remove (j);
+      clockwise[i]->remove (j);
+    }
+    delete clockwise[i];
+  }
+  delete crossBuff;
+  delete clockwise;
+}
+
+/**
+ * Scan the curves into the buffer using crosses rule
+ * @param inter in the allocated  and initialisez intersection array
+ * @param offsetx is the offset in inter buffer
+ * @param offsety is the offset in scanline.
+ * @parm hieght is the hieght of the inter buffer 
+ * @param xy is 0 if x is x y is y 1 vice versa.
+ */
+void
+SGEngine::scanCrosses (SS_InterSection** intersBuff, 
+  int ox, int oy, int height, bool swap)
+{
+  int* crossBuff = new int[height];
+  CHECK_NEW (crossBuff);
+
+  double is;
+  double diff;
+
+  unsigned int swapindx = (swap) ? 1 : 0;
+  unsigned int swapindy = (swap) ? 0 : 1;
+  /* all paths */
   for (unsigned int i=0; i<pathVector.size(); i++)
   {
-    memset (crossBuff, 0, height);
+    memset (crossBuff, 0, height * sizeof (int));
+    /* one path */
     for (unsigned j=0; j+3<pathVector[i].size(); j=j+2)
     {
       bool lastone = (j+3+2>=pathVector[i].size());
@@ -600,6 +768,7 @@ SGEngine::scan (SS_InterSection** intersBuff,
         if (from >= height) from  = height-1;
         diff = y0 - y1;
       }
+      /* scan betbeen  y0..y1 */
       for (int k=from; ; k+=increment)
       {
          double cline = double (k);
@@ -607,26 +776,18 @@ SGEngine::scan (SS_InterSection** intersBuff,
                ? (cline >= y0 && cline < y1) 
                : (cline >= y1 && cline < y0);
 
-         char lastCross = crossBuff[k];
+         int lastCross = crossBuff[k];
          if (lastCross == 0)
          {
-           lastCross = (char) -increment;
-         }
-         else if (!crosses && increment * (int) lastCross < 0)
-         {
-#if 0
-            /* This is just a sanity check */
-            if (increment > 0 && y0 > cline) crosses = true;
-            if (increment < 0 && y1 < cline) crosses = true;
-#endif
+           lastCross =  -increment;
          }
 
-         if (increment * (int) lastCross > 0 || !crosses)
+         if (increment * lastCross > 0 || !crosses)
          {
             if (k==to) break;
             continue;
          }
-         crossBuff[k] = (char) increment;
+         crossBuff[k] = increment;
 
          if (diff < 0.1)
          {
@@ -639,7 +800,7 @@ SGEngine::scan (SS_InterSection** intersBuff,
          intersBuff[k]->appendSorted ((int)is);
          /*
           * This happend if we got inside/out it all wrong
-          * Mainly happens when we get a straign line first and we
+          * Mainly happens when we get a straight line first and we
           * think is is positive.
           */
          if (lastone && (intersBuff[k]->size() & 1)!=0)
