@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -454,7 +455,7 @@ SWHandler::SWHandler(unsigned int msec, bool iswine)
   }
   else
   {
-    fprintf (stderr, "Waiting for (I think ON is a better word) multiple objects. Throw 'em at me!\n");
+//    fprintf (stderr, "Waiting for (I think ON is a better word) multiple objects. Throw 'em at me!\n");
     _windowsSelectHookup = winSelectHack; 
     timer = 0;
   }
@@ -722,15 +723,15 @@ WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nC
   if (OpenClipboard (hwnd))
   {
      UINT format = 0;
-      fprintf (stderr, "clipboard format start %d\n", hwnd);
+//      fprintf (stderr, "clipboard format start %d\n", hwnd);
      while (format=EnumClipboardFormats (format))
      {
        char buff[128];
        buff[127] = 0;
        GetClipboardFormatName (format, buff, 127);
-       fprintf (stderr, "clipboard format[%u]=[%s]\n", format, buff);
+//       fprintf (stderr, "clipboard format[%u]=[%s]\n", format, buff);
      }
-      fprintf (stderr, "clipboard format end\n");
+//      fprintf (stderr, "clipboard format end\n");
      CloseClipboard ();
   }
   else
@@ -1332,7 +1333,7 @@ _eventHandler (HWND hwnd, UINT message, WPARAM wParam,LPARAM lParam)
          {
            ImmGetCompositionStringW(himc, GCS_RESULTSTR, lpstr, 
              size + sizeof (WCHAR));
-           /* opy unicode over */ 
+           /* copy unicode over */ 
            SV_UCS4 ucs4v;
            SString str (lpstr, (unsigned int)size); 
            for (unsigned int i=0; i+1<str.size(); i=i+2)
@@ -1340,7 +1341,8 @@ _eventHandler (HWND hwnd, UINT message, WPARAM wParam,LPARAM lParam)
               SS_UCS2 u = ((SS_UCS2*) str.array())[i/2];
               ucs4v.append ((SS_UCS4) u);
            }
-           SEncoder utf8enc;
+           /* add surrogates together */
+           SEncoder utf8enc ("utf-8");
            SString out = utf8enc.encode (ucs4v);
            keyData.key = SWindowListener::Key_Send;
            sendKeyChar (&keyData, out);
@@ -1488,10 +1490,12 @@ winSelectHack (int readSize, fd_set *ro, int writeSize, fd_set *wo,
   {
     return 0;
   }
-
-  /* Nothing only timer - or not even exceptSize - forget that */
   SBinVector<WSAEVENT> events;
   SV_INT vmap;
+
+// Win95 does not have it.
+#ifdef HAVE_WS2_32_DLL
+  /* Nothing only timer - or not even exceptSize - forget that */
   unsigned int i;
   /* go through read and write */
   for (i=0; i<readSize; i++)
@@ -1515,6 +1519,10 @@ winSelectHack (int readSize, fd_set *ro, int writeSize, fd_set *wo,
   /* maybe WSA_INFINITE should be a timer ... */
   idx = WSAWaitForMultipleEvents(events.size(), events.array(),
     FALSE, millisec, FALSE);
+#else
+  DWORD idx;
+#endif
+
   /*
    * We need this one because WSAWaitForMultipleEvents 
    * will not react to window events. in case events.array is
@@ -1525,10 +1533,12 @@ winSelectHack (int readSize, fd_set *ro, int writeSize, fd_set *wo,
   if (readSize == 0 && ro != 0) FD_ZERO (ro);
   if (writeSize == 0 && wo != 0) FD_ZERO (wo);
   int lerr = GetLastError();
+#ifdef HAVE_WS2_32_DLL
   for (i=0; i<events.size(); i++)
   {
     WSACloseEvent (events[i]);
   }
+#endif
   if (idx == 0xFFFFFFFF)
   {
     fprintf (stderr, "SEventBSD::WaitForMultipleObjects error (%d)\n", lerr);
@@ -1664,6 +1674,18 @@ SW32Window::bitpoint (const SColor& clr, int _x, int _y)
   dcout (mydc);
 }
 
+void
+SW32Window::bitpoints (const SColor& clr, const int* _x, const int* _y, 
+         unsigned int _size)
+{
+  bool mydc = dcin();
+  COLORREF ref = PALETTERGB (clr.red, clr.green, clr.blue);
+  for (unsigned int i=0; i<_size; i++)
+  {
+    ::SetPixel ((HDC)cdc, _x[i], _y[i], ref);
+  }
+  dcout (mydc);
+}
 
 /**
  * Afetr this size things wont be cached.
@@ -1671,7 +1693,7 @@ SW32Window::bitpoint (const SColor& clr, int _x, int _y)
 void
 SW32Window::setPixmapCacheSize(unsigned int _size)
 {
-  fprintf (stderr, "setPixmapCacheSize=%u\n", _size);
+//  fprintf (stderr, "setPixmapCacheSize=%u\n", _size);
   imageCache.setSize (_size);
   maskCache.setSize (_size);
 }
@@ -2382,7 +2404,7 @@ SW32Window::getClipUTF8()
    SW32Window* w = windowHashtable.get (clipboardOwner);
    if (w != 0) /* local guy */
    {
-      SEncoder utf8enc;
+      SEncoder utf8enc ("utf-8-s");
       SString out = utf8enc.encode (clipData);
       return SString (out);
    }
@@ -2408,6 +2430,18 @@ SW32Window::getClipUTF8()
          unsigned int lsize = GlobalSize (h);
          cld = SString (str, lsize);
          GlobalUnlock (h);
+         /* we need to terminate ucs2 if 0 is seen */
+         if (ucs)
+         {
+           for (unsigned int i=0; i+1<lsize; i=i+2)
+           {
+             if (cld[i] == 0 && cld[i+1]==0)
+             {
+                cld.truncate (i);
+                break;
+             }
+           }
+         }
        }
      }
      else 
@@ -2437,10 +2471,10 @@ SW32Window::getClipUTF8()
      return SString (cld);
    }
    impl->encoder.clear();
-   SEncoder ucsenc("ucs-2-le");
+   SEncoder ucsenc("utf-16-le");
    SV_UCS4 ucstext = (ucs) ?  ucsenc.decode (cld) : impl->encoder.decode (cld);
 
-   /* stupid windows dows not know how to copy u+0000 - this
+   /* stupid windows does not know how to copy u+0000 - this
     * is the terminating character.
     */
    for (unsigned int i=0; i<ucstext.size(); i++)
@@ -2451,14 +2485,14 @@ SW32Window::getClipUTF8()
        break;
      }
    }
-   SEncoder utf8enc;
+   SEncoder utf8enc ("utf-8-s");
    return SString (utf8enc.encode (ucstext));
 }
 
 void
 SW32Window::putClipUTF8(const SString& utf8)
 {
-  SEncoder utf8enc;
+  SEncoder utf8enc ("utf-8-s");
   clipData = utf8enc.decode (utf8);
 
   if (!clipChained)
@@ -2485,7 +2519,7 @@ putClipUnicodeText()
 {
   // They sell this API for money! Unbelievable!
   if (clipData.size() == 0) return 1;
-  SEncoder ucs2enc("ucs-2-le");
+  SEncoder ucs2enc("utf-16-le");
   SV_UCS4 v = clipData;
   v.append (0);
   SString out = ucs2enc.encode (v);
@@ -2553,7 +2587,7 @@ static int
 putClipUtf8Text()
 {
   if (clipData.size() == 0) return 1;
-  SEncoder utf8enc;
+  SEncoder utf8enc ("utf-8-s");
   SString out = utf8enc.encode (clipData);
   /* this is our null */
   SString nl; nl.append ((char)0);
@@ -2708,9 +2742,11 @@ sendKeyChar (KeyData* kd, const SString& _s)
   SString s = _s;
   if (s.size() == 0) return;
   unsigned char c0 = (unsigned char) s[0];
+
+  /* TABS and controls are handled in processKey */
   if (s.size() == 1 && c0 < 0x20)
   {
-     if (!kd->ctrl0 && !kd->ctrl1)  return;
+     return;
   }
   /* 
    * Deal with only 'pressed' keys. If focus changes while repeat,
@@ -2769,11 +2805,11 @@ sendKeyReleased (KeyData* kd, SW32Window* wn, SWindowListener* ln)
 
   if (kd->shift0)
   { 
-    ln->keyReleased (wn, SWindowListener::Key_Alt_L, s, ctrl, shift, meta);
+    ln->keyReleased (wn, SWindowListener::Key_Shift_L, s, ctrl, shift, meta);
   } 
   if (kd->shift1)
   { 
-    ln->keyReleased (wn, SWindowListener::Key_Alt_R, s, ctrl, shift, meta);
+    ln->keyReleased (wn, SWindowListener::Key_Shift_R, s, ctrl, shift, meta);
   } 
 
   kd->shift0 = false;
@@ -2843,9 +2879,18 @@ processKey (KeyData* kdin, bool syskey, int keycod, bool isdown)
     kd.ctrl1 = true; 
     kd.key = SWindowListener::Key_Control_R;
     break;
-  case VK_SHIFT: kd.shift0 = true; break;
-  case VK_LSHIFT: kd.shift0 = true; break;
-  case VK_RSHIFT: kd.shift1 = true; break;
+  case VK_SHIFT: 
+     kd.shift0 = true; 
+    kd.key = SWindowListener::Key_Shift_R;
+    break;
+  case VK_LSHIFT: 
+    kd.key = SWindowListener::Key_Shift_L;
+    kd.shift0 = true; 
+    break;
+  case VK_RSHIFT: 
+    kd.key = SWindowListener::Key_Shift_R;
+    kd.shift1 = true; 
+    break;
 
   case VK_MENU: kd.key = SWindowListener::Key_Alt_L; kd.meta0 = true; break;
   case VK_LMENU: kd.key = SWindowListener::Key_Alt_L; kd.meta0 = true; break;
@@ -2928,44 +2973,35 @@ processKey (KeyData* kdin, bool syskey, int keycod, bool isdown)
     kdin->ctrl0 =  ! kd.ctrl0; kdin->ctrl0 = ! kd.ctrl0;
   }
 
-  /* check if we need to send shift ctrl meta */
-  /* no need to send - we will have it later */
+  /*  */
   bool ctrl = kdin->ctrl0 || kdin->ctrl1; 
   bool shift = kdin->shift0 || kdin->shift1; 
   bool meta = kdin->meta0 || kdin->meta1; 
-  if (!ckey || syskey)
+
+  SW32Window* wn = windowHashtable.get (currentFocusWindow);
+  SWindowListener* ln = listenerHashtable.get (currentFocusWindow);
+
+  if (wn != 0 || ln != 0)
   {
-    SW32Window* wn = windowHashtable.get (currentFocusWindow);
-    SWindowListener* ln = listenerHashtable.get (currentFocusWindow);
-    if (wn != 0 || ln != 0)
+    SString s;
+    /* WM_IME_COMPOSITION does not give us TAB */
+    if (keycod == VK_TAB && isdown) 
     {
-      SString s;
-      if (isdown)
-      {
-        if (!sendAcceleratorPressed ((int) kdin->key, ctrl, shift, meta))
-        {
-           ln->keyPressed (wn, kdin->key, s, ctrl, shift, meta);
-        }
-      }
-      else
-      {
-        if (!sendAcceleratorReleased ())
-        {
-          ln->keyReleased (wn, kdin->key, s, ctrl, shift, meta);
-        }
-      }
+      s.append ("\t");
     }
-  }
-  /* shift will get sent the "normal" way. */
-  if (ctrl | meta) 
-  {
     if (isdown)
-    {
-      sendAcceleratorPressed ((int) kdin->key, ctrl, shift, meta);
+    { 
+      if (!sendAcceleratorPressed ((int) kdin->key, ctrl, shift, meta))
+      {
+         ln->keyPressed (wn, kdin->key, s, ctrl, shift, meta);
+      }
     }
     else
     {
-      sendAcceleratorReleased ();
+      if (!sendAcceleratorReleased ())
+      {
+        ln->keyReleased (wn, kdin->key, s, ctrl, shift, meta);
+      }
     }
   }
 }
@@ -3053,6 +3089,13 @@ sendAcceleratorReleased ()
 bool
 SW32Window::startInputMethod (const SString& name, const SProperties& prop)
 {
+  if (name == "x-none" || name == "x-ascii" || name == "x-utf-8")
+  {
+    if (imname == name) return true;
+    stopInputMethod();
+    imname = name;
+    return true;
+  }
   HIMC himc = ImmGetContext((HWND)id);
   if (!himc) return false;
   /* ImmSetOpenStatus */
@@ -3090,6 +3133,7 @@ SW32Window::stopInputMethod ()
 void
 SW32Window::setInputMethodProperties (const SProperties& properties)
 {
+  if (!isVisible()) return;
   HIMC himc = ImmGetContext((HWND)id);
   if (!himc) return;
 
@@ -3098,25 +3142,8 @@ SW32Window::setInputMethodProperties (const SProperties& properties)
     fprintf (stderr, "InputStyle is not present in properties.\n");
     return;
   }
-  int style = 0;
+
   SString s = properties["InputStyle"];
-  if (s=="root")
-  {
-    style = 1;
-  }
-  else if (s =="off-the-spot")
-  {
-    style = 2;
-  }
-  else if (s == "over-the-spot")
-  {
-    style = 3;
-  }
-  else
-  {
-    fprintf (stderr, "InputStyle is bad - %*.*s.\n", SSARGS(s));
-    return;
-  }
 
   /* ok. now I can tell you windows can not set InputStyle sorry */
    
@@ -3129,34 +3156,122 @@ SW32Window::setInputMethodProperties (const SProperties& properties)
     sscanf (lsp.array(), "%d", &spacing);
   }
 
-  if (properties.get ("Color"))
+  /* What to do with this? */
+  if (properties.get ("InputClientColor"))
   {
-    SString col = properties["Color"];
+    SString col = properties["InputClientColor"];
     col.append ((char)0);
     unsigned long bg, fg;
     sscanf (col.array(), "%lu,%lu", &bg, &fg);
     SColor xbg = SColor((SS_WORD32)bg);
     SColor xfg = SColor((SS_WORD32)fg);
   }
-  /* over the spot */
-  // style style == 3
-  if (properties.get ("InputSpot"))
+
+  /* XXX: no idea how to do this... */
+  if (s == "preedit-over-status-under" 
+       && properties.get ("InputSpot")
+       && properties.get ("InputStatusLocation")
+       && properties.get ("InputStatusSize")
+     )
+  {
+    SString spotLocation = properties["InputSpot"];
+    spotLocation.append ((char)0);
+    int _x, _y;
+    sscanf (spotLocation.array(), "%d,%d", &_x, &_y);
+    COMPOSITIONFORM form;
+    form.dwStyle = CFS_POINT;
+    form.ptCurrentPos.x = _x;
+    form.ptCurrentPos.y = _y;
+    form.rcArea.left = 0;
+    form.rcArea.top = 0;
+    form.rcArea.right =  (int) getWidth();
+    form.rcArea.bottom = (int) getHeight();
+    ImmSetCompositionWindow (himc, &form);
+
+    SString sl = properties["InputStatusLocation"];
+    sl.append ((char)0);
+    int statusX, statusY;
+    sscanf (sl.array(), "%d,%d", &statusX, &statusY);
+
+    SString ss = properties["InputStatusSize"];
+    ss.append ((char)0);
+    int statusWidth, statusHeight;
+    sscanf (ss.array(), "%d,%d", &statusWidth, &statusHeight);
+
+    POINT point;
+    point.x = statusX;
+    point.y = statusY;
+    ImmSetStatusWindowPos(himc, &point);
+  }
+  else if (s == "preedit-under-status-under" 
+       && properties.get ("InputSpot")
+       && properties.get ("InputStatusLocation")
+       && properties.get ("InputStatusSize")
+       && properties.get ("InputClientLocation")
+       && properties.get ("InputClientSize")
+       )
+  {
+
+    SString spotLocation = properties["InputSpot"];
+    spotLocation.append ((char)0);
+    int _x, _y;
+    sscanf (spotLocation.array(), "%d,%d", &_x, &_y);
+
+    SString sl = properties["InputStatusLocation"];
+    sl.append ((char)0);
+    int statusX, statusY;
+    sscanf (sl.array(), "%d,%d", &statusX, &statusY);
+
+    SString ss = properties["InputStatusSize"];
+    ss.append ((char)0);
+    int statusWidth, statusHeight;
+    sscanf (ss.array(), "%d,%d", &statusWidth, &statusHeight);
+
+    SString cl = properties["InputClientLocation"];
+    cl.append ((char)0);
+    int clientX, clientY;
+    sscanf (cl.array(), "%d,%d", &clientX, &clientY);
+
+    SString cs = properties["InputClientSize"];
+    cs.append ((char)0);
+    int clientWidth, clientHeight;
+    sscanf (cs.array(), "%d,%d", &clientWidth, &clientHeight);
+
+    POINT point;
+    point.x = statusX;
+    point.y = statusY;
+    ImmSetStatusWindowPos(himc, &point);
+
+    COMPOSITIONFORM form;
+    form.dwStyle = CFS_RECT;
+    form.ptCurrentPos.x = clientX; // starting from.
+    form.ptCurrentPos.y = clientY;
+    form.rcArea.left = clientX;    // next line
+    form.rcArea.top = clientY;
+    form.rcArea.right =  clientX +  clientWidth;
+    form.rcArea.bottom =  clientY + clientHeight;
+    ImmSetCompositionWindow (himc, &form);
+  }
+  else if (s == "preedit-root-status-root")
+  {
+    COMPOSITIONFORM form;
+    form.dwStyle = CFS_DEFAULT;
+    form.ptCurrentPos.x = (int) getWidth();
+    form.ptCurrentPos.y = (int) getHeight();
+    form.rcArea.left = 0;
+    form.rcArea.top = 0;
+    form.rcArea.right =  (int) getWidth();
+    form.rcArea.bottom = (int) getHeight();
+    ImmSetCompositionWindow (himc, &form);
+  }
+  /* All the input styles */
+  else if (s == "preedit-over-status-over" && properties.get ("InputSpot"))
   {
     SString spotLocation = properties["InputSpot"];
     spotLocation.append ((char)0);
     int _x, _y;
     sscanf (spotLocation.array(), "%d,%d", &_x, &_y);
-    POINT point;
-    point.x = _x;
-    point.y = _y;
-    if (style==2)
-    {
-      ImmSetStatusWindowPos(himc, &point);
-      return;
-    }
-    //fprintf (stderr, "Setting input spot=%d %d\n", _x, _y);
     COMPOSITIONFORM form;
-    //form.dwStyle = (style==1)?CFS_DEFAULT : CFS_POINT;
     form.dwStyle = CFS_POINT;
     form.ptCurrentPos.x = _x;
     form.ptCurrentPos.y = _y;
@@ -3177,3 +3292,9 @@ SW32Window::getInputMethod ()
 {
   return SString(imname);
 }
+
+unsigned long
+SW32Window::getWindowID() const
+{
+  return (unsigned long) id;
+}
