@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -42,7 +43,7 @@ static SStringVector scriptCodeMap("*-iso10646-yudit,*-iso10646-dev,*-iso10646-b
  */
 
 static bool inited= false;
-static char* defaultNames[] =
+static const char* defaultNames[] =
 {
 "*-iso10646-1", "ucs2",
 "*-iso10646-dev", "ucs2",
@@ -75,6 +76,8 @@ static char* defaultNames[] =
 "*-ksc5601.1987-0", "ksc-5601-l",
 "*-ksc5601*", "ksc-5601-l",
 
+"*-ksx1001*", "ksx-1001",
+
 "*-jisx0208.1983-0", "jis-0208",
 "*-jisx0208*", "jis-0208",
 
@@ -87,6 +90,9 @@ static char* defaultNames[] =
 "*-jisx0212.1990-0", "jis-0212",
 "*-jisx0212*-0", "jis-0212",
 
+"*-jisx0213*-1", "jis-0213-1",
+"*-jisx0213*-2", "jis-0213-2",
+
 "*-big5.hku-0", "big-5",
 "*-big5-0", "big-5",            
 "*-big5*", "big-5",            
@@ -128,6 +134,9 @@ public:
   SProperties ligatures;
   unsigned int ligatureCount;
   int scriptCode;
+  /* Still no agreement on this - I am using : */ 
+  /* http://mail.nl.linux.org/linux-utf8/2000-12/msg00042.html */
+  unsigned int plane;
 };
 
 
@@ -139,6 +148,7 @@ SX11FontCache::SX11FontCache(SX11Impl *_impl,  const SString& _xlfd)
   fontDescent = 0.0;
   fontWidth = 0.0;
   loaded = false;
+  plane = 0;
 }
 
 /**
@@ -150,6 +160,7 @@ SX11FontCache::load ()
 {
   if (loaded) return (font!=0);
   scriptCode = -1;
+  plane = 0;
   loaded = true;
   char* fontName = xlfd.cString();
   font = XLoadQueryFont (impl->display, fontName);
@@ -211,6 +222,13 @@ SX11FontCache::load ()
   /* Find out what map is used for this font */
   SString m;
   unsigned int i;
+  for (i=0; i<scriptCodeMap.size(); i++)
+  {
+     if (xlfd.match (scriptCodeMap[i]))
+     {
+        scriptCode = (int)i;
+     }
+  }
   for (i=0; i+1<mapNameVector.size(); i=i+2)
   {
     if (xlfd.match (mapNameVector[i]))
@@ -219,12 +237,27 @@ SX11FontCache::load ()
       break;
     }
   }
-  for (i=0; i<scriptCodeMap.size(); i++)
+  for (i=1; i<17; i++)
   {
-     if (xlfd.match (scriptCodeMap[i]))
-     {
-        scriptCode = (int)i;
-     }
+    /* http://mail.nl.linux.org/linux-utf8/2000-12/msg00024.html */
+    SString s("*-iso10646p2-");
+    s.print (i);
+    if (xlfd.match (s))
+    {
+      plane = i;
+      if ((m.size() ==0)) m=SString("ucs2");
+      break;
+    }
+/* Hmm.
+    SString s2("*-iso10646.2-");
+    s2.print (i);
+    if (xlfd.match (s2))
+    {
+      plane = i;
+      if ((m.size() ==0)) m=SString("ucs2");
+      break;
+    }
+*/
   }
   if (m.size() ==0)
   {
@@ -337,7 +370,7 @@ SX11FontCache::draw (SCanvas* canvas, const SPen& pen, double x, double y,
 {
   /* ZWJ and ZWNJ */
   /* ZWJ and ZWNJ */
-  if (scriptCode != -1 
+  if (plane == 0 && scriptCode != -1 
         && (glyphno == 0x200d 
          || glyphno == 0x200c))
   {
@@ -345,7 +378,7 @@ SX11FontCache::draw (SCanvas* canvas, const SPen& pen, double x, double y,
   }
   bool covered = isCoveredScipt ((SS_UCS4) glyphno, scriptCode);
   /* pango subscript and superscirpt ra - zero width */
-  if (!covered && scriptCode != -1 && (glyphno == 0xc97e || glyphno == 0xc97f))
+  if (plane == 0 && !covered && scriptCode != -1 && (glyphno == 0xc97e || glyphno == 0xc97f))
   {
     covered = true;
   }
@@ -398,13 +431,13 @@ double
 SX11FontCache::width (SS_UCS2 glyphno)
 {
   /* ZWJ and ZWNJ */
-  if (scriptCode != -1 && (glyphno == 0x200d || glyphno == 0x200c))
+  if (plane == 0 && scriptCode != -1 && (glyphno == 0x200d || glyphno == 0x200c))
   {
       return SD_ZW_FOTNTHACK;
   }
   /* pango subscript and superscirpt ra - zero width */
   bool covered = isCoveredScipt ((SS_UCS4) glyphno, scriptCode);
-  if (!covered && scriptCode != -1 && (glyphno == 0xc97e || glyphno == 0xc97f))
+  if (plane == 0 && !covered && scriptCode != -1 && (glyphno == 0xc97e || glyphno == 0xc97f))
   {
     covered = true;
   }
@@ -462,8 +495,28 @@ SX11Font::SX11Font (SX11Impl* _impl, const SString& _encoding)
   isencoded = (_encoding.size()!=0);
   if (isencoded && !encoder.isOK())
   {
-    fprintf (stderr, "SX11Font: unknown encoding : '%*.*s'\n",
-       SSARGS (_encoding));
+    if (_encoding != "unicode"
+     && _encoding != "mslvt"
+     && _encoding != "nojamo"
+     && _encoding != "indic"
+     && _encoding != "deva"
+     && _encoding != "beng"
+     && _encoding != "guru"
+     && _encoding != "gujr"
+     && _encoding != "orya"
+     && _encoding != "taml"
+     && _encoding != "telu"
+     && _encoding != "knda"
+     && _encoding != "mlym"
+     && _encoding != "sinh"
+     && _encoding != "thai"
+     && _encoding != "lao"
+     && _encoding != "tibt"
+    )
+    {
+      fprintf (stderr, "SX11Font: unknown encoding : '%*.*s'\n",
+         SSARGS (_encoding));
+    }
     isencoded = false;
   }
 }
@@ -538,7 +591,7 @@ SX11Font::draw (const SString& xlfd, SCanvas* canvas, const SPen& pen,
       SV_UCS4 ucs4; ucs4.append (g); SV_UCS4 decd;
       unsigned int lifted = encoder.lift (ucs4, 0, false, &decd);
       /* like TTF we fallback to original font encoding  */
-      if (lifted != 0)
+      if (lifted == 0)
       {
          glyphno = fc->map.encode (g);
       } 
@@ -563,7 +616,22 @@ SX11Font::draw (const SString& xlfd, SCanvas* canvas, const SPen& pen,
   }
   else
   {
-    glyphno = fc->map.encode (g);
+    if (fc->plane == 0)
+    {
+      /* faster */
+      if (fc->map.isUMap()) {
+        glyphno = fc->map.encode (g);
+      } else {
+        SV_UCS4 ucs4; ucs4.append (g); SV_UCS4 decd;
+        unsigned int lifted = fc->map.lift (ucs4, 0, false, &decd);
+        if (lifted==1 && decd.size()==1) glyphno = decd[0];
+      }
+    }
+    else
+    {
+       SS_UCS4  start = fc->plane*0x10000;;
+       glyphno =  (g < start) ? 0 : fc->map.encode (g-start);
+    }
   }
   if (glyphno==0) return false;
   double dw = fc->draw (canvas, pen, matrix.t0, matrix.t1, glyphno);
@@ -602,8 +670,11 @@ SX11FontCache::getRawPangoLigatures (unsigned int script,
       {
         int type0 = getCharType (ino[curin]);
         int type1 = getCharType (ino[curin+1]);
-        if (type0 == SD_INDIC_CONSONANT && 
-           type1 == SD_INDIC_HALANT)
+        if ((type0 == SD_INDIC_CONSONANT_BASE
+            || type0 == SD_INDIC_CONSONANT_POST_BASE
+            || type0 == SD_INDIC_CONSONANT_BELOW_BASE
+            ) 
+           && type1 == SD_INDIC_HALANT)
         {
           /* PANGO intermediate half form */
           SS_UCS4 gl = 0;
@@ -906,7 +977,21 @@ SX11Font::width (const SString& xlfd, SS_UCS4 g, double *width_)
   }
   else
   {
-    glyphno = fc->map.encode (g);
+    if (fc->plane == 0)
+    {
+      if (fc->map.isUMap()) {
+        glyphno = fc->map.encode (g);
+      } else {
+        SV_UCS4 ucs4; ucs4.append (g); SV_UCS4 decd;
+        unsigned int lifted = fc->map.lift (ucs4, 0, false, &decd);
+        if (lifted==1 && decd.size()==1) glyphno = decd[0];
+      }
+    }
+    else
+    {
+       SS_UCS4  start = fc->plane*0x10000;;
+       glyphno =  (g < start) ? 0 : fc->map.encode (g-start);
+    }
   }
   if (glyphno==0) return false;
   double dw;
