@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -36,7 +37,10 @@ static SWindowListener::SKey keySymOf (KeySym sym);
 static SX11Window* keyboardFocusWindow = 0;;
 static SX11Window* lastKeyboardFocusWindow = 0;;
 static bool isAncestor (Display * d, long root, long parent, long w);
-int handleX11Errors (Display *, XErrorEvent *);
+extern "C"
+{
+  int handleX11Errors (Display *, XErrorEvent *);
+}
 
 /**
  * This should be 1.
@@ -186,18 +190,41 @@ SX11Impl::doX()
   }
   if (!XPending(display)) 
   {
-     XFlush (display);
-     if (!XPending(display)) return false;
+     return false;
   }
 
   /* block - true */
-  if (XPending(display)) {
+  {
     /* we dont process things in a nomral socket event looop. */
     /* everything is moved to job - done. */
     XEvent event;
     /* NextEvent sucks in all events from socket, and pass the first queued. */
     XNextEvent (display, &event);
 
+    /* XIC needs this  */
+
+    if (keyboardFocusWindow != 0 && 
+        (event.type == KeyPress || event.type == KeyRelease))
+    {
+      event.xkey.window = (Window)keyboardFocusWindow->getID();
+      event.xkey.subwindow = None;
+      /* This precaution is because if input method dies....*/
+      if (!keyboardFocusWindow->isAsciiInput())
+      {
+        if (XFilterEvent (&event, event.xkey.window))
+        {
+          return true;
+        }
+      }
+    }
+    else if (XFilterEvent(&event, None))
+    {
+      return true;
+    }
+
+    /* Pixmap Copy */
+    if (event.xany.type == NoExpose) return true;
+
     /* event handlers are our first priority */
     SString ekey((long)event.xany.window);
     ekey.append (SString ((long) event.xany.type));
@@ -367,16 +394,76 @@ SX11Impl::doX()
       {
         /* events that hapen during conversion that dont have sent_event
          * set should be ignored */
-        char asci[16];
-        KeySym key;
-        int cnt = XLookupString (&event.xkey, asci, sizeof (asci), &key, 0);
-        SString s(asci, cnt);
+
+        /* Don't try to make this bigger because it will screw up 
+           Xutf8LookupString at and above 65 (try newline in ami) */
+        char asci[64]; 
+        KeySym key = 0;
+        XIC ic = sw->getIC();
+        SString enc = sw->getICEncoding();
+        Status status_return;
+        int cnt = 0; 
+        SString s;
+        if (sw->isAsciiInput() || ic==0)
+        {
+           cnt = XLookupString (&event.xkey, 
+              asci, sizeof (asci), &key, 0);
+           status_return = 1;
+           s = SString (asci, (unsigned int)cnt);
+        }
+        else if (enc.size()>0)
+        {
+          cnt = XmbLookupString (ic, &event.xkey, 
+             asci, sizeof (asci), &key, &status_return);
+          if (status_return==XLookupChars || XLookupBoth)
+          {
+             SEncoder dec (enc);
+             SString str (asci, (unsigned int)cnt);
+             SV_UCS4 sv = dec.decode (str);
+             SEncoder enc ("utf-8-s");
+             s = enc.encode (sv);
+          }
+        }
+        else
+        {
+#ifdef X_HAVE_UTF8_STRING
+          cnt = Xutf8LookupString (ic, &event.xkey, 
+             asci, sizeof (asci), &key, &status_return);
+          if (status_return==XLookupChars || XLookupBoth)
+          {
+             s = SString (asci, (unsigned int)cnt);
+          }
+
+#else
+          cnt = XmbLookupString (ic, &event.xkey, 
+             asci, sizeof (asci), &key, &status_return);
+          if (status_return==XLookupChars || status_return==XLookupBoth)
+          {
+             s = SString (asci, (unsigned int)cnt);
+          }
+#endif
+        }
+        if (cnt == 0 && (key & 0xff000000) == 0x01000000)
+        {
+           SV_UCS4 ucs4;
+           ucs4.append ((SS_UCS4)(key & 0x00ffffff));
+           SEncoder enc ("utf-8-s");
+           s = enc.encode (ucs4);
+           status_return = 0;
+        }
         bool ctrl = (event.xkey.state & ControlMask) ? true : false;
         bool shift = (event.xkey.state & ShiftMask) ? true : false;
         bool meta = (event.xkey.state & Mod1Mask) ? true : false;
-        SWindowListener::SKey skey =  keySymOf(key);
+        SWindowListener::SKey skey =  status_return
+         ? keySymOf(key) : SWindowListener::Key_Undefined;
         if (event.type==KeyPress) 
         {
+          /* Let kinput2 deal with the event. */
+          /* FIXME: Works fine, but too naive.  Ideally, we should devise
+           * a mechanism that harmonizes the key bindings of yudit with
+           * those of an XIM. */
+          if (sw->isKInput() && (ctrl || meta)) break;
+
           if (sendAcceleratorPressed ((int) skey, ctrl, shift, meta)) break;
           /* in case we forgot to get this. */
           if (sendAcceleratorReleased ()) break;
@@ -385,10 +472,14 @@ SX11Impl::doX()
         {
           if (sendAcceleratorReleased ()) break;
         }
-        if (!event.xkey.send_event && sw->getInputMethod().size() > 0
-           && sw == keyboardFocusWindow)
+        if (!event.xkey.send_event && sw->isKInput() 
+              && sw == keyboardFocusWindow)
         {
           //fprintf (stderr, "Ignoreing keys.\n");
+          if (event.type==KeyRelease)
+          {
+            l->keyReleased (sw, skey, s, ctrl, shift, meta);
+          }
           break;
         }
         if (isOKToDeliver (sw->getID()))
@@ -512,13 +603,18 @@ SX11Impl::doX()
        break;
     case ButtonPress:
        if (!isOKToDeliver (sw->getID())) break;
+       //Enable Mousewheel functions (Addition by Maarten van Gompel <proycon@anaproy.homeip.net>
        if (event.xbutton.button == Button1) {
          l->buttonPressed (sw, 0, event.xbutton.x, event.xbutton.y);
        } else if (event.xbutton.button == Button2) {
          l->buttonPressed (sw, 1, event.xbutton.x, event.xbutton.y);
        } else if (event.xbutton.button == Button3) {
          l->buttonPressed (sw, 2, event.xbutton.x, event.xbutton.y);
-       }
+       } else if (event.xbutton.button == Button4) {
+         l->buttonPressed (sw, 3, event.xbutton.x, event.xbutton.y);
+       } else if (event.xbutton.button == Button5) {
+         l->buttonPressed (sw, 4, event.xbutton.x, event.xbutton.y);
+	}
        break;
     case DestroyNotify:
       break;
@@ -538,6 +634,7 @@ SX11Impl::done (const SEventSource* s)
   hasJob = false;
   //static unsigned int ccount=0;
   doXLoop (); /* this will process readable */
+
   SRedrawEventTable t;
   /* redraw block */
   do {
@@ -565,6 +662,11 @@ SX11Impl::done (const SEventSource* s)
           continue;
         }
 
+         //Pixmap e = imageCache.get (i, j);
+         //if (e != 0) XFreePixmap (impl->display, e);
+       // Flicker free
+       //Pixmap p = XCreatePixmap (display, (Window) wid, 
+       //         evt->width, evt->height, depth)
 
         //if (!swid->mapNotified) continue;
         if (evt->clear)
@@ -575,7 +677,6 @@ SX11Impl::done (const SEventSource* s)
              evt->y, evt->width, evt->height, 0);
           }
         }
-        /* FIXME: why do I have such funny events? I have a bug in merge... */
 
         /* This is the first expose. */
         if (evt->width == 0 || evt->height == 0)
@@ -727,22 +828,6 @@ SX11Impl::getWindow (SWindowListener* l, const SString& name)
                 rectangle.width, rectangle.height, border, foreground,
                 background);
   char* n = name.cString();
-  XSelectInput (display, w, 
-      ExposureMask 
-    | StructureNotifyMask 
-    | SubstructureNotifyMask 
-    | ButtonPressMask 
-    | ButtonReleaseMask 
-    | Button1MotionMask 
-    | Button2MotionMask 
-    | Button3MotionMask 
-    | KeyPressMask 
-    | KeyReleaseMask 
-    | KeymapStateMask
-    | PropertyChangeMask
-    | EnterWindowMask
-    | LeaveWindowMask
-    | FocusChangeMask);
 
   XClassHint    classHints;
   classHints.res_name =  n;
@@ -859,6 +944,8 @@ keySymOf (KeySym sym)
   case XK_z: return SWindowListener::Key_z;
   case XK_Z: return SWindowListener::Key_Z;
 
+  case XK_Shift_L: return SWindowListener::Key_Shift_R;
+  case XK_Shift_R: return SWindowListener::Key_Shift_L;
   case XK_Control_L: return SWindowListener::Key_Control_R;
   case XK_Control_R: return SWindowListener::Key_Control_L;
   case XK_Alt_L: return SWindowListener::Key_Alt_L;
@@ -1130,7 +1217,7 @@ SX11Impl::clipEvent (const XEvent& event)
   SEncoder utf8e;
   SString  str;
 
-  if (sw->getID() != (long)event.xany.window)
+  if (sw == 0 || sw->getID() != (long)event.xany.window)
   {
     fprintf (stderr, "oops SX11Impl::clipEvent - clear.\n");
     return;
@@ -1775,22 +1862,25 @@ SX11Impl::getTopLevelWindow (Window w)
 /**
  * does it make us more robust?
  */
-int
-handleX11Errors (Display *display, XErrorEvent *event)
+extern "C"
 {
-  char buffer_return[256];
-  buffer_return[0] = 0;
-  XGetErrorText (display, event->error_code, 
-       buffer_return, sizeof (buffer_return)-1);
-  buffer_return[255] = 0;
-  if ( event->request_code == X_SetInputFocus)
+  int
+  handleX11Errors (Display *display, XErrorEvent *event)
   {
-    //fprintf(stderr, "XSetInputFocus failed (not an error).\n");
+    char buffer_return[256];
+    buffer_return[0] = 0;
+    XGetErrorText (display, event->error_code, 
+         buffer_return, sizeof (buffer_return)-1);
+    buffer_return[255] = 0;
+    if ( event->request_code == X_SetInputFocus)
+    {
+      //fprintf(stderr, "XSetInputFocus failed (not an error).\n");
+    }
+    else
+    {
+      fprintf(stderr, "X11Impl.cpp request_code=%d:\n%s\n", event->request_code, 
+          buffer_return);
+    }
+    return 0;
   }
-  else
-  {
-    fprintf(stderr, "X11Impl.cpp request_code=%d:\n%s\n", event->request_code, 
-        buffer_return);
-  }
-  return 0;
 }
