@@ -1,6 +1,7 @@
 /** 
  *  Yudit Unicode Editor Source File
  *
+ *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
  *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
@@ -59,11 +60,27 @@ void
 SX11Window::setPixmapCacheOn (bool on)
 {
   pixmapCacheOn=on;
-  needClear = true;
+  needClear = false;
 }
 
+#define SD_YUDIT_EVENT_MASK ( ExposureMask \
+  | StructureNotifyMask \
+  | SubstructureNotifyMask \
+  | ButtonPressMask \
+  | ButtonReleaseMask \
+  | Button1MotionMask \
+  | Button2MotionMask \
+  | Button3MotionMask \
+  | KeyPressMask \
+  | KeyReleaseMask \
+  | KeymapStateMask \
+  | PropertyChangeMask \
+  | EnterWindowMask \
+  | LeaveWindowMask \
+  | FocusChangeMask)
+
 SX11Window::SX11Window(const SString& n, SX11Impl* _impl, long _id) : name (n),
- inputMethod (_impl, _id),
+ inputMethod (_impl, _id, SD_YUDIT_EVENT_MASK),
  pen (SColor(0), SColor(0xffffffff)),
  background (_impl, SS_WORD32(0xffffffff))
 {
@@ -92,6 +109,7 @@ SX11Window::SX11Window(const SString& n, SX11Impl* _impl, long _id) : name (n),
   setBackground (background);
   engine = 0;
   shown = false;
+  XSelectInput (impl->display, (Window) id, SD_YUDIT_EVENT_MASK);
 }
 
 SX11Window::~SX11Window()
@@ -374,12 +392,20 @@ SX11Window::putImage (int _x, int _y, const SImage& im)
     {
       createShadedPixmap (im, &pixmap, (createMask) ? &mask : 0);
     }
+#if 0
+    SString id = key;
+fprintf (stderr, "New Map...");
+    for (unsigned int i=0; i<id.size(); i++)
+    {
+      fprintf (stderr, "%02X ", (unsigned int) ((unsigned char)id[i]));
+    }
+fprintf (stderr, "\n");
+#endif
   }
   else
   {
-//fprintf (stderr, "Using cahce X11\n");
+//fprintf (stderr, "Using cache X11\n");
   }
- 
   XCopyArea(impl->display, mask, (Window)id, gc_draw_and, 0, 0, 
      im.getWidth(), im.getHeight(), _x, _y);
   XCopyArea(impl->display, pixmap, (Window)id, gc_draw_or, 0, 0, 
@@ -654,7 +680,7 @@ SX11Window::move (int _x, int _y)
  * @param len is the unber of chars
  */
 void
-SX11Window::bitfont (const SPen& pen, double x, double y, 
+SX11Window::bitfont (const SPen& _pen, double x, double y, 
   void* native, char* data, unsigned int len)
 {
   if (gcFD != (unsigned long) native)
@@ -662,11 +688,11 @@ SX11Window::bitfont (const SPen& pen, double x, double y,
     XSetFont (impl->display, gc, (Font) native);
     gcFD  = (unsigned long) native;
   }
-  if (gcFG != (unsigned long) pen.getForeground().getValue())
+  if (gcFG != (unsigned long) _pen.getForeground().getValue())
   {
     XSetForeground (impl->display, gc,
-       SX11Color(impl, pen.getForeground()).getPixelValue());
-    gcFG = (unsigned long) pen.getForeground().getValue();
+       SX11Color(impl, _pen.getForeground()).getPixelValue());
+    gcFG = (unsigned long) _pen.getForeground().getValue();
   }
   XDrawString16 (impl->display, (Window) id, gc, (int)x, (int)y, 
     (XChar2b*) data, len/2);
@@ -768,11 +794,38 @@ SX11Window::bitpoint (const SColor& fg, int x, int y)
   XDrawPoint (impl->display, (Window) id, gc, x, y);
 }
 
+void
+SX11Window::bitpoints (const SColor& fg, const int* _x, const int* _y, 
+         unsigned int _size)
+{
+  if (_size==0) return;
+  if (gcFG != (unsigned long)  fg.getValue())
+  {
+    XSetForeground (impl->display, gc, SX11Color(impl, fg).getPixelValue());
+    gcFG = (unsigned long)  fg.getValue();
+  }
+  XPoint* p = new XPoint[_size];
+  CHECK_NEW (p);
+  for (unsigned int i=0; i<_size; i++)
+  {
+     p[i].x = _x[i];
+     p[i].y = _y[i];
+  }
+  XDrawPoints (impl->display, (Window) id, gc, p, _size, CoordModeOrigin);
+  delete p;
+}
+
 
 void
 SX11Window::getKeyboardFocus()
 {
   impl->getKeyboardFocus (this);
+  XSetInputFocus (impl->display, (Window) id, RevertToNone, CurrentTime);
+  XIC xic = inputMethod.getIC();
+  if (xic)
+  {
+    XSetICFocus (xic);
+  }
   return;
 }
 
@@ -852,9 +905,9 @@ SX11Window::redraw (bool clear, int _x, int _y, unsigned int _width, unsigned in
  * @param properties provide some attributes to the input method.
  */
 bool
-SX11Window::startInputMethod (const SString& name, const SProperties& prop)
+SX11Window::startInputMethod (const SString& _name, const SProperties& prop)
 {
-  return inputMethod.start (name, prop);
+  return inputMethod.start (_name, prop);
 }
 
 void
@@ -1154,3 +1207,31 @@ SX11Window::wait ()
   }
   SEventHandler::remove (impl->job);
 }
+
+XIC
+SX11Window::getIC ()
+{
+  return inputMethod.getIC();
+}
+const SString&
+SX11Window::getICEncoding () const
+{
+  return inputMethod.getICEncoding();
+}
+
+bool
+SX11Window::isAsciiInput ()
+{
+  return inputMethod.isAsciiInput();
+}
+bool
+SX11Window::isKInput ()
+{
+  return inputMethod.isKInput();
+}
+
+unsigned long
+SX11Window::getWindowID() const
+{
+  return (unsigned long) id;
+}
